#include "Board.h"
#include "tmc/ic/TMC2041/TMC2041.h"

#include "../tmc/stepDir.h"

#define VM_MIN  50   // 5V
#define VM_MAX  286  // 286V (26V + 10%)

#define TMC2041_MAX_VELOCITY STEPDIR_MAX_VELOCITY

#define MOTORS 2

static uint32 right(uint8 motor, int32 velocity);
static uint32 left(uint8 motor, int32 velocity);
static uint32 rotate(uint8 motor, int32 velocity);
static uint32 stop(uint8 motor);
static uint32 moveTo(uint8 motor, int32 position);
static uint32 moveBy(uint8 motor, int32 *ticks);
static uint32 moveProfile(uint8 motor, int32 position);
static uint32 GAP(uint8 type, uint8 motor, int32 *value);
static uint32 SAP(uint8 type, uint8 motor, int32 value);

static void readRegister(uint8 address, int32 *value);
static void writeRegister(uint8 address, int32 value);
static uint32 getMeasuredSpeed(uint8 motor, int32 *value);

static void periodicJob(uint32 tick);
static void checkErrors(uint32 tick);
static void deInit(void);
static uint32 userFunction(uint8 type, uint8 motor, int32 *value);

static uint8 reset();
static void enableDriver(DriverState state);

StepDirectionChannelType stepDirDummmy;
static StepDirectionChannelType *StepDirCh[] = {&stepDirDummmy, &stepDirDummmy};

SPIChannelTypeDef *TMC2041_SPIChannel;
TMC2041TypeDef TMC2041;
ConfigurationTypeDef *TMC2041_config;

typedef struct
{
	IOPinTypeDef  *REFL1_STEP1;
	IOPinTypeDef  *REFL2_STEP2;
	IOPinTypeDef  *REFR1_DIR1;
	IOPinTypeDef  *REFR2_DIR2;
	IOPinTypeDef  *DRV_ENN;
	IOPinTypeDef  *SWIOP;
	IOPinTypeDef  *SWION;
	IOPinTypeDef  *SWSEL;
	IOPinTypeDef  *INT;
	IOPinTypeDef  *PP;
	IOPinTypeDef  *CSN;
} PinsTypeDef;

static PinsTypeDef Pins;

void tmc2041_writeDatagram(u8 motor, uint8 address, uint8 x1, uint8 x2, uint8 x3, uint8 x4)
{
	UNUSED(motor);

	TMC2041_SPIChannel->readWrite(address | TMC2041_WRITE_BIT, FALSE);
	TMC2041_SPIChannel->readWrite(x1, FALSE);
	TMC2041_SPIChannel->readWrite(x2, FALSE);
	TMC2041_SPIChannel->readWrite(x3, FALSE);
	TMC2041_SPIChannel->readWrite(x4, TRUE);

	int value = x1;
	value <<= 8;
	value |= x2;
	value <<= 8;
	value |= x3;
	value <<= 8;
	value |= x4;

	TMC2041_config->shadowRegister[address & 0x7F] = value;
}

void tmc2041_writeInt(u8 motor, uint8 address, int value)
{
	tmc2041_writeDatagram(motor, address, 0xFF & (value>>24), 0xFF & (value>>16), 0xFF & (value>>8), 0xFF & (value>>0));
}

int tmc2041_readInt(u8 motor, uint8 address)
{
	UNUSED(motor);
	address &= 0x7F;

	if(!IS_READABLE(TMC2041.registerAccess[address]))
		return TMC2041_config->shadowRegister[address]; // register not readable -> softwarecopy

	TMC2041_SPIChannel->readWrite(address, FALSE);
	TMC2041_SPIChannel->readWrite(0, FALSE);
	TMC2041_SPIChannel->readWrite(0, FALSE);
	TMC2041_SPIChannel->readWrite(0, FALSE);
	TMC2041_SPIChannel->readWrite(0, TRUE);

	TMC2041_SPIChannel->readWrite(address, FALSE);
	int value = TMC2041_SPIChannel->readWrite(0, FALSE);
	value <<= 8;
	value |= TMC2041_SPIChannel->readWrite(0, FALSE);
	value <<=	8;
	value |= TMC2041_SPIChannel->readWrite(0, FALSE);
	value <<= 8;
	value |= TMC2041_SPIChannel->readWrite(0, TRUE);

	return value;
}

static uint32 rotate(uint8 motor, int32 velocity)
{
	if(motor >= MOTORS)
		return ERROR_MOTOR;

	StepDirCh[motor]->rampMode        = SD_RAMP_MODE_VELOCITY;
	StepDirCh[motor]->targetVelocity  = velocity;

	return ERROR_NONE;
}

static uint32 right(uint8 motor, int32 velocity)
{
	return rotate(motor, velocity);
}

static uint32 left(uint8 motor, int32 velocity)
{
	return rotate(motor, -velocity);
}

static uint32 stop(uint8 motor)
{
	return rotate(motor, 0);
}

static uint32 moveTo(uint8 motor, int32 position)
{
	if(motor >= MOTORS)
		return ERROR_MOTOR;

	StepDirCh[motor]->targetPosition  = position;
	StepDirCh[motor]->rampMode        = SD_RAMP_MODE_POSITIONING;
	StepDirCh[motor]->useProfile      = 0;

	return ERROR_NONE;
}

static uint32 moveBy(uint8 motor, int32 *ticks)
{
	if(motor >= MOTORS)
		return ERROR_MOTOR;

	// determine actual position and add numbers of ticks to move
	*ticks = StepDirCh[motor]->actualPosition + *ticks;

	return moveTo(motor, *ticks);
}

static uint32 moveProfile(uint8 motor, int32 position)
{
	if(motor >= MOTORS)
		return ERROR_MOTOR;

	StepDirCh[motor]->rampMode        = SD_RAMP_MODE_POSITIONING;
	StepDirCh[motor]->useProfile      = 0;
	StepDirCh[motor]->profileFrom     = StepDirCh[motor]->actualPosition;
	StepDirCh[motor]->targetPosition  = StepDirCh[motor]->profileFrom;
	StepDirCh[motor]->profileTo       = position;
	StepDirCh[motor]->useProfile      = 1;

	return ERROR_NONE;
}

// \xml <?xml version="1.0" encoding="UTF-8" standalone="no"?>
// \xml <tmc:product_module author="?" xmlns:tmc="http://www.trinamic.com" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://www.trinamic.com https://www.trinamic.com/fileadmin/xml/TMC_Schema.xsd">
// \xml <name>TMC2041-EVAL</name>
// \xml <name_fw_version>2041</name_fw_version>
// \xml <hardware_version>1.3</hardware_version>
// \xml <software_version>3.01</software_version>
static uint32 handleParameter(u8 readWrite, u8 motor, u8 type, int32 *value)
{
	u32 errors = ERROR_NONE;
	int tempValue;

	if(motor >= MOTORS)
		return ERROR_MOTOR;

	// \xml <axis_parameters axis="0">
	switch(type)
	{
	case 0:
		// \xml <tmc:axis_parameter type="0" name="Target position" identifier="par::TargetPosition">
		// \xml   <desc>The desired target position in position mode.</desc>
		// \xml   <tmc:value_signed min="#s32_MIN#" max="#s32_MAX#" default="0" unit="[usteps]" access="RW"></tmc:value_signed>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->targetPosition;
		} else if(readWrite == WRITE) {
			StepDirCh[motor]->targetPosition = *value;
		}
		break;
	case 1:
		// \xml <tmc:axis_parameter type="1" name="Actual position" identifier="par::ActualPosition">
		// \xml   <desc>The actual position of the motor. Stop the motor before overwriting it. Should normally only be overwritten for reference position setting.</desc>
		// \xml   <tmc:value_signed min="#s32_MIN#" max="#s32_MAX#" default="0" unit="[usteps]" access="RW"></tmc:value_signed>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->actualPosition;
		} else if(readWrite == WRITE) {
			StepDirCh[motor]->actualPosition = *value;
		}
		break;
	case 2:
		// todo XML 3: Velocity ist signed, nicht unsigned (LH)
		// \xml <tmc:axis_parameter type="2" name="Target speed" identifier="par::TargetVelocity">
		// \xml   <desc>The desired speed in velocity mode. Not valid in position mode.</desc>
		// \xml   <tmc:value_unsigned min="0" max="#TMC2041_MAX_VELOCITY#" default="0" unit="[pps]" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->targetVelocity;
		} else if(readWrite == WRITE) {
			StepDirCh[motor]->targetVelocity = *value;
		}
		break;
	case 3:
		// todo CHECK 3: min max actually velocity min and velocity max? (JE) #1
		// \xml <tmc:axis_parameter type="3" name="Actual speed" identifier="par::ActualVelocity">
		// \xml   <desc>The actual speed of the motor.</desc>
		// \xml   <tmc:value_signed min="#s32_MIN#" max="#s32_MAX#" default="0" unit="[pps]" access="R"></tmc:value_signed>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->actualVelocity;
			*value = CAST_Sn_TO_S32(*value, 24);
		} else if(readWrite == WRITE) {
			errors |= ERROR_TYPE;
		}
		break;
	case 4:
		// \xml <tmc:axis_parameter type="4" name="Maximum speed" identifier="par::MaxVelocity">
		// \xml   <desc>The maximum speed used for positioning ramps.</desc>
		// \xml   <tmc:value_unsigned min="0" max="#TMC2041_MAX_VELOCITY#" default="0" unit="[pps]" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value	= StepDirCh[motor]->maxPositioningSpeed;
		} else if(readWrite == WRITE) {
			StepDirCh[motor]->maxPositioningSpeed = abs(*value);
		}
		break;
	case 5:
		// \xml <tmc:axis_parameter type="5" name="Maximum acceleration" identifier="par::MaxAcceleration">
		// \xml   <desc>Maximum acceleration in positioning ramps. Acceleration and deceleration value in velocity mode.</desc>
		// \xml   <tmc:value_unsigned min="0" max="#u16_MAX#" default="0" unit="[pps/s]" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value	= StepDirCh[motor]->actualAcceleration;
		} else if(readWrite == WRITE) {
			StepDirCh[motor]->actualAcceleration = *value;
		}
		break;
	case 6:
		// \xml <tmc:axis_parameter type="6" name="Maximum current" identifier="par::MaxCurrent">
		// \xml   <desc>Motor current used when motor is running. Where 0 = 3.125% up to 31 = 100% of maximum module current. Calculated by: maximum module current * (value + 1)/ 32.</desc>
		// \xml   <tmc:value_unsigned min="0" max="31" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2041_FIELD_READ(motor, TMC2041_IHOLD_IRUN(motor), TMC2041_IRUN_MASK, TMC2041_IRUN_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2041_FIELD_UPDATE(motor, TMC2041_IHOLD_IRUN(motor), TMC2041_IRUN_MASK, TMC2041_IRUN_SHIFT, *value); // todo CHECK 3: check functionality (ED) #1
		}
		break;
	case 7:
		// \xml <tmc:axis_parameter type="7" name="Standby current" identifier="par::StandbyCurrent">
		// \xml   <desc>The current used when the motor is not running. Where 0 = 3.125% up to 31 = 100% of maximum module current. Calculated by: maximum module current * (value + 1)/ 32. This value should be as low as possible so that the motor can cool down when it is not moving. Please see also parameter 214.</desc>
		// \xml   <tmc:value_unsigned min="0" max="31" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2041_FIELD_READ(motor, TMC2041_IHOLD_IRUN(motor), TMC2041_IHOLD_MASK, TMC2041_IHOLD_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2041_FIELD_UPDATE(motor, TMC2041_IHOLD_IRUN(motor), TMC2041_IHOLD_MASK, TMC2041_IHOLD_SHIFT, *value);
		}
		break;
	case 8:
		// \xml <tmc:axis_parameter type="8" name="Position reached flag" identifier="par::PositionReachedFlag">
		// \xml	  <desc>This flag is always set when target position and actual position are equal.</desc>
		// \xml	  <tmc:value_bool desc_false="Target position not reached" desc_true="Target position reached" default="false" access="R"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->targetReached;
		} else if(readWrite == WRITE)
			errors |= ERROR_TYPE;
		break;
	case 28:
		// \xml <tmc:axis_parameter type="28" name="High speed fullstep mode" identifier="par::HighSpeedFullstepMode">
		// \xml   <desc>Switch to fullstep mode when measured speed is higher than axis parameter 23 when set to 1.</desc>
		// \xml   <tmc:value_bool desc_false="Inactive" desc_true="Active" default="false" access="RW"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2041_FIELD_READ(motor, TMC2041_CHOPCONF(motor), TMC2041_VHIGHFS_MASK, TMC2041_VHIGHFS_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2041_FIELD_UPDATE(motor, TMC2041_CHOPCONF(motor), TMC2041_VHIGHFS_MASK, TMC2041_VHIGHFS_SHIFT, *value);
		}
		break;
	case 29:
		if(readWrite == READ) {
			*value = StepDir.ch1->actualVelocity;
		} else if(readWrite == WRITE)
			errors |= ERROR_TYPE;
		break;
	case 140:
		// \xml <tmc:axis_parameter type="140" name="Microstep Resolution" identifier="par::MicrostepResolution">
		// \xml   <desc>Microstep resolutions per full step:</desc>
		// \xml   <tmc:value_choice default="256" unit="" access="RW">
		// \xml     <choice_element value="1" desc="fullstep"></choice_element>
		// \xml     <choice_element value="2" desc="halfstep"></choice_element>
		// \xml     <choice_element value="4" desc="4 microsteps"></choice_element>
		// \xml     <choice_element value="8" desc="8 microsteps"></choice_element>
		// \xml     <choice_element value="16" desc="16 microsteps"></choice_element>
		// \xml     <choice_element value="32" desc="32 microsteps"></choice_element>
		// \xml     <choice_element value="64" desc="64 microsteps"></choice_element>
		// \xml     <choice_element value="128" desc="128 microsteps"></choice_element>
		// \xml     <choice_element value="256" desc="256 microsteps"></choice_element>
		// \xml   </tmc:value_choice>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2041_FIELD_READ(motor, TMC2041_CHOPCONF(motor), TMC2041_MRES_MASK, TMC2041_MRES_SHIFT);
		} else if(readWrite == WRITE) {
			switch(*value)
			{
			case 1:    *value = 8;   break;
			case 2:    *value = 7;   break;
			case 4:    *value = 6;   break;
			case 8:    *value = 5;   break;
			case 16:   *value = 4;   break;
			case 32:   *value = 3;   break;
			case 64:   *value = 2;   break;
			case 128:  *value = 1;   break;
			case 256:  *value = 0;   break;
			default:   *value = -1;  break;
			}

			if(*value != -1)
			{
				TMC2041_FIELD_UPDATE(motor, TMC2041_CHOPCONF(motor), TMC2041_MRES_MASK, TMC2041_MRES_SHIFT, *value);
			}
			//else TMCL.reply->Status = REPLY_INVALID_VALUE;
		}
		break;
	case 162:
		// \xml <tmc:axis_parameter type="162" name="Chopper blank time" identifier="par::ChopperBlankTime">
		// \xml   <desc>Selects the comparator blank time. This time needs to safely cover the switching event and the duration of the ringing on the sense resistor. Normally leave at the default value.</desc>
		// \xml   <tmc:value_unsigned min="0" max="3" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2041_FIELD_READ(motor, TMC2041_CHOPCONF(motor), TMC2041_TBL_MASK, TMC2041_TBL_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2041_FIELD_UPDATE(motor, TMC2041_CHOPCONF(motor), TMC2041_TBL_MASK, TMC2041_TBL_SHIFT, *value);
		}
		break;
	case 163:
		// \xml <tmc:axis_parameter type="163" name="Constant TOff Mode" identifier="par::ConstantTOffMode">
		// \xml   <desc>Selection of the chopper mode.</desc>
		// \xml   <tmc:value_bool desc_false="spread cycle" desc_true="classic constant off time" default="false" access="RW"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2041_FIELD_READ(motor, TMC2041_CHOPCONF(motor), TMC2041_CHM_MASK, TMC2041_CHM_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2041_FIELD_UPDATE(motor, TMC2041_CHOPCONF(motor), TMC2041_CHM_MASK, TMC2041_CHM_SHIFT, *value);
		}
		break;
	case 164:
		// \xml <tmc:axis_parameter type="164" name="Disable fast decay comparator" identifier="par::DisableFastDecayComparator">
		// \xml   <desc>See parameter 163. For "classic const. off time", setting this parameter to "1" will disable current comparator usage for termination of fast decay cycle.</desc>
		// \xml   <tmc:value_bool desc_false="enable comparator" desc_true="disable comparator" default="false" access="RW"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2041_FIELD_READ(motor, TMC2041_CHOPCONF(motor), TMC2041_DISFDCC_MASK, TMC2041_DISFDCC_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2041_FIELD_UPDATE(motor, TMC2041_CHOPCONF(motor), TMC2041_DISFDCC_MASK, TMC2041_DISFDCC_SHIFT, *value);
		}
		break;
	case 165:
		// \xml <tmc:axis_parameter type="165" name="Chopper hysteresis end / fast decay time" identifier="par::ChopperHysteresisEnd">
		// \xml   <desc>See parameter 163. For "spread cycle" chopper mode this parameter will set / return the hysteresis end setting (hysteresis end value after a number of decrements). For "classic const. off time" chopper mode this parameter will set / return the fast decay time.</desc>
		// \xml   <tmc:value_unsigned min="0" max="15" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		tempValue = tmc2041_readInt(motor, TMC2041_CHOPCONF(motor));
		if(readWrite == READ) {
			if(tempValue & TMC2041_CHM_MASK) // Chopper hysteresis end
			{
				*value = TMC2041_FIELD_READ(motor, TMC2041_CHOPCONF(motor), TMC2041_HEND_MASK, TMC2041_HEND_SHIFT);
			}
			else // fast decay time
			{
				*value = TMC2041_FIELD_READ(motor, TMC2041_CHOPCONF(motor), TMC2041_TFD_ALL_MASK, TMC2041_TFD_ALL_SHIFT);
				if(tempValue & TMC2041_TFD_3_MASK) // add MSB of fast decay time to *value
					*value |= 1<<3;
			}
		} else if(readWrite == WRITE) {
			if(tempValue & TMC2041_CHM_MASK) // Chopper hysteresis end
			{
				TMC2041_FIELD_UPDATE(motor, TMC2041_CHOPCONF(motor), TMC2041_HEND_MASK, TMC2041_HEND_SHIFT, *value);
			}
			else // fast decay time
			{
				TMC2041_FIELD_UPDATE(motor, TMC2041_CHOPCONF(motor), TMC2041_TFD_ALL_MASK, TMC2041_TFD_ALL_SHIFT, *value);
				TMC2041_FIELD_UPDATE(motor, TMC2041_CHOPCONF(motor), TMC2041_TFD_3_MASK, TMC2041_TFD_3_SHIFT, (*value & (1<<3))? 1:0);
			}
		}
		break;
	case 166:
		// \xml <tmc:axis_parameter type="166" name="Chopper hysteresis start / sine wave offset" identifier="par::ChopperHysteresisStart">
		// \xml   <desc>See parameter 163. For "spread cycle" chopper mode this parameter will set / return the Hysteresis start setting (please note that this value is an offset to the hysteresis end value). For "classic const. off time" chopper mode this parameter will set / return the sine wave offset.</desc>
		// \xml   <tmc:value_unsigned min="0" max="8" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		tempValue = tmc2041_readInt(motor, TMC2041_CHOPCONF(motor));
		if(readWrite == READ) {
			if(tempValue & TMC2041_CHM_MASK) // Chopper hysteresis start
			{
				*value = TMC2041_FIELD_READ(motor, TMC2041_CHOPCONF(motor), TMC2041_HSTRT_MASK, TMC2041_HSTRT_SHIFT);
			}
			else // sine wave offset
			{
				*value = TMC2041_FIELD_READ(motor, TMC2041_CHOPCONF(motor), TMC2041_OFFSET_MASK, TMC2041_OFFSET_SHIFT);
			}
		} else if(readWrite == WRITE) {
			if(tempValue & TMC2041_CHM_MASK) // Chopper hysteresis start
			{
				TMC2041_FIELD_UPDATE(motor, TMC2041_CHOPCONF(motor), TMC2041_HSTRT_MASK, TMC2041_HSTRT_SHIFT, *value);
			}
			else // sine wave offset
			{
				TMC2041_FIELD_UPDATE(motor, TMC2041_CHOPCONF(motor), TMC2041_OFFSET_MASK, TMC2041_OFFSET_SHIFT, *value);
			}
		}
		break;
	case 167:
		// \xml <tmc:axis_parameter type="167" name="Chopper off time" identifier="par::TOff">
		// \xml   <desc>The off time setting controls the minimum chopper frequency. An off time within the range of 5us to 20us will fit. Off time setting for constant t Off chopper: NCLK = 12 + 32  tOFF (Minimum is 64 clocks) Setting this parameter to zero completely disables all driver transistors and the motor can freewheel.</desc>
		// \xml   <tmc:value_unsigned min="0" max="15" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2041_FIELD_READ(motor, TMC2041_CHOPCONF(motor), TMC2041_TOFF_MASK, TMC2041_TOFF_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2041_FIELD_UPDATE(motor, TMC2041_CHOPCONF(motor), TMC2041_TOFF_MASK, TMC2041_TOFF_SHIFT, *value);
		}
		break;
	case 168:
		// \xml <tmc:axis_parameter type="168" name="smartEnergy current minimum (SEIMIN)" identifier="par::SEIMIN">
		// \xml   <desc>Sets the lower motor current limit for coolStep operation by scaling the maximum current (see axis parameter 6) value.</desc>
		// \xml   <tmc:value_bool desc_false="1/2 of CS" desc_true="1/4 of CS" default="false" access="RW"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2041_FIELD_READ(motor, TMC2041_COOLCONF(motor), TMC2041_SEIMIN_MASK, TMC2041_SEIMIN_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2041_FIELD_UPDATE(motor, TMC2041_COOLCONF(motor), TMC2041_SEIMIN_MASK, TMC2041_SEIMIN_SHIFT, *value);
		}
		break;
	case 169:
		// \xml <tmc:axis_parameter type="169" name="smartEnergy current down step" identifier="par::SECDS">
		// \xml   <desc>Sets the number of stallGuard2 readings above the upper threshold necessary for each current decrement of the motor current. Number of stallGuard2 measurements per decrement: Scaling: 0. . . 3: 32, 8, 2, 1. 0: slow decrement, 3: fast decrement</desc>
		// \xml   <tmc:value_unsigned min="0" max="3" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2041_FIELD_READ(motor, TMC2041_COOLCONF(motor), TMC2041_SEDN_MASK, TMC2041_SEDN_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2041_FIELD_UPDATE(motor, TMC2041_COOLCONF(motor), TMC2041_SEDN_MASK, TMC2041_SEDN_SHIFT, *value);
		}
		break;
	case 170:
		// \xml <tmc:axis_parameter type="170" name="smartEnergy hysteresis" identifier="par::smartEnergyHysteresis">
		// \xml   <desc>Sets the distance between the lower and the upper threshold for stallGuard2 reading. Above the upper threshold the motor current becomes decreased. Hysteresis: ([AP172] + 1) * 32. Upper stallGuard threshold: ([AP172] + [AP170] + 1) * 32</desc>
		// \xml   <tmc:value_unsigned min="0" max="15" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2041_FIELD_READ(motor, TMC2041_COOLCONF(motor), TMC2041_SEMAX_MASK, TMC2041_SEMAX_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2041_FIELD_UPDATE(motor, TMC2041_COOLCONF(motor), TMC2041_SEMAX_MASK, TMC2041_SEMAX_SHIFT, *value);
		}
		break;
	case 171:
		// \xml <tmc:axis_parameter type="171" name="smartEnergy current up step" identifier="par::SECUS">
		// \xml   <desc>Sets the current increment step. The current becomes incremented for each measured stallGuard2 value below the lower threshold see smartEnergy hysteresis start). Current increment step size: Scaling: 0. . . 3: 1, 2, 4, 8. 0: slow increment, 3: fast increment / fast reaction to rising load</desc>
		// \xml   <tmc:value_unsigned min="0" max="3" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2041_FIELD_READ(motor, TMC2041_COOLCONF(motor), TMC2041_SEUP_MASK, TMC2041_SEUP_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2041_FIELD_UPDATE(motor, TMC2041_COOLCONF(motor), TMC2041_SEUP_MASK, TMC2041_SEUP_SHIFT, *value);
		}
		break;
	case 172:
		// \xml <tmc:axis_parameter type="172" name="smartEnergy hysteresis start" identifier="par::smartEnergyHysteresisStart">
		// \xml   <desc>The lower threshold for the stallGuard2 value (see smart Energy current up step).</desc>
		// \xml   <tmc:value_unsigned min="0" max="15" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2041_FIELD_READ(motor, TMC2041_COOLCONF(motor), TMC2041_SEMIN_MASK, TMC2041_SEMIN_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2041_FIELD_UPDATE(motor, TMC2041_COOLCONF(motor), TMC2041_SEMIN_MASK, TMC2041_SEMIN_SHIFT, *value);
		}
		break;
	case 173:
		// \xml <tmc:axis_parameter type="173" name="stallGuard2 filter enable" identifier="par::SG2FilterEnable">
		// \xml   <desc>Enables the stallGuard2 filter for more precision of the movement. If set, reduces the measurement frequency to one measurement per four fullsteps. In most cases it is expedient to set the filtered mode before using coolStep. Use the standard mode for step loss detection.</desc>
		// \xml   <tmc:value_bool desc_false="standard mode" desc_true="filtered mode" default="false" access="RW"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2041_FIELD_READ(motor, TMC2041_COOLCONF(motor), TMC2041_SFILT_MASK, TMC2041_SFILT_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2041_FIELD_UPDATE(motor, TMC2041_COOLCONF(motor), TMC2041_SFILT_MASK, TMC2041_SFILT_SHIFT, *value);
		}
		break;
	case 174:
		// \xml <tmc:axis_parameter type="174" name="stallGuard2 threshold" identifier="par::SG2Threshold">
		// \xml   <desc>This signed value controls stallGuard2 threshold level for stall output and sets the optimum measurement range for readout. A lower value gives a higher sensitivity. Zero is the starting value. A higher value makes stallGuard2 less sensitive and requires more torque to indicate a stall.</desc>
		// \xml   <tmc:value_signed min="-64" max="63" default="0" unit="" access="RW"></tmc:value_signed>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2041_FIELD_READ(motor, TMC2041_COOLCONF(motor), TMC2041_SGT_MASK, TMC2041_SGT_SHIFT);
			*value = CAST_Sn_TO_S32(*value, 7);
		} else if(readWrite == WRITE) {
			TMC2041_FIELD_UPDATE(motor, TMC2041_COOLCONF(motor), TMC2041_SGT_MASK, TMC2041_SGT_SHIFT, *value);
		}
		break;
	case 179:
		// \xml <tmc:axis_parameter type="179" name="VSense" identifier="par::VSense">
		// \xml   <desc>Sense resistor voltage:</desc>
		// \xml   <tmc:value_bool desc_false="High sense resistor voltage" desc_true="Low sense resistor voltage" default="false" access="RW"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2041_FIELD_READ(motor, TMC2041_CHOPCONF(motor), TMC2041_VSENSE_MASK, TMC2041_VSENSE_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2041_FIELD_UPDATE(motor, TMC2041_CHOPCONF(motor), TMC2041_VSENSE_MASK, TMC2041_VSENSE_SHIFT, *value);
		}
		break;
	case 180:
		// \xml <tmc:axis_parameter type="180" name="smartEnergy actual current" identifier="par::smartEnergyActualCurrent">
		// \xml   <desc>This status value provides the actual motor current setting as controlled by coolStep. The value goes up to the CS value and down to the portion of CS as specified by SEIMIN. Actual motor current scaling factor: 0. . . 31: 1/32, 2/32, . . . 32/32</desc>
		// \xml   <tmc:value_unsigned min="0" max="31" default="0" unit="" access="R"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2041_FIELD_READ(motor, TMC2041_DRVSTATUS(motor), TMC2041_CS_ACTUAL_MASK, TMC2041_CS_ACTUAL_SHIFT);
		} else if(readWrite == WRITE)
			errors |= ERROR_TYPE;
		break;
	case 181:
		// \xml <tmc:axis_parameter type="181" name="smartEnergy stall velocity" identifier="par::smartEnergyStallVelocity">
		// \xml   <desc>Velocity from which stop on stall feature is active</desc>
		// \xml   <tmc:value_unsigned min="0" max="#TMC2041_MAX_VELOCITY#" default="0" unit="[pps]" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->stopOnStallVelocityTreshold;
		} else if(readWrite == WRITE) {
			StepDirCh[motor]->stopOnStallVelocityTreshold = *value;
		}
		break;
	case 184:
		// \xml <tmc:axis_parameter type="184" name="Random TOff mode" identifier="par::RandomTOffMode">
		// \xml   <desc>Enable / disable random TOff mode.</desc>
		// \xml   <tmc:value_bool desc_false="Chopper off time is fixed" desc_true="Chopper off time is random" default="false" access="RW"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2041_FIELD_READ(motor, TMC2041_CHOPCONF(motor), TMC2041_RNDTF_MASK, TMC2041_RNDTF_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2041_FIELD_UPDATE(motor, TMC2041_CHOPCONF(motor), TMC2041_RNDTF_MASK, TMC2041_RNDTF_SHIFT, *value);
		}
		break;
	case 206:
		// \xml <tmc:axis_parameter type="206" name="Load value" identifier="par::LoadValue">
		// \xml   <desc>Actual current control scaling for monitoring smart energy current scaling or automatic current scaling.</desc>
		// \xml   <tmc:value_unsigned min="0" max="#u10_MAX#" default="0" unit="" access="R"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2041_FIELD_READ(motor, TMC2041_DRVSTATUS(motor), TMC2041_SG_RESULT_MASK, TMC2041_SG_RESULT_SHIFT);
		} else if(readWrite == WRITE) {
			errors |= ERROR_TYPE;
		}
		break;
	default:
		errors |= ERROR_TYPE;
		break;
	}
	// \xml </axis_parameters>

	// \xml <axis_parameters axis="1">
	//copy axis parameters from above ~~ stands for automatic insertion from following tag
	// \xml 	~~<axis_parameters axis="0">
	// \xml </axis_parameters>
	return errors;
}
// \xml </tmc:product_module>

static uint32 SAP(uint8 type, uint8 motor, int32 value)
{
	return handleParameter(WRITE, motor, type, &value);
}

static uint32 GAP(uint8 type, uint8 motor, int32 *value)
{
	return handleParameter(READ, motor, type, value);
}

static uint32 getMeasuredSpeed(uint8 motor, int32 *value)
{
	if(motor >= MOTORS)
		return ERROR_MOTOR;

	switch(motor)
	{
	case 0:
		*value = StepDir.ch1->actualVelocity;
		break;
	default:
		return ERROR_MOTOR;
		break;
	}
	return ERROR_NONE;
}

static void writeRegister(uint8 address, int32 value)
{
	tmc2041_writeInt(0, address, value);
}

static void readRegister(uint8 address, int32 *value)
{
	*value = tmc2041_readInt(0, address);
}

static void periodicJob(uint32 tick)
{
	int motor;
	for(motor = 0; motor < MOTORS; motor++)
	{
		tmc2041_periodicJob(motor, tick, &TMC2041, TMC2041_config);
	}
}

static void checkErrors(uint32 tick)
{
	UNUSED(tick);
	Evalboards.ch2.errors = 0;
}

static uint32 userFunction(uint8 type, uint8 motor, int32 *value)
{
	uint32 errors = 0;

	UNUSED(motor);

	switch(type)
	{
	case 1:  // read interrupt pin INT
		*value = (HAL.IOs->config->isHigh(Pins.INT)) ? 1 : 0;
		break;
	case 2:  // read position compare pin PP
		*value = (HAL.IOs->config->isHigh(Pins.PP)) ? 1 : 0;
		break;
	default:
		errors |= ERROR_TYPE;
		break;
	}

	return errors;
}

static void deInit(void)
{
	enableDriver(DRIVER_DISABLE);
	HAL.IOs->config->reset(Pins.DRV_ENN);
	HAL.IOs->config->reset(Pins.INT);
	HAL.IOs->config->reset(Pins.PP);

	StepDir_deInit();
	Timer.deInit();
}

static uint8 reset()
{
	for(uint8 motor = 0; motor < MOTORS; motor++)
		if(StepDirCh[motor]->actualVelocity != 0)
			return 0;

	tmc2041_reset(TMC2041_config);
	StepDir_init(MOTORS);
	StepDirCh[0]           = StepDir.ch1;
	StepDirCh[0]->stepOut  = Pins.REFL1_STEP1;
	StepDirCh[0]->dirOut   = Pins.REFR1_DIR1;

	StepDirCh[1]           = StepDir.ch2;
	StepDirCh[1]->stepOut  = Pins.REFL2_STEP2;
	StepDirCh[1]->dirOut   = Pins.REFR2_DIR2;

	return 1;
}

static uint8 restore()
{
	return tmc2041_restore(TMC2041_config);
}

static void enableDriver(DriverState state)
{
	if(state == DRIVER_USE_GLOBAL_ENABLE)
		state = Evalboards.driverEnable;

	if(state ==  DRIVER_DISABLE)
		HAL.IOs->config->setHigh(Pins.DRV_ENN);
	else if((state == DRIVER_ENABLE) && (Evalboards.driverEnable == DRIVER_ENABLE))
		HAL.IOs->config->setLow(Pins.DRV_ENN);
}

void TMC2041_init(void)
{
	tmc2041_initConfig(&TMC2041);

	Pins.DRV_ENN      = &HAL.IOs->pins->DIO0;
	Pins.PP           = &HAL.IOs->pins->DIO4;
	Pins.INT          = &HAL.IOs->pins->DIO5;
	Pins.REFL1_STEP1  = &HAL.IOs->pins->DIO6;
	Pins.REFR1_DIR1   = &HAL.IOs->pins->DIO7;
	Pins.REFL2_STEP2  = &HAL.IOs->pins->DIO8;
	Pins.REFR2_DIR2   = &HAL.IOs->pins->DIO9;
	Pins.CSN          = &HAL.IOs->pins->SPI2_CSN0;

	HAL.IOs->config->toOutput(Pins.REFL1_STEP1);
	HAL.IOs->config->toOutput(Pins.REFR1_DIR1);
	HAL.IOs->config->toOutput(Pins.REFL2_STEP2);
	HAL.IOs->config->toOutput(Pins.REFR2_DIR2);
	HAL.IOs->config->toOutput(Pins.DRV_ENN);
	HAL.IOs->config->toOutput(Pins.CSN);

	TMC2041_SPIChannel = &HAL.SPI->ch2;
	TMC2041_SPIChannel->CSN = Pins.CSN;

	TMC2041_config = &TMCDriver.config;

	StepDir_init(MOTORS);
	StepDirCh[0] = StepDir.ch1;
	StepDirCh[0]->stepOut  = Pins.REFL1_STEP1;
	StepDirCh[0]->dirOut   = Pins.REFR1_DIR1;

	StepDirCh[1] = StepDir.ch2;
	StepDirCh[1]->stepOut  = Pins.REFL2_STEP2;
	StepDirCh[1]->dirOut   = Pins.REFR2_DIR2;

	Evalboards.ch2.config->reset        = reset;
	Evalboards.ch2.config->restore      = restore;
	Evalboards.ch2.config->state        = CONFIG_RESET;
	Evalboards.ch2.config->configIndex  = 0;

	Evalboards.ch2.rotate               = rotate;
	Evalboards.ch2.right                = right;
	Evalboards.ch2.left                 = left;
	Evalboards.ch2.stop                 = stop;
	Evalboards.ch2.GAP                  = GAP;
	Evalboards.ch2.SAP                  = SAP;
	Evalboards.ch2.moveTo               = moveTo;
	Evalboards.ch2.moveBy               = moveBy;
	Evalboards.ch2.moveProfile          = moveProfile;
	Evalboards.ch2.writeRegister        = writeRegister;
	Evalboards.ch2.readRegister         = readRegister;
	Evalboards.ch2.periodicJob          = periodicJob;
	Evalboards.ch2.userFunction         = userFunction;
	Evalboards.ch2.getMeasuredSpeed     = getMeasuredSpeed;
	Evalboards.ch2.enableDriver         = enableDriver;
	Evalboards.ch2.checkErrors          = checkErrors;
	Evalboards.ch2.numberOfMotors       = MOTORS;
	Evalboards.ch2.VMMin                = VM_MIN;
	Evalboards.ch2.VMMax                = VM_MAX;
	Evalboards.ch2.deInit               = deInit;

	enableDriver(DRIVER_USE_GLOBAL_ENABLE);

	Timer.init();
	Timer.setDuty(0);
}
