#include "Rhino_standalone.h"

#include "Board.h"
#include "tmc/ic/TMC5130/TMC5130_Register.h"
#include "../tmc/stepDir.h"
#include "../hal/Timer.h"

//#define VM_MIN  50   // VM[V/10] min
//#define VM_MAX  480  // VM[V/10] max

#define MOTORS 1

#define TMC2100_MAX_VELOCITY      STEPDIR_MAX_VELOCITY
#define TMC2100_MAX_ACCELERATION  16777215

static uint32 right(uint8 motor, int32 velocity);
static uint32 left(uint8 motor, int32 velocity);
static uint32 rotate(uint8 motor, int32 velocity);
static uint32 stop(uint8 motor);
static uint32 moveTo(uint8 motor, int32 position);
static uint32 moveBy(uint8 motor, int32 *ticks);
static uint32 GAP(uint8 type, uint8 motor, int32 *value);
static uint32 SAP(uint8 type, uint8 motor, int32 value);
static void readRegister(uint8 address, int32 *value);
static void writeRegister(uint8 address, int32 value);

//static void periodicJob	(uint32 tick);
//static void checkErrors	(uint32 tick);
static void deInit(void);
//static uint32 userFunction(uint8 type, uint8 motor, int32 *value);

static uint8 reset();
static void enableDriver(DriverState state);

static u32 setStandAloneSettings(uint8 i, int32 value);
static u32 getStandAloneSettings(uint8 i, int32 *value);

static IO_States lastEnable = IOS_OPEN; //TMCRhinoSA.Enable_StandStillPowerDownSettings.ENABLED_034;

typedef struct
{
	IOPinTypeDef  *CFG3;
	IOPinTypeDef  *CFG2;
	IOPinTypeDef  *CFG1;
	IOPinTypeDef  *CFG0;
	IOPinTypeDef  *STEP;
	IOPinTypeDef  *DIR;
	IOPinTypeDef  *CFG4;
	IOPinTypeDef  *CFG5;
	IOPinTypeDef  *ERROR;
	IOPinTypeDef  *INDEX;
	IOPinTypeDef  *CFG6_ENN;
	IOPinTypeDef  *AIN_REF_SW;
	IOPinTypeDef  *AIN_REF_PWM;

} PinsTypeDef;

static PinsTypeDef Pins;

StepDirectionChannelType stepDirDummmy;
static StepDirectionChannelType *StepDirCh[] = {&stepDirDummmy};

static uint32 rotate(uint8 motor, int32 velocity)
{
	if(motor >= MOTORS)
		return ERROR_MOTOR;

	StepDirCh[motor]->rampMode        = SD_RAMP_MODE_VELOCITY;
	StepDirCh[motor]->targetVelocity  = velocity;

	return ERROR_NONE;
}

static uint32 right(uint8 motor, int32 velocity)
{
	return rotate(motor, velocity);
}

static uint32 left(uint8 motor, int32 velocity)
{
	return rotate(motor, -velocity);
}

static uint32 stop(uint8 motor)
{
	return rotate(motor, 0);
}

static uint32 moveTo(uint8 motor, int32 position)
{
	if(motor >= MOTORS)
		return ERROR_MOTOR;

	StepDirCh[motor]->targetPosition  = position;
	StepDirCh[motor]->rampMode        = SD_RAMP_MODE_POSITIONING;
	StepDirCh[motor]->useProfile      = 0;

	return ERROR_NONE;
}

static uint32 moveBy(uint8 motor, int32 *ticks)
{
	if(motor >= MOTORS)
		return ERROR_MOTOR;

	// determine actual position and add numbers of ticks to move
	*ticks = StepDirCh[motor]->actualPosition + *ticks;

	return moveTo(motor, *ticks);
}

// \xml <?xml version="1.0" encoding="UTF-8" standalone="no"?>
// \xml <tmc:product_module author="BS" xmlns:tmc="http://www.trinamic.com" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://www.trinamic.com https://www.trinamic.com/fileadmin/xml/TMC_Schema.xsd">
// \xml <name>TMC2100-EVAL</name>
// \xml <name_fw_version>2100</name_fw_version>
// \xml <hardware_version>1.0</hardware_version>
// \xml <software_version>2.91</software_version>
static uint32 handleParameter(u8 readWrite, u8 motor, u8 type, int32 *value)
{
	u32 errors = ERROR_NONE;

	if(motor >= MOTORS)
		return ERROR_MOTOR;

	// \xml <axis_parameters axis="0">
	switch(type)
	{
	case 0:
		// \xml <tmc:axis_parameter type="0" name="Target position" identifier="par::TargetPosition">
		// \xml   <desc>The desired target position in position mode.</desc>
		// \xml   <tmc:value_signed min="#s32_MIN#" max="#s32_MAX#" default="0" unit="[usteps]" access="RW"></tmc:value_signed>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->targetPosition;
		} else if(readWrite == WRITE) {
			StepDirCh[motor]->targetPosition = *value;
		}
		break;
	case 1:
		// \xml <tmc:axis_parameter type="1" name="Actual position" identifier="par::ActualPosition">
		// \xml   <desc>The actual position of the motor. Stop the motor before overwriting it. Should normally only be overwritten for reference position setting.</desc>
		// \xml   <tmc:value_signed min="#s32_MIN#" max="#s32_MAX#" default="0" unit="[usteps]" access="RW"></tmc:value_signed>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->actualPosition;
		} else if(readWrite == WRITE) {
			StepDirCh[motor]->actualPosition = *value;
			//set also target position to prevent movement
			StepDirCh[motor]->targetPosition = *value;
		}
		break;
	case 2:
		// \xml <tmc:axis_parameter type="2" name="Target speed" identifier="par::TargetVelocity">
		// \xml   <desc>The desired speed in velocity mode. Not valid in position mode.</desc>
		// \xml   <tmc:value_signed min="-#TMC2100_MAX_VELOCITY#" max="#TMC2100_MAX_VELOCITY#" default="0" unit="[pps]" access="RW"></tmc:value_signed>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->targetVelocity;
		} else if(readWrite == WRITE) {
			StepDirCh[motor]->targetVelocity = *value;
		}
		break;
	case 3:
		// \xml <tmc:axis_parameter type="3" name="Actual speed" identifier="par::ActualVelocity">
		// \xml   <desc>The actual speed of the motor.</desc>
		// \xml   <tmc:value_signed min="-#TMC2100_MAX_VELOCITY#" max="#TMC2100_MAX_VELOCITY#" default="0" unit="[pps]" access="R"></tmc:value_signed>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->actualVelocity;
		} else if(readWrite == WRITE) {
			errors |= ERROR_TYPE;
		}
		break;
	case 4:
		// \xml <tmc:axis_parameter type="4" name="Maximum speed" identifier="par::MaxVelocity">
		// \xml   <desc>The maximum speed used for positioning ramps.</desc>
		// \xml   <tmc:value_unsigned min="0" max="#TMC2100_MAX_VELOCITY#" default="150000" unit="[pps]" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->maxPositioningSpeed;
		} else if(readWrite == WRITE) {
			StepDirCh[motor]->maxPositioningSpeed = abs(*value);
		}
		break;
	case 5:
		// \xml <tmc:axis_parameter type="5" name="Maximum acceleration" identifier="par::MaxAcceleration">
		// \xml   <desc>Maximum acceleration in positioning ramps. Acceleration and deceleration value in velocity mode.</desc>
		// \xml   <tmc:value_unsigned min="0" max="#TMC2100_MAX_ACCELERATION#" default="1000" unit="[pps/s]" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->actualAcceleration;
		} else if(readWrite == WRITE) {
			StepDirCh[motor]->actualAcceleration = *value;
		}
		break;
	// getStandaloneSettin APs have been copied from from types 14-19 but also left here for backwards-compatibility
	case 6:
	case 7:
	case 8:
	case 9:
	case 10:
	case 11:
		// \xml <tmc:axis_parameter type="8" name="Position reached flag" identifier="par::PositionReachedFlag">
		// \xml   <desc>This flag is always set when target position and actual position are equal.</desc>
		// \xml   <tmc:value_bool desc_false="Target position not reached" desc_true="Target position reached" default="false" access="R"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			if(type == 8)
				*value = StepDirCh[motor]->targetReached;
			else
				errors |= ERROR_TYPE;
		} else if(readWrite == WRITE) {
			errors |= setStandAloneSettings(type-6, *value);
		}
		break;
	case 12:
		if(readWrite == READ) {
			*value = (HAL.IOs->config->isHigh(Pins.AIN_REF_SW))? 1 : 0;
		} else if(readWrite == WRITE) {
			if(*value)
				HAL.IOs->config->setHigh(Pins.AIN_REF_SW);
			else
				HAL.IOs->config->setLow(Pins.AIN_REF_SW);
		}
	break;

	case 13:
		if(readWrite == READ) {
			*value = (HAL.IOs->config->isHigh(Pins.AIN_REF_PWM)) ? 1 : 0;
		} else if(readWrite == WRITE) {
			if(((uint32) *value) > 10000)
				errors |= ERROR_TYPE;
			else
				Timer.setDuty(*value);
		}
	break;

	// getStandaloneSettin APs have been copied from from types 6-11 but also left there for backwards-compatibility

	case 14:
		// \xml <tmc:axis_parameter type="14" name="Chopper off time" identifier="par::TOff">
		// \xml   <desc>The off time setting controls the minimum chopper frequency.</desc>
		// \xml   <tmc:value_choice default="0" unit="" access="RW">
		// \xml     <choice_element value="0" desc="140 Tclk (recommended)"></choice_element>
		// \xml     <choice_element value="1" desc="236 Tclk"></choice_element>
		// \xml     <choice_element value="2" desc="332 Tclk"></choice_element>
		// \xml   </tmc:value_choice>
		// \xml </tmc:axis_parameter>
	case 15:
		// \xml <tmc:axis_parameter type="15" name="Microstep Resolution, Interpolation and Chopper" identifier="par::MicrostepResoIntChopper">
		// \xml   <desc>Microstep resolutions per full step:</desc>
		// \xml   <tmc:value_choice default="0" unit="" access="RW">
		// \xml     <choice_element value="0" desc="fullstep, no intpol, spreadCycle"></choice_element>
		// \xml     <choice_element value="1" desc="halfstep, no intpol, spreadCycle"></choice_element>
		// \xml     <choice_element value="2" desc="halfstep, intpol to 256 uSteps, spreadCycle"></choice_element>
		// \xml     <choice_element value="3" desc="4 (quarterstep), no intpol, spreadCycle"></choice_element>
		// \xml     <choice_element value="4" desc="16 usteps, no intpol, spreadCycle"></choice_element>
		// \xml     <choice_element value="5" desc="4 (quarterstep), intpol to 256 uSteps, spreadCycle"></choice_element>
		// \xml     <choice_element value="6" desc="16 usteps, intpol to 256 uSteps, spreadCycle"></choice_element>
		// \xml     <choice_element value="7" desc="4 (quarterstep), intpol to 256 uSteps, stealthChop"></choice_element>
		// \xml     <choice_element value="8" desc="16 usteps, intpol to 256 uSteps, stealthChop"></choice_element>
		// \xml   </tmc:value_choice>
		// \xml </tmc:axis_parameter>
	case 16:
		// \xml <tmc:axis_parameter type="16" name="Maximum current" identifier="par::max. current">
		// \xml   <desc>Motor current used when motor is running.</desc>
		// \xml   <tmc:value_choice default="0" unit="" access="RW">
		// \xml     <choice_element value="0" desc="Internal reference voltage. Current scale set by external sense resistors only."></choice_element>
		// \xml     <choice_element value="1" desc="Internal sense resistors. Use AIN as reference current."></choice_element>
		// \xml     <choice_element value="2" desc="External reference voltage on pin AIN. Current scale set by sense resistors and scaled by AIN."></choice_element>
		// \xml   </tmc:value_choice>
		// \xml </tmc:axis_parameter>
	case 17:
		// \xml <tmc:axis_parameter type="17" name="Chopper hysteresis" identifier="par::ChopperHysteresis">
		// \xml   <desc>Tuning of zero crossing precision</desc>
		// \xml   <tmc:value_choice default="0" unit="" access="RW">
		// \xml     <choice_element value="0" desc="5 (recommended)"></choice_element>
		// \xml     <choice_element value="1" desc="9"></choice_element>
		// \xml     <choice_element value="2" desc="13"></choice_element>
		// \xml   </tmc:value_choice>
		// \xml </tmc:axis_parameter>
	case 18:
		// \xml <tmc:axis_parameter type="18" name="Chopper blank time" identifier="par::ChopperBlankTime">
		// \xml   <desc>Selects the comparator blank time. This time needs to safely cover the switching event and the duration of the ringing on the sense resistor.</desc>
		// \xml   <tmc:value_choice default="0" unit="" access="RW">
		// \xml     <choice_element value="0" desc="16 (for stealthChop)"></choice_element>
		// \xml     <choice_element value="1" desc="24 (recommended)"></choice_element>
		// \xml     <choice_element value="2" desc="36"></choice_element>
		// \xml   </tmc:value_choice>
		// \xml </tmc:axis_parameter>
	case 19:
		// \xml <tmc:axis_parameter type="19" name="Power down delay" identifier="par::power down delay">
		// \xml   <desc>Standstill period before the current will be ramped down to standby current. The standard value is 200 (which means 2000ms).</desc>
		// \xml   <tmc:value_choice default="0" unit="" access="RW">
		// \xml     <choice_element value="0" desc="N"></choice_element>
		// \xml     <choice_element value="1" desc="Driver disabled"></choice_element>
		// \xml     <choice_element value="2" desc="Y"></choice_element>
		// \xml   </tmc:value_choice>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			getStandAloneSettings(type - 14, value);
		} else if(readWrite == WRITE) {
			errors |= setStandAloneSettings(type - 14, *value);
		}
	break;

	case 140:
		// \xml <tmc:axis_parameter type="140" name="Microstep Resolution" identifier="par::MicrostepResolution">
		// \xml   <desc>Microstep resolutions per full step:</desc>
		// \xml   <tmc:value_choice default="256" unit="" access="R">
		// \xml     <choice_element value="1" desc="fullstep"></choice_element>
		// \xml     <choice_element value="2" desc="halfstep"></choice_element>
		// \xml     <choice_element value="4" desc="4 microsteps"></choice_element>
		// \xml     <choice_element value="16" desc="16 microsteps"></choice_element>
		// \xml   </tmc:value_choice>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			errors |= ERROR_TYPE;
		} else if(readWrite == WRITE) {
			// microstep resolution without interpolation

			errors |= getStandAloneSettings(1, value);

			IO_States CFG[7];
			TMCRhinoSA.getPins(CFG);

			if((CFG[2] == IOS_LOW)        && (CFG[1] == IOS_LOW))   *value = 1;
			else if((CFG[2] == IOS_LOW)   && (CFG[1] == IOS_HIGH))  *value = 2;
			else if((CFG[2] == IOS_LOW)   && (CFG[1] == IOS_OPEN))  *value = 2;
			else if((CFG[2] == IOS_HIGH)  && (CFG[1] == IOS_LOW))   *value = 4;
			else if((CFG[2] == IOS_HIGH)  && (CFG[1] == IOS_HIGH))  *value = 16;
			else if((CFG[2] == IOS_HIGH)  && (CFG[1] == IOS_OPEN))  *value = 4;
			else if((CFG[2] == IOS_OPEN)  && (CFG[1] == IOS_LOW))   *value = 16;
			else if((CFG[2] == IOS_OPEN)  && (CFG[1] == IOS_HIGH))  *value = 4;
			else if((CFG[2] == IOS_OPEN)  && (CFG[1] == IOS_OPEN))  *value = 16;
		}
		break;
	default:
		errors |= ERROR_TYPE;
		break;
	}
	// \xml </axis_parameters>
	return errors;
}
// \xml </tmc:product_module>

static uint32 SAP(uint8 type, uint8 motor, int32 value)
{
	return handleParameter(WRITE, motor, type, &value);
}

static uint32 GAP(uint8 type, uint8 motor, int32 *value)
{
	return handleParameter(READ, motor, type, value);
}

static void writeRegister(uint8 address, int32 value)
{
	UNUSED(address);

	TMCRhinoSA.setInt(value);
}

static void readRegister(uint8 address, int32 *value)
{
	UNUSED(address);

	*value = TMCRhinoSA.getInt();
}

static u32 setStandAloneSettings(uint8 i, int32 value)
{
	TMCRhinoTypeStandAloneConfigDef config;

	u32 errors = ERROR_NONE;

	if(i > 6)
	{
		errors |= ERROR_TYPE;
		return errors;
	}

	if((value < 0) || (value > 2))
	{
		if((i != 1) || (value > 0x0F))
		{
			errors |= ERROR_VALUE;
			return errors;
		}
	}

	TMCRhinoSA.getConfig(&config);

	switch(i)
	{
	case 0:
		config.chopperOffTime = value;
		break;
	case 1:
		switch(value)
		{
		case 0:
			config.microstepResolution1 = TMCRhinoSA.MicrostepResolutionSettings1.BY1_INTERPOL0;
			config.microstepResolution2 = TMCRhinoSA.MicrostepResolutionSettings2.BY1_INTERPOL0;
			break;
		case 1:
			config.microstepResolution1 = TMCRhinoSA.MicrostepResolutionSettings1.BY2_INTERPOL0;
			config.microstepResolution2 = TMCRhinoSA.MicrostepResolutionSettings2.BY2_INTERPOL0;
			break;
		case 2:
			config.microstepResolution1 = TMCRhinoSA.MicrostepResolutionSettings1.BY2_INTERPOL256;
			config.microstepResolution2 = TMCRhinoSA.MicrostepResolutionSettings2.BY2_INTERPOL256;
			break;
		case 3:
			config.microstepResolution1 = TMCRhinoSA.MicrostepResolutionSettings1.BY4_INTERPOL0;
			config.microstepResolution2 = TMCRhinoSA.MicrostepResolutionSettings2.BY4_INTERPOL0;
			break;
		case 4:
			config.microstepResolution1 = TMCRhinoSA.MicrostepResolutionSettings1.BY4_INTERPOL256;
			config.microstepResolution2 = TMCRhinoSA.MicrostepResolutionSettings2.BY4_INTERPOL256;
			break;
		case 5:
			config.microstepResolution1 = TMCRhinoSA.MicrostepResolutionSettings1.BY16_INTERPOL0;
			config.microstepResolution2 = TMCRhinoSA.MicrostepResolutionSettings2.BY16_INTERPOL0;
			break;
		case 6:
			config.microstepResolution1 = TMCRhinoSA.MicrostepResolutionSettings1.BY16_INTERPOL256;
			config.microstepResolution2 = TMCRhinoSA.MicrostepResolutionSettings2.BY16_INTERPOL256;
			break;
		case 7:
			config.microstepResolution1 = TMCRhinoSA.MicrostepResolutionSettings1.PWM_CHOP_BY4_INTERPOL256;
			config.microstepResolution2 = TMCRhinoSA.MicrostepResolutionSettings2.PWM_CHOP_BY4_INTERPOL256;
			break;
		case 8:
			config.microstepResolution1 = TMCRhinoSA.MicrostepResolutionSettings1.PWM_CHOP_BY16_INTERPOL256;
			config.microstepResolution2 = TMCRhinoSA.MicrostepResolutionSettings2.PWM_CHOP_BY16_INTERPOL256;
			break;
		}
		break;
	case 2:
		config.currentSetting = value;
		break;
	case 3:
		config.chopperHysteresis = value;
		break;
	case 4:
		config.chopperBlankTime = value;
		break;
	case 5:
		config.enableStandStillPowerDown = lastEnable = value;
		break;
	}

	TMCRhinoSA.setConfig(&config);

	return errors;
}

static u32 getStandAloneSettings(uint8 i, int32 *value)
{
	TMCRhinoTypeStandAloneConfigDef config;

	u32 errors = ERROR_NONE;

	if(i > 6)
	{
		errors |= ERROR_TYPE;
		return errors;
	}

	TMCRhinoSA.getConfig(&config);

	switch(i)
	{
	case 0:
		*value = config.chopperOffTime;
		break;
	case 1:
		if(     (config.microstepResolution1 == TMCRhinoSA.MicrostepResolutionSettings1.BY1_INTERPOL0) &&
			    (config.microstepResolution2 == TMCRhinoSA.MicrostepResolutionSettings2.BY1_INTERPOL0))
			*value = 0;
		else if((config.microstepResolution1 == TMCRhinoSA.MicrostepResolutionSettings1.BY2_INTERPOL0) &&
				(config.microstepResolution2 == TMCRhinoSA.MicrostepResolutionSettings2.BY2_INTERPOL0))
			*value = 1;
		else if((config.microstepResolution1 == TMCRhinoSA.MicrostepResolutionSettings1.BY2_INTERPOL256) &&
				(config.microstepResolution2 == TMCRhinoSA.MicrostepResolutionSettings2.BY2_INTERPOL256))
			*value = 2;
		else if((config.microstepResolution1 == TMCRhinoSA.MicrostepResolutionSettings1.BY4_INTERPOL0) &&
				(config.microstepResolution2 == TMCRhinoSA.MicrostepResolutionSettings2.BY4_INTERPOL0))
			*value = 3;
		else if((config.microstepResolution1 == TMCRhinoSA.MicrostepResolutionSettings1.BY4_INTERPOL256) &&
				(config.microstepResolution2 == TMCRhinoSA.MicrostepResolutionSettings2.BY4_INTERPOL256))
			*value = 4;
		else if((config.microstepResolution1 == TMCRhinoSA.MicrostepResolutionSettings1.BY16_INTERPOL0) &&
				(config.microstepResolution2 == TMCRhinoSA.MicrostepResolutionSettings2.BY16_INTERPOL0))
			*value = 5;
		else if((config.microstepResolution1 == TMCRhinoSA.MicrostepResolutionSettings1.BY16_INTERPOL256) &&
				(config.microstepResolution2 == TMCRhinoSA.MicrostepResolutionSettings2.BY16_INTERPOL256))
			*value = 6;
		else if((config.microstepResolution1 == TMCRhinoSA.MicrostepResolutionSettings1.PWM_CHOP_BY4_INTERPOL256) &&
				(config.microstepResolution2 == TMCRhinoSA.MicrostepResolutionSettings2.PWM_CHOP_BY4_INTERPOL256))
			*value = 7;
		else if((config.microstepResolution1 == TMCRhinoSA.MicrostepResolutionSettings1.PWM_CHOP_BY16_INTERPOL256) &&
				(config.microstepResolution2 == TMCRhinoSA.MicrostepResolutionSettings2.PWM_CHOP_BY16_INTERPOL256))
			*value = 8;
		else
			errors |= ERROR_VALUE;
		break;
	case 2:
		*value = config.currentSetting;
		break;
	case 3:
		*value = config.chopperHysteresis;
		break;
	case 4:
		*value = config.chopperBlankTime;
		break;
	case 5:
		*value = config.enableStandStillPowerDown;
		break;
	case 6:
		*value = config.chopperOffTime;
		break;
	}
	return errors;
}

void TMC2100_init(void)
{
	Pins.AIN_REF_PWM  = &HAL.IOs->pins->DIO11;
	Pins.AIN_REF_SW   = &HAL.IOs->pins->DIO10;
	Pins.CFG0         = &HAL.IOs->pins->SPI2_SDO;
	Pins.CFG1         = &HAL.IOs->pins->SPI2_SDI;
	Pins.CFG2         = &HAL.IOs->pins->SPI2_SCK;
	Pins.CFG3         = &HAL.IOs->pins->SPI2_CSN0;
	Pins.CFG4         = &HAL.IOs->pins->DIO13;
	Pins.CFG5         = &HAL.IOs->pins->DIO12;
	Pins.CFG6_ENN     = &HAL.IOs->pins->DIO0;
	Pins.DIR          = &HAL.IOs->pins->DIO7;
	Pins.ERROR        = &HAL.IOs->pins->DIO15;
	Pins.INDEX        = &HAL.IOs->pins->DIO16;
	Pins.STEP         = &HAL.IOs->pins->DIO6;

	HAL.IOs->config->toOutput(Pins.STEP);
	HAL.IOs->config->toOutput(Pins.DIR);
	HAL.IOs->config->toOutput(Pins.AIN_REF_PWM);
	HAL.IOs->config->toOutput(Pins.AIN_REF_SW);

	HAL.IOs->config->toInput(Pins.ERROR);
	HAL.IOs->config->toInput(Pins.INDEX);

	TMCRhinoSA.CFGPins[0] = Pins.CFG0;
	TMCRhinoSA.CFGPins[1] = Pins.CFG1;
	TMCRhinoSA.CFGPins[2] = Pins.CFG2;
	TMCRhinoSA.CFGPins[3] = Pins.CFG3;
	TMCRhinoSA.CFGPins[4] = Pins.CFG4;
	TMCRhinoSA.CFGPins[5] = Pins.CFG5;
	TMCRhinoSA.CFGPins[6] = Pins.CFG6_ENN;

	for(uint8 i = 0; i < 7; i++)
	{
		HAL.IOs->config->toOutput(TMCRhinoSA.CFGPins[i]);
		HAL.IOs->config->setLow(TMCRhinoSA.CFGPins[i]);
	}

	Evalboards.ch2.config->state = CONFIG_READY;

	StepDir_init(MOTORS);
	StepDirCh[0]           = StepDir.ch2;
	StepDirCh[0]->stepOut  = Pins.STEP;
	StepDirCh[0]->dirOut   = Pins.DIR;

	Evalboards.ch2.rotate          = rotate;
	Evalboards.ch2.right           = right;
	Evalboards.ch2.left            = left;
	Evalboards.ch2.stop            = stop;
	Evalboards.ch2.GAP             = GAP;
	Evalboards.ch2.SAP             = SAP;
	Evalboards.ch2.moveTo          = moveTo;
	Evalboards.ch2.moveBy          = moveBy;
	Evalboards.ch2.writeRegister   = writeRegister;
	Evalboards.ch2.readRegister    = readRegister;
	Evalboards.ch2.enableDriver    = enableDriver;
	Evalboards.ch2.numberOfMotors  = MOTORS;
	Evalboards.ch2.deInit          = deInit;

	enableDriver(DRIVER_USE_GLOBAL_ENABLE);
	reset();

#if defined(Startrampe)
	Pins.AIN_REF_PWM->configuration.GPIO_Mode = GPIO_Mode_AF;
	GPIO_PinAFConfig(Pins.AIN_REF_PWM->port, Pins.AIN_REF_PWM->bit, GPIO_AF_TIM1);
#elif defined(Landungsbruecke)
	HAL.IOs->config->toOutput(Pins.AIN_REF_PWM);
	Pins.AIN_REF_PWM->configuration.GPIO_Mode = GPIO_Mode_AF4;
#endif

	HAL.IOs->config->set(Pins.AIN_REF_PWM);
	Timer.init();
	Timer.setDuty(0);
}

static void deInit(void)
{
	enableDriver(DRIVER_DISABLE);

	HAL.IOs->config->reset(Pins.AIN_REF_PWM);
	HAL.IOs->config->reset(Pins.AIN_REF_SW);
	HAL.IOs->config->reset(Pins.CFG0);
	HAL.IOs->config->reset(Pins.CFG1);
	HAL.IOs->config->reset(Pins.CFG2);
	HAL.IOs->config->reset(Pins.CFG3);
	HAL.IOs->config->reset(Pins.CFG4);
	HAL.IOs->config->reset(Pins.CFG5);
	HAL.IOs->config->reset(Pins.CFG6_ENN);
	HAL.IOs->config->reset(Pins.DIR);
	HAL.IOs->config->reset(Pins.ERROR);
	HAL.IOs->config->reset(Pins.INDEX);
	HAL.IOs->config->reset(Pins.STEP);

	StepDir_deInit();
	Timer.deInit();
}

static uint8 reset()
{
	if(StepDirCh[0]->actualVelocity && !VitalSignsMonitor.brownOut)
		return 0;

	TMCRhinoSA.reset();

	return 1;
}

static void enableDriver(DriverState state)
{
	TMCRhinoTypeStandAloneConfigDef config;
	TMCRhinoSA.getConfig(&config);

	if(state == DRIVER_USE_GLOBAL_ENABLE)
		state = Evalboards.driverEnable;

	if(state == DRIVER_DISABLE)
	{
		lastEnable = config.enableStandStillPowerDown;
		config.enableStandStillPowerDown = TMCRhinoSA.EnableStandStillPowerDownSettings.DISABLED;
	}
	else
	{
		config.enableStandStillPowerDown = lastEnable;
	}
}
