#include "../tmc/stepDir.h"
#include "Board.h"
#include "tmc/ic/TMC2130/TMC2130.h"

#define MOTORS         1  // number of motors for this board
#define DEFAULT_MOTOR  0

#define TMC2130_EVAL_VM_MIN  50   // VM[V/10] min
#define TMC2130_EVAL_VM_MAX  480  // VM[V/10] max +5%

#define TMC2130_MAX_VELOCITY      STEPDIR_MAX_VELOCITY
#define TMC2130_MAX_ACCELERATION  16777215

typedef struct
{
	IOPinTypeDef *REFL_STEP;
	IOPinTypeDef *REFR_DIR;
	IOPinTypeDef *DRV_ENN_CFG6;
	IOPinTypeDef *ENCA_DCIN_CFG5;
	IOPinTypeDef *ENCB_DCEN_CFG4;
	IOPinTypeDef *ENCN_DCO;
	IOPinTypeDef *DIAG0;
	IOPinTypeDef *DIAG1;
	IOPinTypeDef *AIN_REF_SW;
	IOPinTypeDef *AIN_REF_PWM;
} PinsTypeDef;

static PinsTypeDef Pins;

StepDirectionChannelType stepDirDummmy;
static StepDirectionChannelType *StepDirCh[] = {&stepDirDummmy};

SPIChannelTypeDef *TMC2130_SPIChannel[MOTORS];
TMC2130TypeDef Rhino_TMC2130[MOTORS];
ConfigurationTypeDef *TMC2130_config[MOTORS];

// => SPI wrapper (extend the SPI functions to allow motor selection in your application)
void tmc2130_writeDatagram(uint8 motor, uint8 address, uint8 x1, uint8 x2, uint8 x3, uint8 x4)
{
	TMC2130_SPIChannel[motor]->readWrite(address | 0x80, FALSE);
	TMC2130_SPIChannel[motor]->readWrite(x1, FALSE);
	TMC2130_SPIChannel[motor]->readWrite(x2, FALSE);
	TMC2130_SPIChannel[motor]->readWrite(x3, FALSE);
	TMC2130_SPIChannel[motor]->readWrite(x4, TRUE);

	int value = x1;
	value <<= 8;
	value |= x2;
	value <<= 8;
	value |= x3;
	value <<= 8;
	value |= x4;

	TMC2130_config[motor]->shadowRegister[address & 0x7F] = value;
}

void tmc2130_writeInt(uint8 motor, uint8 address, int value)
{
	tmc2130_writeDatagram(motor, address, 0xFF & (value>>24), 0xFF & (value>>16), 0xFF & (value>>8), 0xFF & (value>>0));
}

int tmc2130_readInt(u8 motor, uint8 address)
{
	address &= 0x7F;

	// register not readable -> return shadow register copy
	if(!IS_READABLE(Rhino_TMC2130[motor].registerAccess[address]))
		return TMC2130_config[motor]->shadowRegister[address];

	TMC2130_SPIChannel[motor]->readWrite(address, FALSE);
	TMC2130_SPIChannel[motor]->readWrite(0, FALSE);
	TMC2130_SPIChannel[motor]->readWrite(0, FALSE);
	TMC2130_SPIChannel[motor]->readWrite(0, FALSE);
	TMC2130_SPIChannel[motor]->readWrite(0, TRUE);

	TMC2130_SPIChannel[motor]->readWrite(address, FALSE);
	int value = TMC2130_SPIChannel[motor]->readWrite(0, FALSE);
	value <<= 8;
	value |= TMC2130_SPIChannel[motor]->readWrite(0, FALSE);
	value <<= 8;
	value |= TMC2130_SPIChannel[motor]->readWrite(0, FALSE);
	value <<= 8;
	value |= TMC2130_SPIChannel[motor]->readWrite(0, TRUE);

	return value;
}
// <= SPI wrapper

static uint32 rotate(uint8 motor, int32 velocity)
{
	if(motor >= MOTORS)
		return ERROR_MOTOR;

	StepDirCh[motor]->rampMode        = SD_RAMP_MODE_VELOCITY;
	StepDirCh[motor]->targetVelocity  = velocity;

	return ERROR_NONE;
}

static uint32 right(uint8 motor, int32 velocity)
{
	return rotate(motor, velocity);
}

static uint32 left(uint8 motor, int32 velocity)
{
	return rotate(motor, -velocity);
}

static uint32 stop(uint8 motor)
{
	return rotate(motor, 0);
}

static uint32 moveTo(uint8 motor, int32 position)
{
	if(motor >= MOTORS)
		return ERROR_MOTOR;

	StepDirCh[motor]->targetPosition  = position;
	StepDirCh[motor]->rampMode        = SD_RAMP_MODE_POSITIONING;
	StepDirCh[motor]->useProfile      = 0;

	return ERROR_NONE;
}

static uint32 moveBy(uint8 motor, int32 *ticks)
{
	if(motor >= MOTORS)
		return ERROR_MOTOR;

	// determine actual position and add numbers of ticks to move
	*ticks = StepDirCh[motor]->actualPosition + *ticks;

	return moveTo(motor, *ticks);
}

static uint32 moveProfile(uint8 motor, int32 position)
{
	if(motor >= MOTORS)
		return ERROR_MOTOR;

	StepDirCh[motor]->rampMode        = SD_RAMP_MODE_POSITIONING;
	StepDirCh[motor]->useProfile      = 0;
	StepDirCh[motor]->profileFrom     = StepDir.ch1->actualPosition;
	StepDirCh[motor]->targetPosition  = StepDirCh[motor]->profileFrom;
	StepDirCh[motor]->profileTo       = position;
	StepDirCh[motor]->useProfile      = 1;

	return ERROR_NONE;
}

// \xml <?xml version="1.0" encoding="UTF-8" standalone="no"?>
// \xml <tmc:product_module author="BS" xmlns:tmc="http://www.trinamic.com" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://www.trinamic.com https://www.trinamic.com/fileadmin/xml/TMC_Schema.xsd">
// \xml <name>TMC2130-EVAL</name>
// \xml <name_fw_version>2130</name_fw_version>
// \xml <hardware_version>1.0</hardware_version>
// \xml <software_version>2.91</software_version>
static uint32 handleParameter(u8 readWrite, u8 motor, u8 type, int32 *value)
{
	u32 errors = ERROR_NONE;

	if(motor >= MOTORS)
		return ERROR_MOTOR;

	int32 tempValue;

	// \xml <axis_parameters axis="0">
	switch(type)
	{
	case 0:
		// \xml <tmc:axis_parameter type="0" name="Target position" identifier="par::TargetPosition">
		// \xml   <desc>The desired target position in position mode.</desc>
		// \xml   <tmc:value_signed min="#s32_MIN#" max="#s32_MAX#" default="0" unit="[usteps]" access="RW"></tmc:value_signed>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->targetPosition;
		} else if(readWrite == WRITE) {
			StepDirCh[motor]->targetPosition = *value;
		}
		break;
	case 1:
		// \xml <tmc:axis_parameter type="1" name="Actual position" identifier="par::ActualPosition">
		// \xml   <desc>The actual position of the motor. Stop the motor before overwriting it. Should normally only be overwritten for reference position setting.</desc>
		// \xml   <tmc:value_signed min="#s32_MIN#" max="#s32_MAX#" default="0" unit="[usteps]" access="RW"></tmc:value_signed>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->actualPosition;
		} else if(readWrite == WRITE) {
			StepDirCh[motor]->actualPosition = *value;
			//set also target position to prevent movement
			StepDirCh[motor]->targetPosition = *value;
		}
		break;
	case 2:
		// \xml <tmc:axis_parameter type="2" name="Target speed" identifier="par::TargetVelocity">
		// \xml   <desc>The desired speed in velocity mode. Not valid in position mode.</desc>
		// \xml   <tmc:value_signed min="-#TMC2130_MAX_VELOCITY#" max="#TMC2130_MAX_VELOCITY#" default="0" unit="[pps]" access="RW"></tmc:value_signed>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->targetVelocity;
		} else if(readWrite == WRITE) {
			StepDirCh[motor]->targetVelocity = *value;
		}
		break;
	case 3:
		// \xml <tmc:axis_parameter type="3" name="Actual speed" identifier="par::ActualVelocity">
		// \xml   <desc>The actual speed of the motor.</desc>
		// \xml   <tmc:value_signed min="-#TMC2130_MAX_VELOCITY#" max="#TMC2130_MAX_VELOCITY#" default="0" unit="[pps]" access="R"></tmc:value_signed>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->actualVelocity;
		} else if(readWrite == WRITE) {
			errors |= ERROR_TYPE;
		}
		break;
	case 4:
		// \xml <tmc:axis_parameter type="4" name="Maximum speed" identifier="par::MaxVelocity">
		// \xml   <desc>The maximum speed used for positioning ramps.</desc>
		// \xml   <tmc:value_unsigned min="0" max="#TMC2130_MAX_VELOCITY#" default="150000" unit="[pps]" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->maxPositioningSpeed;
		} else if(readWrite == WRITE) {
			StepDirCh[motor]->maxPositioningSpeed = abs(*value);
		}
		break;
	case 5:
		// \xml <tmc:axis_parameter type="5" name="Maximum acceleration" identifier="par::MaxAcceleration">
		// \xml   <desc>Maximum acceleration in positioning ramps. Acceleration and deceleration value in velocity mode.</desc>
		// \xml   <tmc:value_unsigned min="0" max="#TMC2130_MAX_ACCELERATION#" default="1000" unit="[pps/s]" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->actualAcceleration;
		} else if(readWrite == WRITE) {
			StepDirCh[motor]->actualAcceleration = *value;
		}
		break;
	case 6:
		// \xml <tmc:axis_parameter type="6" name="Maximum current" identifier="par::MaxCurrent">
		// \xml   <desc>Motor current used when motor is running. Where 0 = 3.125% up to 31 = 100% of maximum module current. Calculated by: maximum module current * (value + 1)/ 32.</desc>
		// \xml   <tmc:value_unsigned min="0" max="31" default="24" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2130_FIELD_READ(motor, TMC2130_IHOLD_IRUN, TMC2130_IRUN_MASK, TMC2130_IRUN_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2130_FIELD_UPDATE(motor, TMC2130_IHOLD_IRUN, TMC2130_IRUN_MASK, TMC2130_IRUN_SHIFT, *value);
		}
		break;
	case 7:
		// \xml <tmc:axis_parameter type="7" name="Standby current" identifier="par::StandbyCurrent">
		// \xml   <desc>The current used when the motor is not running. Where 0 = 3.125% up to 31 = 100% of maximum module current. Calculated by: maximum module current * (value + 1)/ 32. This value should be as low as possible so that the motor can cool down when it is not moving. Please see also parameter 214.</desc>
		// \xml   <tmc:value_unsigned min="0" max="31" default="3" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2130_FIELD_READ(motor, TMC2130_IHOLD_IRUN, TMC2130_IHOLD_MASK, TMC2130_IHOLD_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2130_FIELD_UPDATE(motor, TMC2130_IHOLD_IRUN, TMC2130_IHOLD_MASK, TMC2130_IHOLD_SHIFT, *value);
		}
		break;
	case 8:
		// \xml <tmc:axis_parameter type="8" name="Position reached flag" identifier="par::PositionReachedFlag">
		// \xml   <desc>This flag is always set when target position and actual position are equal.</desc>
		// \xml   <tmc:value_bool desc_false="Target position not reached" desc_true="Target position reached" default="false" access="R"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->targetReached;
		} else if(readWrite == WRITE) {
			errors |= ERROR_TYPE;
		}
		break;
	case 21:
		// todo CHECK 3: Funktionalitt prfen (JE)
		// todo XML 3: XML Beschreibung hinzufgen (JE)
		if(readWrite == READ) {
			errors |= ERROR_TYPE;
		} else if(readWrite == WRITE) {
			tmc2130_writeInt(motor, TMC2130_TPOWERDOWN, *value);
		}
		break;
	case 23:
		// \xml <tmc:axis_parameter type="23" name="Speed threshold for high speed mode" identifier="par::THIGH">
		// \xml   <desc>Speed threshold for de-activating coolStep and switching into a different chopper mode and fullstepping to maximize torque.</desc>
		// \xml	  <tmc:value_unsigned min="0" max="1048574" default="0" unit="[pps]" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			tempValue = tmc2130_readInt(motor, TMC2130_THIGH);
			*value = MIN(0xFFFFF, (1<<24) / ((tempValue)? tempValue:1));
		} else if(readWrite == WRITE) {
			*value = MIN(0xFFFFF, (1<<24) / ((*value)? *value:1));
			tmc2130_writeInt(motor, TMC2130_THIGH, *value);
		}
		break;
	case 24:
		// \xml <tmc:axis_parameter type="24" name="Minimum speed for switching to dcStep" identifier="par::VDCMIN">
		// \xml   <desc>DcStep is enabled above this velocity.</desc>
		// \xml   <tmc:value_unsigned min="0" max="1048574" default="0" unit="[pps]" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = tmc2130_readInt(motor, TMC2130_VDCMIN);
		} else if(readWrite == WRITE) {
			tmc2130_writeInt(motor, TMC2130_VDCMIN, *value);
		}
		break;
	case 26:
		// \xml <tmc:axis_parameter type="26" name="High speed fullstep mode" identifier="par::HighSpeedFullstepMode">
		// \xml   <desc>Switch to fullstep mode when measured speed is higher than axis parameter 23 when set to 1.</desc>
		// \xml   <tmc:value_bool desc_false="Inactive" desc_true="Active" default="false" access="RW"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2130_FIELD_READ(motor, TMC2130_CHOPCONF, TMC2130_VHIGHFS_MASK, TMC2130_VHIGHFS_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2130_FIELD_UPDATE(motor, TMC2130_CHOPCONF, TMC2130_VHIGHFS_MASK, TMC2130_VHIGHFS_SHIFT, *value);
		}
		break;
	case 27:
		// \xml <tmc:axis_parameter type="27" name="High speed chopper mode" identifier="par::HighSpeedChopperMode">
		// \xml   <desc>Switch to other chopper mode when measured speed is higher than axis parameter 22 when set to 1.</desc>
		// \xml   <tmc:value_bool desc_false="Inactive" desc_true="Active" default="false" access="RW"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2130_FIELD_READ(motor, TMC2130_CHOPCONF, TMC2130_VHIGHCHM_MASK, TMC2130_VHIGHCHM_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2130_FIELD_UPDATE(motor, TMC2130_CHOPCONF, TMC2130_VHIGHCHM_MASK, TMC2130_VHIGHCHM_SHIFT, *value);
		}
		break;
	case 28:
		// \xml <tmc:axis_parameter type="28" name="Internal RSense" identifier="par::internal_Rsense">
		// \xml   <desc>Activation of internal resistors to use current supplied into AIN as reference for internal sense resistor</desc>
		// \xml   <tmc:value_bool desc_false="Normal Operation" desc_true="Internal Sense Resistors" default="false" access="RW"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2130_FIELD_READ(motor, TMC2130_GCONF, TMC2130_INTERNAL_RSENSE_MASK, TMC2130_INTERNAL_RSENSE_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2130_FIELD_UPDATE(motor, TMC2130_GCONF, TMC2130_INTERNAL_RSENSE_MASK, TMC2130_INTERNAL_RSENSE_SHIFT, *value);
		}
		break;
	case 29:
		// todo XML CHECK 2: check description #1
		// \xml <tmc:axis_parameter type="29" name="Measured Speed" identifier="par::MeasuredSpeed">
		// \xml   <desc>Current speed value measured from the TMC2130 current position or from encoder.</desc>
		// \xml   <tmc:value_signed min="#s32_MIN#" max="#s32_MAX#" default="0" unit="" access="R"></tmc:value_signed>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDir.ch1->actualVelocity;
		} else if(readWrite == WRITE) {
			errors |= ERROR_TYPE;
		}
		break;
	case 140:
		// \xml <tmc:axis_parameter type="140" name="Microstep Resolution" identifier="par::MicrostepResolution">
		// \xml   <desc>Microstep resolutions per full step:</desc>
		// \xml   <tmc:value_choice default="256" unit="" access="RW">
		// \xml     <choice_element value="1" desc="fullstep"></choice_element>
		// \xml     <choice_element value="2" desc="halfstep"></choice_element>
		// \xml     <choice_element value="4" desc="4 microsteps"></choice_element>
		// \xml     <choice_element value="8" desc="8 microsteps"></choice_element>
		// \xml     <choice_element value="16" desc="16 microsteps"></choice_element>
		// \xml     <choice_element value="32" desc="32 microsteps"></choice_element>
		// \xml     <choice_element value="64" desc="64 microsteps"></choice_element>
		// \xml     <choice_element value="128" desc="128 microsteps"></choice_element>
		// \xml     <choice_element value="256" desc="256 microsteps"></choice_element>
		// \xml   </tmc:value_choice>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = 256 >> TMC2130_FIELD_READ(motor, TMC2130_CHOPCONF, TMC2130_MRES_MASK, TMC2130_MRES_SHIFT);
		} else if(readWrite == WRITE) {
			switch(*value)
			{
			case 1:    *value = 8;   break;
			case 2:    *value = 7;   break;
			case 4:    *value = 6;   break;
			case 8:    *value = 5;   break;
			case 16:   *value = 4;   break;
			case 32:   *value = 3;   break;
			case 64:   *value = 2;   break;
			case 128:  *value = 1;   break;
			case 256:  *value = 0;   break;
			default:   *value = -1;  break;
			}

			if(*value != -1)
			{
				TMC2130_FIELD_UPDATE(motor, TMC2130_CHOPCONF, TMC2130_MRES_MASK, TMC2130_MRES_SHIFT, *value);
			}
			else
			{
				errors |= ERROR_VALUE;
			}
		}
		break;
	case 162:
		// \xml <tmc:axis_parameter type="162" name="Chopper blank time" identifier="par::ChopperBlankTime">
		// \xml   <desc>Selects the comparator blank time. This time needs to safely cover the switching event and the duration of the ringing on the sense resistor. Normally leave at the default value.</desc>
		// \xml   <tmc:value_unsigned min="0" max="3" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2130_FIELD_READ(motor, TMC2130_CHOPCONF, TMC2130_TBL_MASK, TMC2130_TBL_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2130_FIELD_UPDATE(motor, TMC2130_CHOPCONF, TMC2130_TBL_MASK, TMC2130_TBL_SHIFT, *value);
		}
		break;
	case 163:
		// \xml <tmc:axis_parameter type="163" name="Constant TOff Mode" identifier="par::ConstantTOffMode">
		// \xml   <desc>Selection of the chopper mode.</desc>
		// \xml   <tmc:value_bool desc_false="spread cycle" desc_true="classic constant off time" default="false" access="RW"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2130_FIELD_READ(motor, TMC2130_CHOPCONF, TMC2130_CHM_MASK, TMC2130_CHM_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2130_FIELD_UPDATE(motor, TMC2130_CHOPCONF, TMC2130_CHM_MASK, TMC2130_CHM_SHIFT, *value);
		}
		break;
	case 164:
		// \xml <tmc:axis_parameter type="164" name="Disable fast decay comparator" identifier="par::DisableFastDecayComparator">
		// \xml   <desc>See parameter 163. For "classic const. off time", setting this parameter to "1" will disable current comparator usage for termination of fast decay cycle.</desc>
		// \xml   <tmc:value_bool desc_false="enable comparator" desc_true="disable comparator" default="false" access="RW"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2130_FIELD_READ(motor, TMC2130_CHOPCONF, TMC2130_DISFDCC_MASK, TMC2130_DISFDCC_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2130_FIELD_UPDATE(motor, TMC2130_CHOPCONF, TMC2130_DISFDCC_MASK, TMC2130_DISFDCC_SHIFT, *value);
		}
		break;
	case 165:
		// \xml <tmc:axis_parameter type="165" name="Chopper hysteresis end / fast decay time" identifier="par::ChopperHysteresisEnd">
		// \xml   <desc>See parameter 163. For "spread cycle" chopper mode this parameter will set / return the hysteresis end setting (hysteresis end value after a number of decrements). For "classic const. off time" chopper mode this parameter will set / return the fast decay time.</desc>
		// \xml   <tmc:value_unsigned min="0" max="15" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			if(tmc2130_readInt(motor, TMC2130_CHOPCONF) & (1<<14))
			{
				*value = TMC2130_FIELD_READ(motor, TMC2130_CHOPCONF, TMC2130_HEND_MASK, TMC2130_HEND_SHIFT);
			}
			else
			{
				tempValue = tmc2130_readInt(motor, TMC2130_CHOPCONF);
				*value = (tmc2130_readInt(motor, TMC2130_CHOPCONF) >> 4) & 0x07;
				if(tempValue & (1<<11))
					*value |= 1<<3;
			}
		} else if(readWrite == WRITE) {
			if(tmc2130_readInt(motor, TMC2130_CHOPCONF) & (1<<14))
			{
				TMC2130_FIELD_UPDATE(motor, TMC2130_CHOPCONF, TMC2130_HEND_MASK, TMC2130_HEND_SHIFT, *value);
			}
			else
			{
				tempValue = tmc2130_readInt(motor, TMC2130_CHOPCONF);

				TMC2130_FIELD_UPDATE(motor, TMC2130_CHOPCONF, TMC2130_TFD_3_MASK, TMC2130_TFD_3_SHIFT, (*value & (1<<3))? 1:0);
				TMC2130_FIELD_UPDATE(motor, TMC2130_CHOPCONF, TMC2130_TFD_ALL_MASK, TMC2130_TFD_ALL_SHIFT, *value);
			}
		}
		break;
	case 166:
		// \xml <tmc:axis_parameter type="166" name="Chopper hysteresis start / sine wave offset" identifier="par::ChopperHysteresisStart">
		// \xml   <desc>See parameter 163. For "spread cycle" chopper mode this parameter will set / return the Hysteresis start setting (please note that this value is an offset to the hysteresis end value). For "classic const. off time" chopper mode this parameter will set / return the sine wave offset.</desc>
		// \xml   <tmc:value_unsigned min="0" max="8" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			if(tmc2130_readInt(motor, TMC2130_CHOPCONF) & (1<<14))
			{
				*value = TMC2130_FIELD_READ(motor, TMC2130_CHOPCONF, TMC2130_HSTRT_MASK, TMC2130_HSTRT_SHIFT);
			}
			else
			{
				tempValue = tmc2130_readInt(motor, TMC2130_CHOPCONF);
				*value = (tmc2130_readInt(motor, TMC2130_CHOPCONF) >> 7) & 0x0F;
				if(tempValue & (1<<11))
					*value |= 1<<3;
			}
		} else if(readWrite == WRITE) {
			if(tmc2130_readInt(motor, TMC2130_CHOPCONF) & (1<<14))
			{
				TMC2130_FIELD_UPDATE(motor, TMC2130_CHOPCONF, TMC2130_HSTRT_MASK, TMC2130_HSTRT_SHIFT, *value);
			}
			else
			{
				TMC2130_FIELD_UPDATE(motor, TMC2130_CHOPCONF, TMC2130_OFFSET_MASK, TMC2130_OFFSET_SHIFT, *value);
			}
		}
		break;
	case 167:
		// \xml <tmc:axis_parameter type="167" name="Chopper off time" identifier="par::TOff">
		// \xml   <desc>The off time setting controls the minimum chopper frequency. An off time within the range of 5us to 20us will fit. Off time setting for constant t Off chopper: NCLK = 12 + 32  tOFF (Minimum is 64 clocks) Setting this parameter to zero completely disables all driver transistors and the motor can freewheel.</desc>
		// \xml   <tmc:value_unsigned min="0" max="15" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2130_FIELD_READ(motor, TMC2130_CHOPCONF, TMC2130_TOFF_MASK, TMC2130_TOFF_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2130_FIELD_UPDATE(motor, TMC2130_CHOPCONF, TMC2130_TOFF_MASK, TMC2130_TOFF_SHIFT, *value);
		}
		break;
	case 168:
		// \xml <tmc:axis_parameter type="168" name="smartEnergy current minimum (SEIMIN)" identifier="par::SEIMIN">
		// \xml   <desc>Sets the lower motor current limit for coolStep operation by scaling the maximum current (see axis parameter 6) value.</desc>
		// \xml   <tmc:value_bool desc_false="1/2 of CS" desc_true="1/4 of CS" default="false" access="RW"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2130_FIELD_READ(motor, TMC2130_COOLCONF, TMC2130_SEIMIN_MASK, TMC2130_SEIMIN_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2130_FIELD_UPDATE(motor, TMC2130_COOLCONF, TMC2130_SEIMIN_MASK, TMC2130_SEIMIN_SHIFT, *value);
		}
		break;
	case 169:
		// \xml <tmc:axis_parameter type="169" name="smartEnergy current down step" identifier="par::SECDS">
		// \xml   <desc>Sets the number of stallGuard2 readings above the upper threshold necessary for each current decrement of the motor current. Number of stallGuard2 measurements per decrement: Scaling: 0. . . 3: 32, 8, 2, 1. 0: slow decrement, 3: fast decrement</desc>
		// \xml   <tmc:value_unsigned min="0" max="3" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2130_FIELD_READ(motor, TMC2130_COOLCONF, TMC2130_SEDN_MASK, TMC2130_SEDN_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2130_FIELD_UPDATE(motor, TMC2130_COOLCONF, TMC2130_SEDN_MASK, TMC2130_SEDN_SHIFT, *value);
		}
		break;
	case 170:
		// \xml <tmc:axis_parameter type="170" name="smartEnergy hysteresis" identifier="par::smartEnergyHysteresis">
		// \xml   <desc>Sets the distance between the lower and the upper threshold for stallGuard2 reading. Above the upper threshold the motor current becomes decreased. Hysteresis: ([AP172] + 1) * 32. Upper stallGuard threshold: ([AP172] + [AP170] + 1) * 32</desc>
		// \xml   <tmc:value_unsigned min="0" max="15" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2130_FIELD_READ(motor, TMC2130_COOLCONF, TMC2130_SEMAX_MASK, TMC2130_SEMAX_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2130_FIELD_UPDATE(motor, TMC2130_COOLCONF, TMC2130_SEMAX_MASK, TMC2130_SEMAX_SHIFT, *value);
		}
		break;
	case 171:
		// \xml <tmc:axis_parameter type="171" name="smartEnergy current up step" identifier="par::SECUS">
		// \xml   <desc>Sets the current increment step. The current becomes incremented for each measured stallGuard2 value below the lower threshold see smartEnergy hysteresis start). Current increment step size: Scaling: 0. . . 3: 1, 2, 4, 8. 0: slow increment, 3: fast increment / fast reaction to rising load</desc>
		// \xml   <tmc:value_unsigned min="0" max="3" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2130_FIELD_READ(motor, TMC2130_COOLCONF, TMC2130_SEUP_MASK, TMC2130_SEUP_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2130_FIELD_UPDATE(motor, TMC2130_COOLCONF, TMC2130_SEUP_MASK, TMC2130_SEUP_SHIFT, *value);
		}
		break;
	case 172:
		// \xml <tmc:axis_parameter type="172" name="smartEnergy hysteresis start" identifier="par::smartEnergyHysteresisStart">
		// \xml   <desc>The lower threshold for the stallGuard2 value (see smart Energy current up step).</desc>
		// \xml   <tmc:value_unsigned min="0" max="15" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2130_FIELD_READ(motor, TMC2130_COOLCONF, TMC2130_SEMIN_MASK, TMC2130_SEMIN_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2130_FIELD_UPDATE(motor, TMC2130_COOLCONF, TMC2130_SEMIN_MASK, TMC2130_SEMIN_SHIFT, *value);
		}
		break;
	case 173:
		// \xml <tmc:axis_parameter type="173" name="stallGuard2 filter enable" identifier="par::SG2FilterEnable">
		// \xml   <desc>Enables the stallGuard2 filter for more precision of the movement. If set, reduces the measurement frequency to one measurement per four fullsteps. In most cases it is expedient to set the filtered mode before using coolStep. Use the standard mode for step loss detection.</desc>
		// \xml   <tmc:value_bool desc_false="standard mode" desc_true="filtered mode" default="false" access="RW"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2130_FIELD_READ(motor, TMC2130_COOLCONF, TMC2130_SFILT_MASK, TMC2130_SFILT_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2130_FIELD_UPDATE(motor, TMC2130_COOLCONF, TMC2130_SFILT_MASK, TMC2130_SFILT_SHIFT, *value);
		}
		break;
	case 174:
		// \xml <tmc:axis_parameter type="174" name="stallGuard2 threshold" identifier="par::SG2Threshold">
		// \xml   <desc>This signed value controls stallGuard2 threshold level for stall output and sets the optimum measurement range for readout. A lower value gives a higher sensitivity. Zero is the starting value. A higher value makes stallGuard2 less sensitive and requires more torque to indicate a stall.</desc>
		// \xml   <tmc:value_signed min="-64" max="63" default="0" unit="" access="RW"></tmc:value_signed>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2130_FIELD_READ(motor, TMC2130_COOLCONF, TMC2130_SGT_MASK, TMC2130_SGT_SHIFT);
			*value = CAST_Sn_TO_S32(*value, 7);
		} else if(readWrite == WRITE) {
			TMC2130_FIELD_UPDATE(motor, TMC2130_COOLCONF, TMC2130_SGT_MASK, TMC2130_SGT_SHIFT, *value);
		}
		break;
	case 179:
		// \xml <tmc:axis_parameter type="179" name="VSense" identifier="par::VSense">
		// \xml   <desc>Sense resistor voltage:</desc>
		// \xml   <tmc:value_bool desc_false="High sense resistor voltage" desc_true="Low sense resistor voltage" default="false" access="RW"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2130_FIELD_READ(motor, TMC2130_CHOPCONF, TMC2130_VSENSE_MASK, TMC2130_VSENSE_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2130_FIELD_UPDATE(motor, TMC2130_CHOPCONF, TMC2130_VSENSE_MASK, TMC2130_VSENSE_SHIFT, *value);
		}
		break;
	case 180:
		// \xml <tmc:axis_parameter type="180" name="smartEnergy actual current" identifier="par::smartEnergyActualCurrent">
		// \xml   <desc>This status value provides the actual motor current setting as controlled by coolStep. The value goes up to the CS value and down to the portion of CS as specified by SEIMIN. Actual motor current scaling factor: 0. . . 31: 1/32, 2/32, . . . 32/32</desc>
		// \xml   <tmc:value_unsigned min="0" max="31" default="0" unit="" access="R"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2130_FIELD_READ(motor, TMC2130_DRV_STATUS, TMC2130_CS_ACTUAL_MASK, TMC2130_CS_ACTUAL_SHIFT);
		} else if(readWrite == WRITE) {
			errors |= ERROR_TYPE;
		}
		break;
	case 181:
		// \xml <tmc:axis_parameter type="181" name="smartEnergy stall velocity" identifier="par::smartEnergyStallVelocity">
		// \xml   <desc>Velocity from which stop on stall feature is active</desc>
		// \xml   <tmc:value_unsigned min="0" max="#TMC2130_MAX_VELOCITY#" default="0" unit="[pps]" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->stopOnStallVelocityTreshold;
		} else if(readWrite == WRITE) {
			StepDirCh[motor]->stopOnStallVelocityTreshold = *value;
			StepDir.ch1->stallFlagIn = 0;
		}
		break;
	case 182:
		// \xml <tmc:axis_parameter type="182" name="smartEnergy threshold speed" identifier="par::smartEnergyThresholdSpeed">
		// \xml   <desc>Above this speed coolStep becomes enabled.</desc>
		// \xml   <tmc:value_unsigned min="0" max="1048574" default="0" unit="[pps]" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			tempValue = tmc2130_readInt(motor, TMC2130_TCOOLTHRS);
			*value = MIN(0xFFFFF, (1<<24) / ((tempValue)? tempValue:1));
		} else if(readWrite == WRITE) {
			*value = MIN(0xFFFFF, (1<<24) / ((*value)? *value:1));
			tmc2130_writeInt(motor, TMC2130_TCOOLTHRS, *value);
		}
		break;
	case 184:
		// \xml <tmc:axis_parameter type="184" name="Random TOff mode" identifier="par::RandomTOffMode">
		// \xml   <desc>Enable / disable random TOff mode.</desc>
		// \xml   <tmc:value_bool desc_false="Chopper off time is fixed" desc_true="Chopper off time is random" default="false" access="RW"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2130_FIELD_READ(motor, TMC2130_CHOPCONF, TMC2130_RNDTF_MASK, TMC2130_RNDTF_SHIFT);
		} else if(readWrite == WRITE) {
			FIELD_WRITE(tmc2130_writeInt, motor, TMC2130_CHOPCONF, TMC2130_RNDTF_MASK, TMC2130_RNDTF_SHIFT, *value);
		}
		break;
	case 185:
		// \xml <tmc:axis_parameter type="185" name="Chopper synchronization" identifier="par::ChopperSynchronization">
		// \xml   <desc>This parameter allows synchronization of the chopper for both phases of a two phase motor in order to avoid the occurrence of a beat, especially at low velocities. 0: chopper sync function chopSync off 1. . . 15: chopper synchronization</desc>
		// \xml   <tmc:value_unsigned min="0" max="15" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2130_FIELD_READ(motor, TMC2130_CHOPCONF, TMC2130_SYNC_MASK, TMC2130_SYNC_SHIFT);
		} else if(readWrite == WRITE) {
			FIELD_WRITE(tmc2130_writeInt, motor, TMC2130_CHOPCONF, TMC2130_SYNC_MASK, TMC2130_SYNC_SHIFT, *value);
		}
		break;
	case 186:
		// \xml <tmc:axis_parameter type="186" name="PWM threshold speed" identifier="par::PWMThresholdSpeed">
		// \xml   <desc>The stealthChop feature will be switched on when the actual velocity falls below this value.</desc>
		// \xml   <tmc:value_unsigned min="0" max="1048574" default="0" unit="[pps]" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			tempValue = tmc2130_readInt(motor, TMC2130_TPWMTHRS);
			*value = MIN(0xFFFFF, (1<<24) / ((tempValue)? tempValue:1));
		} else if(readWrite == WRITE) {
			*value = MIN(0xFFFFF, (1<<24) / ((*value)? *value:1));
			tmc2130_writeInt(motor, TMC2130_TPWMTHRS, *value);
		}
		break;
	case 187:
		// \xml <tmc:axis_parameter type="187" name="PWM gradient" identifier="par::PWMGrad">
		// \xml   <desc>Velocity dependent gradient for PWM amplitude (stealthChop). Setting this value to 0 turns off stealthChop.</desc>
		// \xml   <tmc:value_unsigned min="0" max="15" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2130_FIELD_READ(motor, TMC2130_PWMCONF, TMC2130_PWM_GRAD_MASK, TMC2130_PWM_GRAD_SHIFT);
		} else if(readWrite == WRITE) {
			// Set gradient
			TMC2130_FIELD_UPDATE(motor, TMC2130_PWMCONF, TMC2130_PWM_GRAD_MASK, TMC2130_PWM_GRAD_SHIFT, *value);

			// Enable/disable stealthChop accordingly
			TMC2130_FIELD_UPDATE(motor, TMC2130_GCONF, TMC2130_EN_PWM_MODE_MASK, TMC2130_EN_PWM_MODE_SHIFT, (*value>0));
		}
		break;
	case 188:
		// \xml <tmc:axis_parameter type="188" name="PWM amplitude" identifier="par::PWMAmplitude">
		// \xml   <desc>Maximum PWM amplitude when switching to stealthChop mode. Do not set too low. Values above 64 recommended.</desc>
		// \xml   <tmc:value_unsigned min="0" max="#u8_MAX#" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2130_FIELD_READ(motor, TMC2130_PWMCONF, TMC2130_PWM_AMPL_MASK, TMC2130_PWM_AMPL_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2130_FIELD_UPDATE(motor, TMC2130_PWMCONF, TMC2130_PWM_AMPL_MASK, TMC2130_PWM_AMPL_SHIFT, *value);
		}
		break;
	case 191:
		// \xml <tmc:axis_parameter type="191" name="PWM frequency" identifier="par::PWMFrequency">
		// \xml   <desc>PWM frequency selection for StealthChop.</desc>
		// \xml   <tmc:value_choice default="0" unit="" access="RW">
		// \xml     <choice_element value="0" desc="fpwm = 1/1024 fclk"></choice_element>
		// \xml     <choice_element value="1" desc="fpwm = 1/683 fclk"></choice_element>
		// \xml     <choice_element value="2" desc="fpwm = 1/512 fclk"></choice_element>
		// \xml     <choice_element value="3" desc="fpwm = 1/410 fclk"></choice_element>
		// \xml   </tmc:value_choice>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2130_FIELD_READ(motor, TMC2130_PWMCONF, TMC2130_PWM_FREQ_MASK, TMC2130_PWM_FREQ_SHIFT);
		} else if(readWrite == WRITE) {
			if(*value >= 0 && *value < 4)
			{
				TMC2130_FIELD_UPDATE(motor, TMC2130_PWMCONF, TMC2130_PWM_FREQ_MASK, TMC2130_PWM_FREQ_SHIFT, *value);
			}
			else
			{
				errors |= ERROR_VALUE;
			}
		}
		break;
	case 192:
		// \xml <tmc:axis_parameter type="192" name="PWM autoscale" identifier="par::PWMAutoscale">
		// \xml   <desc>PWM automatic amplitude scaling for StealthChop.</desc>
		// \xml   <tmc:value_bool desc_false="User defined PWM amplitude. The current settings have no influence" desc_true="enable automatic current control" default="false" access="RW"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2130_FIELD_READ(motor, TMC2130_PWMCONF, TMC2130_PWM_AUTOSCALE_MASK, TMC2130_PWM_AUTOSCALE_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2130_FIELD_UPDATE(motor, TMC2130_PWMCONF, TMC2130_PWM_AUTOSCALE_MASK, TMC2130_PWM_AUTOSCALE_SHIFT, (*value)? 1:0);
		}
		break;
	case 204:
		// \xml <tmc:axis_parameter type="204" name="Freewheeling mode" identifier="par::FreewheelingMode">
		// \xml   <desc>Stand still option when the standby current (parameter 7) is set to zero.</desc>
		// \xml   <tmc:value_choice default="0" unit="" access="RWE">
		// \xml     <choice_element value="0" desc="normal operation"></choice_element>
		// \xml     <choice_element value="1" desc="freewheeling"></choice_element>
		// \xml     <choice_element value="2" desc="coil shorted using low side drivers"></choice_element>
		// \xml     <choice_element value="3" desc="coil shorted using high side drivers"></choice_element>
		// \xml   </tmc:value_choice>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2130_FIELD_READ(motor, TMC2130_PWMCONF, TMC2130_FREEWHEEL_MASK, TMC2130_FREEWHEEL_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2130_FIELD_UPDATE(motor, TMC2130_PWMCONF, TMC2130_FREEWHEEL_MASK, TMC2130_FREEWHEEL_SHIFT, *value);
		}
		break;
	case 206:
		// \xml <tmc:axis_parameter type="206" name="Load value" identifier="par::LoadValue">
		// \xml   <desc>Actual current control scaling for monitoring smart energy current scaling or automatic current scaling.</desc>
		// \xml   <tmc:value_unsigned min="0" max="#u10_MAX#" default="0" unit="" access="R"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2130_FIELD_READ(motor, TMC2130_DRV_STATUS, TMC2130_SG_RESULT_MASK, TMC2130_SG_RESULT_SHIFT);
		} else if(readWrite == WRITE) {
			errors |= ERROR_TYPE;
		}
		break;
	default:
		errors |= ERROR_TYPE;
		break;
	}
	// \xml </axis_parameters>
	return errors;
}
// \xml </tmc:product_module>

static uint32 SAP(uint8 type, uint8 motor, int32 value)
{
	return handleParameter(WRITE, motor, type, &value);
}

static uint32 GAP(uint8 type, uint8 motor, int32 *value)
{
	return handleParameter(READ, motor, type, value);
}

static void writeRegister(uint8 address, int32 value)
{
	tmc2130_writeInt(DEFAULT_MOTOR, address, value);
}

static void readRegister(uint8 address, int32 *value)
{
	*value = tmc2130_readInt(DEFAULT_MOTOR, address);
}

static void periodicJob(uint32 tick)
{
	for(int motor = 0; motor < MOTORS; motor++)
	{
		TMC2130_periodicJob(motor, tick, &Rhino_TMC2130[motor], TMC2130_config[motor]);

		if(!StepDirCh[motor]->stallFlagIn && StepDir.ch1->stopOnStallActive)
		{
			tmc2130_readInt(motor, TMC2130_DRV_STATUS);
			if(tmc2130_readInt(motor, TMC2130_DRV_STATUS) & (1<<24))
				StepDir.ch1->stallFlagIn = 1;
		}
	}
}

static uint32 userFunction(uint8 type, uint8 motor, int32 *value)
{
	uint32 errors = 0;

	uint32 uvalue;

	switch(type)
	{
	case 0:	// set analogue current duty
		/*
		 * Current will be defined by analogue value voltage or current signal. In any case this function
		 * will generate a analogue voltage by PWM for up to 50% duty and a switch for the other 50%.
		 * The reference voltage will be AIN_REF = VCC_IO * value/20000 with value = {0..20000}
		 */

		uvalue = (uint32) *value;

		if(uvalue <= 20000)
		{
			HAL.IOs->config->setToState(Pins.AIN_REF_SW, (uvalue > 10000) ? IOS_HIGH : IOS_LOW);
			Timer.setDuty(uvalue%10001);
		}
		else
		{
			errors |= ERROR_VALUE;
		}
		break;
	case 1:	// Use internal clock
		/*
		 * Internel clock will be enabled by calling this function with a value != 0 and unpower and repower the motor supply while keeping usb connected.
		 */
		if(*value)
			HAL.IOs->config->setToState(&HAL.IOs->pins->CLK16, IOS_LOW);
		else
			HAL.IOs->config->reset(&HAL.IOs->pins->CLK16);
		break;
	case 2:	// writing a register at address = motor with value = value and reading back the value
		// DO NOT USE!
		// todo BUG 3: using the motor for both address and motor value can lead to trying to get struct pointer from an array outside its bounds
		//             and then loading a function pointer from that. That is pretty a much guaranteed crash and/or hard fault!!! (LH) #1

		//tmc2130_writeInt(motor, motor, *value);
		//*value = tmc2130_readInt(motor, motor);
		break;
	case 3:	// set DC_EN
		HAL.IOs->config->setToState(Pins.ENCB_DCEN_CFG4, (*value) ? IOS_HIGH : IOS_LOW);
		break;
	case 4: // move a profile
		switch(*value)
		{
		case 0: // deactivate using a profile
			StepDirCh[motor]->useProfile = 0;
			break;
		case 1: // move from this position
			StepDirCh[motor]->profileFrom = StepDir.ch1->actualPosition;
			break;
		case 2: // to this position
			StepDirCh[motor]->profileTo = StepDir.ch1->actualPosition;
			break;
		case 3: // activate using a profile
			StepDirCh[motor]->useProfile      = 1;
			StepDirCh[motor]->targetPosition  = StepDirCh[motor]->profileFrom;
			StepDirCh[motor]->rampMode        = SD_RAMP_MODE_POSITIONING;
			break;
		default: // activate using a profile from current position to current position + value
			StepDirCh[motor]->useProfile      = 0;
			StepDirCh[motor]->profileFrom     = StepDir.ch1->actualPosition;
			StepDirCh[motor]->profileTo       = *value;
			StepDirCh[motor]->targetPosition  = StepDirCh[motor]->profileFrom;
			StepDirCh[motor]->useProfile      = 1;
			break;
		}
		break;
	default:
		errors |= ERROR_TYPE;
		break;
	}
	return errors;
}

static uint32 getMeasuredSpeed(uint8 motor, int32 *value)
{
	if(motor >= MOTORS)
		return ERROR_MOTOR;

	switch(motor)
	{
	case 0:
		*value = StepDir.ch1->actualVelocity;
		break;
	default:
		return ERROR_MOTOR;
		break;
	}
	return ERROR_NONE;
}

static void deInit(void)
{
	HAL.IOs->config->setHigh(Pins.DRV_ENN_CFG6);	// DISABLE DRIVER

	HAL.IOs->config->reset(Pins.REFL_STEP);
	HAL.IOs->config->reset(Pins.REFR_DIR);
	HAL.IOs->config->reset(Pins.AIN_REF_SW);
	HAL.IOs->config->reset(Pins.AIN_REF_PWM);
	HAL.IOs->config->reset(Pins.ENCA_DCIN_CFG5);
	HAL.IOs->config->reset(Pins.ENCB_DCEN_CFG4);
	HAL.IOs->config->reset(Pins.ENCN_DCO);
	HAL.IOs->config->reset(Pins.DIAG0);
	HAL.IOs->config->reset(Pins.DIAG1);
	HAL.IOs->config->reset(Pins.DRV_ENN_CFG6);

	StepDir_deInit();
	Timer.deInit();
}

static uint8 reset()
{
	if(StepDirCh[0]->actualVelocity && !VitalSignsMonitor.brownOut)
		return 0;

	for(uint8 motor = 0; motor < MOTORS; motor++)
		TMC2130_reset(TMC2130_config[motor]);

	StepDir_init(MOTORS);
	StepDirCh[0]           = StepDir.ch1;
	StepDirCh[0]->stepOut  = Pins.REFL_STEP;
	StepDirCh[0]->dirOut   = Pins.REFR_DIR;

	return 1;
}

static uint8 restore()
{
	return TMC2130_restore(TMC2130_config[0]);
}

static void enableDriver(DriverState state)
{
	if(state == DRIVER_USE_GLOBAL_ENABLE)
		state = Evalboards.driverEnable;

	if(state == DRIVER_DISABLE)
		HAL.IOs->config->setHigh(Pins.DRV_ENN_CFG6);
	else if((state == DRIVER_ENABLE) && (Evalboards.driverEnable == DRIVER_ENABLE))
		HAL.IOs->config->setLow(Pins.DRV_ENN_CFG6);
}

void TMC2130_init(void)
{
	for(int i = 0; i < MOTORS; i++)
		TMC2130_initConfig(&Rhino_TMC2130[i]);

	Pins.DRV_ENN_CFG6    = &HAL.IOs->pins->DIO0;
	Pins.REFL_STEP       = &HAL.IOs->pins->DIO6;
	Pins.REFR_DIR        = &HAL.IOs->pins->DIO7;
	Pins.AIN_REF_SW      = &HAL.IOs->pins->DIO10;
	Pins.AIN_REF_PWM     = &HAL.IOs->pins->DIO11;
	Pins.ENCA_DCIN_CFG5  = &HAL.IOs->pins->DIO12;
	Pins.ENCB_DCEN_CFG4  = &HAL.IOs->pins->DIO13;
	Pins.ENCN_DCO        = &HAL.IOs->pins->DIO14;
	Pins.DIAG0           = &HAL.IOs->pins->DIO15;
	Pins.DIAG1           = &HAL.IOs->pins->DIO16;

	HAL.IOs->config->toInput(Pins.DIAG0);
	HAL.IOs->config->toInput(Pins.DIAG1);
	HAL.IOs->config->toInput(Pins.ENCN_DCO);

	HAL.IOs->config->toOutput(Pins.REFL_STEP);
	HAL.IOs->config->toOutput(Pins.REFR_DIR);
	HAL.IOs->config->toOutput(Pins.DRV_ENN_CFG6);
	HAL.IOs->config->toOutput(Pins.ENCB_DCEN_CFG4);
	HAL.IOs->config->toOutput(Pins.ENCA_DCIN_CFG5);
	HAL.IOs->config->toOutput(Pins.AIN_REF_PWM);
	HAL.IOs->config->toOutput(Pins.AIN_REF_SW);

	HAL.IOs->config->setLow(Pins.AIN_REF_PWM);
	HAL.IOs->config->setLow(Pins.AIN_REF_SW);
	HAL.IOs->config->setLow(Pins.ENCN_DCO);
	HAL.IOs->config->setLow(Pins.ENCA_DCIN_CFG5);

	for(int i = 0; i < MOTORS; i++)
	{
		TMC2130_SPIChannel[i] = &HAL.SPI->ch2;
		TMC2130_SPIChannel[i]->CSN = &HAL.IOs->pins->SPI2_CSN0;

		TMC2130_config[i] = Evalboards.ch2.config;
	}

	StepDir_init(MOTORS);
	StepDirCh[0]           = StepDir.ch1;
	StepDirCh[0]->stepOut  = Pins.REFL_STEP;
	StepDirCh[0]->dirOut   = Pins.REFR_DIR;

	Evalboards.ch2.config->reset        = reset;
	Evalboards.ch2.config->restore      = restore;
	Evalboards.ch2.config->state        = CONFIG_RESET;
	Evalboards.ch2.config->configIndex  = 0;

	Evalboards.ch2.rotate               = rotate;
	Evalboards.ch2.right                = right;
	Evalboards.ch2.left                 = left;
	Evalboards.ch2.stop                 = stop;
	Evalboards.ch2.GAP                  = GAP;
	Evalboards.ch2.SAP                  = SAP;
	Evalboards.ch2.moveTo               = moveTo;
	Evalboards.ch2.moveBy               = moveBy;
	Evalboards.ch2.moveProfile          = moveProfile;
	Evalboards.ch2.writeRegister        = writeRegister;
	Evalboards.ch2.readRegister         = readRegister;
	Evalboards.ch2.periodicJob          = periodicJob;
	Evalboards.ch2.userFunction         = userFunction;
	Evalboards.ch2.getMeasuredSpeed     = getMeasuredSpeed;
	Evalboards.ch2.enableDriver         = enableDriver;
	Evalboards.ch2.numberOfMotors       = MOTORS;
	Evalboards.ch2.VMMin                = TMC2130_EVAL_VM_MIN;
	Evalboards.ch2.VMMax                = TMC2130_EVAL_VM_MAX;
	Evalboards.ch2.deInit               = deInit;

	enableDriver(DRIVER_USE_GLOBAL_ENABLE);

#if defined(Startrampe)
	Pins.AIN_REF_PWM->configuration.GPIO_Mode = GPIO_Mode_AF;
	GPIO_PinAFConfig(Pins.AIN_REF_PWM->port, Pins.AIN_REF_PWM->bit, GPIO_AF_TIM1);
#elif defined(Landungsbruecke)
	HAL.IOs->config->toOutput(Pins.AIN_REF_PWM);
	Pins.AIN_REF_PWM->configuration.GPIO_Mode = GPIO_Mode_AF4;
#endif
	HAL.IOs->config->set(Pins.AIN_REF_PWM);

	Timer.init();
	Timer.setDuty(0);
}
