#include "Board.h"
#include "tmc/ic/TMC2208/TMC2208.h"
#include "../tmc/stepDir.h"

#define TMC2208_MAX_VELOCITY      STEPDIR_MAX_VELOCITY
#define TMC2208_MAX_ACCELERATION  16777215

#define ERRORS_VM        (1<<0)
#define ERRORS_VM_UNDER  (1<<1)
#define ERRORS_VM_OVER   (1<<2)

#define VM_MIN  50   // VM[V/10] min
#define VM_MAX  390  // VM[V/10] max

#define MOTORS 1

#define TIMEOUT_VALUE 10 // 10 ms

static uint32 right(uint8 motor, int32 velocity);
static uint32 left(uint8 motor, int32 velocity);
static uint32 rotate(uint8 motor, int32 velocity);
static uint32 stop(uint8 motor);
static uint32 moveTo(uint8 motor, int32 position);
static uint32 moveBy(uint8 motor, int32 *ticks);
static uint32 GAP(uint8 type, uint8 motor, int32 *value);
static uint32 SAP(uint8 type, uint8 motor, int32 value);

static void readRegister (uint8 address, int32 *value);
static void writeRegister (uint8 address, int32 value);

static void checkErrors (uint32 tick);
static void deInit(void);
static uint32 userFunction(uint8 type, uint8 motor, int32 *value);

static void periodicJob(uint32 tick);
static uint8 reset(void);
static void enableDriver(DriverState state);

StepDirectionChannelType stepDirDummmy;
static StepDirectionChannelType *StepDirCh[] = {&stepDirDummmy};

RXTXTypeDef *TMC2208_UARTChannel;
TMC2208TypeDef TMC2208;
ConfigurationTypeDef *TMC2208_config;

// Helper macro - index is always 1 here (channel 1 <-> index 0, channel 2 <-> index 1)
#define TMC2208_CRC(data, length) CRC8(data, length, 1)

typedef struct
{
	IOPinTypeDef  *DRV_ENN;
	IOPinTypeDef  *STEP;
	IOPinTypeDef  *DIR;
	IOPinTypeDef  *MS1;
	IOPinTypeDef  *MS2;
	IOPinTypeDef  *DIAG;
	IOPinTypeDef  *INDEX;
} PinsTypeDef;

static PinsTypeDef Pins;

static uint8 restore(void);

void tmc2208_writeRegister(uint8 address, int32 value)
{
	writeRegister(address, value);
}

void tmc2208_readRegister(uint8 address, int32 *value)
{
	readRegister(address, value);
}

static uint32 rotate(uint8 motor, int32 velocity)
{
	if(motor >= MOTORS)
		return ERROR_MOTOR;

	StepDirCh[motor]->rampMode        = SD_RAMP_MODE_VELOCITY;
	StepDirCh[motor]->targetVelocity  = velocity;

	return ERROR_NONE;
}

static uint32 right(uint8 motor, int32 velocity)
{
	return rotate(motor, velocity);
}

static uint32 left(uint8 motor, int32 velocity)
{
	return rotate(motor, -velocity);
}

static uint32 stop(uint8 motor)
{
	return rotate(motor, 0);
}

static uint32 moveTo(uint8 motor, int32 position)
{
	if(motor >= MOTORS)
		return ERROR_MOTOR;

	StepDirCh[motor]->targetPosition  = position;
	StepDirCh[motor]->rampMode        = SD_RAMP_MODE_POSITIONING;

	return ERROR_NONE;
}

static uint32 moveBy(uint8 motor, int32 *ticks)
{
	if(motor >= MOTORS)
		return ERROR_MOTOR;

	// determine actual position and add numbers of ticks to move
	*ticks = StepDirCh[motor]->actualPosition + *ticks;

	return moveTo(motor, *ticks);
}

// \xml <?xml version="1.0" encoding="UTF-8" standalone="no"?>
// \xml <tmc:product_module author="OK" xmlns:tmc="http://www.trinamic.com" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://www.trinamic.com https://www.trinamic.com/fileadmin/xml/TMC_Schema.xsd">
// \xml <name>TMC2208-EVAL</name>
// \xml <name_fw_version>2208</name_fw_version>
// \xml <hardware_version>1.0</hardware_version>
// \xml <software_version>3.00</software_version>
static uint32 handleParameter(u8 readWrite, u8 motor, u8 type, int32 *value)
{
	u32 errors = ERROR_NONE;


	if(motor >= MOTORS)
		return ERROR_MOTOR;

	// \xml <axis_parameters axis="0">
	switch(type)
	{
	case 0:
		// \xml <tmc:axis_parameter type="0" name="Target position" identifier="par::TargetPosition">
		// \xml   <desc>The desired target position in position mode.</desc>
		// \xml   <tmc:value_signed min="#s32_MIN#" max="#s32_MAX#" default="0" unit="[usteps]" access="RW"></tmc:value_signed>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->targetPosition;
		} else if(readWrite == WRITE) {
			StepDirCh[motor]->targetPosition = *value;
		}
		break;
	case 1:
		// \xml <tmc:axis_parameter type="1" name="Actual position" identifier="par::ActualPosition">
		// \xml   <desc>The actual position of the motor. Stop the motor before overwriting it. Should normally only be overwritten for reference position setting.</desc>
		// \xml   <tmc:value_signed min="#s32_MIN#" max="#s32_MAX#" default="0" unit="[usteps]" access="RW"></tmc:value_signed>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->actualPosition;
		} else if(readWrite == WRITE) {
			StepDirCh[motor]->actualPosition = *value;
			StepDirCh[motor]->targetPosition = *value;
		}
		break;
	case 2:
		// \xml <tmc:axis_parameter type="2" name="Target speed" identifier="par::TargetVelocity">
		// \xml   <desc>The desired speed in velocity mode. Not valid in position mode.</desc>
		// \xml   <tmc:value_signed min="-#TMC2208_MAX_VELOCITY#" max="#TMC2208_MAX_VELOCITY#" default="0" unit="[pps]" access="RW"></tmc:value_signed>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->targetVelocity;
		} else if(readWrite == WRITE) {
			StepDirCh[motor]->targetVelocity = *value;
		}
		break;
	case 3:
		// \xml <tmc:axis_parameter type="3" name="Actual speed" identifier="par::ActualVelocity">
		// \xml   <desc>The actual speed of the motor.</desc>
		// \xml   <tmc:value_signed min="-#TMC2208_MAX_VELOCITY#" max="#TMC2208_MAX_VELOCITY#" default="0" unit="[pps]" access="R"></tmc:value_signed>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->actualVelocity;
		} else if(readWrite == WRITE) {
			errors |= ERROR_TYPE;
		}
		break;
	case 4:
		// \xml <tmc:axis_parameter type="4" name="Maximum speed" identifier="par::MaxVelocity">
		// \xml   <desc>The maximum speed used for positioning ramps.</desc>
		// \xml   <tmc:value_unsigned min="0" max="#TMC2208_MAX_VELOCITY#" default="150000" unit="[pps]" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->maxPositioningSpeed;
		} else if(readWrite == WRITE) {
			StepDirCh[motor]->maxPositioningSpeed = abs(*value);
		}
		break;
	case 5:
		// \xml <tmc:axis_parameter type="5" name="Maximum acceleration" identifier="par::MaxAcceleration">
		// \xml   <desc>Maximum acceleration in positioning ramps. Acceleration and deceleration value in velocity mode.</desc>
		// \xml   <tmc:value_unsigned min="0" max="#TMC2208_MAX_ACCELERATION#" default="1000" unit="[pps/s]" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->actualAcceleration;
		} else if(readWrite == WRITE) {
			StepDirCh[motor]->actualAcceleration = *value;
		}
		break;
	default:
		errors |= ERROR_TYPE;
		break;
	}
	// \xml </axis_parameters>

	return errors;
}
// \xml </tmc:product_module>

static uint32 SAP(uint8 type, uint8 motor, int32 value)
{
	return handleParameter(WRITE, motor, type, &value);
}

static uint32 GAP(uint8 type, uint8 motor, int32 *value)
{
	return handleParameter(READ, motor, type, value);
}

static void writeRegister(uint8 address, int32 value)
{
	uint8 writeData[8];

	writeData[0] = 0x05;                         // Sync byte
	writeData[1] = 0x00;                         // Slave address
	writeData[2] = address | TMC2208_WRITE_BIT;  // Register address with write bit set
	writeData[3] = value >> 24;                  // Register Data
	writeData[4] = value >> 16;                  // Register Data
	writeData[5] = value >> 8;                   // Register Data
	writeData[6] = value & 0xFF;                 // Register Data
	writeData[7] = TMC2208_CRC(writeData, 7);    // Cyclic redundancy check

	TMC2208_UARTChannel->clearBuffers();
	for(uint32 i = 0; i < ARRAY_SIZE(writeData); i++)
		TMC2208_UARTChannel->tx(writeData[i]);

	TMC2208_config->shadowRegister[address & ~TMC2208_WRITE_BIT] = value;
}

static void readRegister(uint8 address, int32 *value)
{
	uint8 readData[8], dataRequest[4];
	uint32 timeout;

	// Clear write bit
	address &= ~TMC2208_WRITE_BIT;

	if(!IS_READABLE(TMC2208.registerAccess[address]))
	{	// Register not readable - shadowRegister copy
		*value = TMC2208_config->shadowRegister[address];
		return;
	}

	dataRequest[0] = 0x05;                         // Sync byte
	dataRequest[1] = 0x00;                         // Slave address
	dataRequest[2] = address;                      // Register address
	dataRequest[3] = TMC2208_CRC(dataRequest, 3);  // Cyclic redundancy check

	TMC2208_UARTChannel->clearBuffers();
	TMC2208_UARTChannel->txN(dataRequest, ARRAY_SIZE(dataRequest));

	// Wait for reply with timeout limit
	timeout = systick_getTick();
	while(TMC2208_UARTChannel->bytesAvailable() < ARRAY_SIZE(readData))
		if(timeSince(timeout) > TIMEOUT_VALUE) // Timeout
			return;

	TMC2208_UARTChannel->rxN(readData, ARRAY_SIZE(readData));
	// Check if the received data is correct (CRC, Sync, Slave address, Register address)
	// todo CHECK 2: Only keep CRC check? Should be sufficient for wrong transmissions (LH) #1
	if(readData[7] != TMC2208_CRC(readData, 7) || readData[0] != 0x05 || readData[1] != 0xFF || readData[2] != address)
		return;

	*value = readData[3] << 24 | readData[4] << 16 | readData[5] << 8 | readData[6];
	return;
}

static void checkErrors(uint32 tick)
{
	UNUSED(tick);
	Evalboards.ch2.errors = 0;
}

static uint32 userFunction(uint8 type, uint8 motor, int32 *value)
{
	uint32 errors = 0;

	UNUSED(type);
	UNUSED(motor);
	UNUSED(*value);

	return errors;
}

static void deInit(void)
{
	enableDriver(DRIVER_DISABLE);
	HAL.IOs->config->reset(Pins.DRV_ENN);

	HAL.IOs->config->reset(Pins.STEP);
	HAL.IOs->config->reset(Pins.DIR);
	HAL.IOs->config->reset(Pins.MS1);
	HAL.IOs->config->reset(Pins.MS2);
	HAL.IOs->config->reset(Pins.DIAG);
	HAL.IOs->config->reset(Pins.INDEX);

	StepDir_deInit();
}

static uint8 reset()
{
	return tmc2208_reset(TMC2208_config);
}

static uint8 restore()
{
	return tmc2208_restore(TMC2208_config);
}

static void enableDriver(DriverState state)
{
	if(state == DRIVER_USE_GLOBAL_ENABLE)
		state = Evalboards.driverEnable;

	if(state == DRIVER_DISABLE)
		HAL.IOs->config->setHigh(Pins.DRV_ENN);
	else if((state == DRIVER_ENABLE) && (Evalboards.driverEnable == DRIVER_ENABLE))
		HAL.IOs->config->setLow(Pins.DRV_ENN);
}

static void periodicJob(uint32 tick)
{
	for(int motor = 0; motor < MOTORS; motor++)
		tmc2208_periodicJob(motor, tick, &TMC2208, TMC2208_config);
}

void TMC2208_init(void)
{
	fillCRC8Table(0x07, TRUE, 1);

	tmc2208_initConfig(&TMC2208);

	Pins.DRV_ENN  = &HAL.IOs->pins->DIO0;
	Pins.STEP     = &HAL.IOs->pins->DIO6;
	Pins.DIR      = &HAL.IOs->pins->DIO7;
	Pins.MS1      = &HAL.IOs->pins->DIO3;
	Pins.MS2      = &HAL.IOs->pins->DIO4;
	Pins.DIAG     = &HAL.IOs->pins->DIO1;
	Pins.INDEX    = &HAL.IOs->pins->DIO2;

	HAL.IOs->config->toOutput(Pins.DRV_ENN);
	HAL.IOs->config->toOutput(Pins.STEP);
	HAL.IOs->config->toOutput(Pins.DIR);
	HAL.IOs->config->toOutput(Pins.MS1);
	HAL.IOs->config->toOutput(Pins.MS2);
	HAL.IOs->config->toInput(Pins.DIAG);
	HAL.IOs->config->toInput(Pins.INDEX);

	HAL.UART->init();

	TMC2208_UARTChannel = HAL.UART;

	TMC2208_config = Evalboards.ch2.config;


	Evalboards.ch2.config->reset        = reset;
	Evalboards.ch2.config->restore      = restore;
	Evalboards.ch2.config->state        = CONFIG_RESET;
	Evalboards.ch2.config->configIndex  = 0;

	Evalboards.ch2.rotate               = rotate;
	Evalboards.ch2.right                = right;
	Evalboards.ch2.left                 = left;
	Evalboards.ch2.stop                 = stop;
	Evalboards.ch2.GAP                  = GAP;
	Evalboards.ch2.SAP                  = SAP;
	Evalboards.ch2.moveTo               = moveTo;
	Evalboards.ch2.moveBy               = moveBy;
	Evalboards.ch2.writeRegister        = writeRegister;
	Evalboards.ch2.readRegister         = readRegister;
	Evalboards.ch2.userFunction         = userFunction;
	Evalboards.ch2.enableDriver         = enableDriver;
	Evalboards.ch2.checkErrors          = checkErrors;
	Evalboards.ch2.numberOfMotors       = MOTORS;
	Evalboards.ch2.VMMin                = VM_MIN;
	Evalboards.ch2.VMMax                = VM_MAX;
	Evalboards.ch2.deInit               = deInit;
	Evalboards.ch2.periodicJob          = periodicJob;

	StepDir_init(MOTORS);
	StepDirCh[0] = StepDir.ch1;

	StepDirCh[0]->stepOut              = Pins.STEP;
	StepDirCh[0]->dirOut               = Pins.DIR;
	StepDirCh[0]->maxPositioningSpeed  = 51200;
	StepDirCh[0]->actualAcceleration   = 51200;

	enableDriver(DRIVER_ENABLE);
};
