#include "../tmc/stepDir.h"

#include "Board.h"
#include "tmc/ic/TMC2660/TMC2660.h"

#define TMC2660_MAX_VELOCITY      STEPDIR_MAX_VELOCITY
#define TMC2660_MAX_ACCELERATION  16777215

#define ERRORS_I_STS          (1<<0)  // stand still current too high
#define ERRORS_I_TIMEOUT_STS  (1<<1)  // current limited in stand still to prevent driver from demage

#define VM_MIN  50   // VM[V/10] min
#define VM_MAX  293  // VM[V/10] max +10%

#define MOTORS 1

#define I_STAND_STILL 5
#define T_STAND_STILL 1000

#define DEFAULT_MOTOR 0

static uint32 right(uint8 motor, int32 velocity);
static uint32 left(uint8 motor, int32 velocity);
static uint32 rotate(uint8 motor, int32 velocity);
static uint32 stop(uint8 motor);
static uint32 moveTo(uint8 motor, int32 position);
static uint32 moveBy(uint8 motor, int32 *ticks);
static uint32 GAP(uint8 type, uint8 motor, int32 *value);
static uint32 SAP(uint8 type, uint8 motor, int32 value);
static void readRegister(uint8 address,	int32 *value);
static void writeRegister(uint8 address, int32 value);
static uint32 getMeasuredSpeed(uint8 motor, int32 *value);

static void periodicJob(uint32 tick);
static void deInit(void);
static uint32 userFunction(uint8 type, uint8 motor, int32 *value);

static uint8 reset();
static void enableDriver(DriverState state);

void tmc2660_writeInt(uint8 motor, uint8 address, int value);
uint32 tmc2660_readInt(uint8 motor, uint8 address);

static void coolStepActivationChanged(uint32 value);

static uint32 compatibilityMode = 1;

StepDirectionChannelType stepDirDummmy;
static StepDirectionChannelType *StepDirCh[] = {&stepDirDummmy};

SPIChannelTypeDef *TMC2660_SPIChannel;
TMC2660TypeDef TMC2660;
ConfigurationTypeDef *TMC2660_config;



typedef struct
{
	IOPinTypeDef  *CSN;
	IOPinTypeDef  *STEP;
	IOPinTypeDef  *DIR;
	IOPinTypeDef  *ENN;
	IOPinTypeDef  *SG_TST;
	IOPinTypeDef  *TEMP_BRIDGE;
} PinsTypeDef;

static PinsTypeDef Pins;

void readWrite(uint32 value)
{	// sending data (value) via spi to TMC262, coping written and received data to shadow register
	static uint8 rdsel = 0; // number of expected read response

// if SGCONF should be written, check whether stand still, or run current should be used
	if(TMC2660_GET_ADDRESS(value) == TMC2660_SGCSCONF)
	{
		value &= ~TMC2660_SET_CS(-1); // clear CS field
		value |= (TMC2660.isStandStillCurrentLimit) ?  TMC2660_SET_CS(TMC2660.standStillCurrentScale) : TMC2660_SET_CS(TMC2660.runCurrentScale); // set current
	}

// write value and read reply to shadow register
	TMC2660_config->shadowRegister[rdsel]  = TMC2660_SPIChannel->readWrite(value>>16, 0);
	TMC2660_config->shadowRegister[rdsel]  <<= 8;
	TMC2660_config->shadowRegister[rdsel]  |= TMC2660_SPIChannel->readWrite(value>>8, 0);
	TMC2660_config->shadowRegister[rdsel]  <<= 8;
	TMC2660_config->shadowRegister[rdsel]  |= TMC2660_SPIChannel->readWrite(value & 0xFF, 1);
	TMC2660_config->shadowRegister[rdsel]  >>= 4;

	TMC2660_config->shadowRegister[TMC2660_RESPONSE_LATEST] = TMC2660_config->shadowRegister[rdsel]; // copy value to latest field

// set virtual read address for next reply given by RDSEL, can only change by setting RDSEL in DRVCONF
	if(TMC2660_GET_ADDRESS(value) == TMC2660_DRVCONF)
		rdsel = TMC2660_GET_RDSEL(value);

// write store written value to shadow register
	TMC2660_config->shadowRegister[TMC2660_GET_ADDRESS(value) | TMC2660_WRITE ] = value;
}

void readImmediately(uint8 rdsel)
{ // sets desired reply in DRVCONF register, resets it to previous settings whilst reading desired reply
	uint32 value, drvConf;

// additional reading to keep all replies up to date
	value = tmc2660_readInt(0, TMC2660_WRITE | TMC2660_DRVCONF);  // buffer value amd  drvConf to write back later
	drvConf = value;
	value &= ~TMC2660_SET_RDSEL(-1);                              // clear RDSEL bits
	value |= TMC2660_SET_RDSEL((rdsel%3));                        // set rdsel
	readWrite(value);                                             // write to chip and readout reply
	readWrite(drvConf);                                           // write to chip and return desired reply
}

// => SPI wrapper
void tmc2660_writeInt(uint8 motor, uint8 address, int value)
{
	UNUSED(motor);

	// tmc2660_writeDatagram(address, 0xFF & (value>>24), 0xFF & (value>>16), 0xFF & (value>>8), 0xFF & (value>>0));
	value &= 0x0FFFFF;

	// store desired cs value, this can be overwritten by current limitation
	if(TMC2660_GET_ADDRESS(value) == TMC2660_SGCSCONF)
		TMC2660.runCurrentScale = TMC2660_GET_CS(value);

	TMC2660_config->shadowRegister[0x7F & (address | TMC2660_WRITE)] = value;
	if(!TMC2660.continuousModeEnable)
		readWrite(value);
}

uint32 tmc2660_readInt(uint8 motor, uint8 address)
{
	UNUSED(motor);

	if(!TMC2660.continuousModeEnable && !(address & TMC2660_WRITE))
		readImmediately(address);

	return TMC2660_config->shadowRegister[0x7F & address];
}

void tmc2660_readWrite(uint8 motor, uint32 value)
{
	UNUSED(motor);

	static uint8 rdsel = 0; // number of expected read response

	// if SGCONF should be written, check whether stand still, or run current should be used
	if(TMC2660_GET_ADDRESS(value) == TMC2660_SGCSCONF)
	{
		value &= ~TMC2660_SET_CS(-1); // clear CS field
		value |= (TMC2660.isStandStillCurrentLimit) ?  TMC2660_SET_CS(TMC2660.standStillCurrentScale) : TMC2660_SET_CS(TMC2660.runCurrentScale); // set current
	}

	// write value and read reply to shadow register
	TMC2660_config->shadowRegister[rdsel] = TMC2660_SPIChannel->readWrite(value>>16, 0);
	TMC2660_config->shadowRegister[rdsel] <<= 8;
	TMC2660_config->shadowRegister[rdsel] |= TMC2660_SPIChannel->readWrite(value>>8, 0);
	TMC2660_config->shadowRegister[rdsel] <<= 8;
	TMC2660_config->shadowRegister[rdsel] |= TMC2660_SPIChannel->readWrite(value & 0xFF, 1);
	TMC2660_config->shadowRegister[rdsel] >>= 4;

	TMC2660_config->shadowRegister[TMC2660_RESPONSE_LATEST] = TMC2660_config->shadowRegister[rdsel]; // copy value to latest field

	// set virtual read address for next reply given by RDSEL, can only change by setting RDSEL in DRVCONF
	if(TMC2660_GET_ADDRESS(value) == TMC2660_DRVCONF)
		rdsel = TMC2660_GET_RDSEL(value);

	// write store written value to shadow register
	TMC2660_config->shadowRegister[TMC2660_GET_ADDRESS(value) | TMC2660_WRITE ] = value;
}

static uint32 userFunction(uint8 type, uint8 motor, int32 *value)
{
	uint32 errors = 0;

	UNUSED(motor);

	switch(type)
	{
	case 0:	// disable continuos read/write mode - used in BoardAssignment.c for the combination TMC43XX + TMC2660
		// In continuos read/write mode settings will be continously written to TMC2660 and all replies are requested rotatory.
		// It's the default mode to prevent TMC2660 from loosing setting on brownout and being alway up to date with all chip states.
		TMC2660.continuousModeEnable = *value ? 0 : 1;
		break;
	case 1:	// disable compatibility mode
		// per default compability mode is enabled,
		// saying firmware works with orl TMC2660-Eval Tool
		// e.g. stallGuard value is only
		compatibilityMode = *value ? 0 : 1;
		break;
	default:
		errors |= ERROR_TYPE;
		break;
	}

	return errors;
}

static uint32 rotate(uint8 motor, int32 velocity)
{
	if(motor >= MOTORS)
		return ERROR_MOTOR;

	TMC2660.isStandStillCurrentLimit  = 0;
	TMC2660.isStandStillOverCurrent   = 0;

	StepDirCh[motor]->rampMode        = SD_RAMP_MODE_VELOCITY;
	StepDirCh[motor]->targetVelocity  = velocity;

	return ERROR_NONE;
}

static uint32 right(uint8 motor, int32 velocity)
{
	return rotate(motor, velocity);
}

static uint32 left(uint8 motor, int32 velocity)
{
	return rotate(motor, -velocity);
}

static uint32 stop(uint8 motor)
{
	return rotate(motor, 0);
}

static uint32 moveTo(uint8 motor, int32 position)
{
	if(motor >= MOTORS)
		return ERROR_MOTOR;

	StepDirCh[motor]->targetPosition  = position;
	StepDirCh[motor]->rampMode        = SD_RAMP_MODE_POSITIONING;

	return ERROR_NONE;
}

static uint32 moveBy(uint8 motor, int32 *ticks)
{
	if(motor >= MOTORS)
		return ERROR_MOTOR;

	// determine actual position and add numbers of ticks to move
	*ticks = StepDirCh[motor]->actualPosition + *ticks;

	return moveTo(motor, *ticks);
}

// \xml <?xml version="1.0" encoding="UTF-8" standalone="no"?>
// \xml <tmc:product_module author="?" xmlns:tmc="http://www.trinamic.com" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://www.trinamic.com https://www.trinamic.com/fileadmin/xml/TMC_Schema.xsd">
// \xml <name>TMC2660-EVAL</name>
// \xml <name_fw_version>2660</name_fw_version>
// \xml <hardware_version>1.0</hardware_version>
// \xml <software_version>2.91</software_version>
static uint32 handleParameter(u8 readWrite, u8 motor, u8 type, int32 *value)
{
	u32 errors = ERROR_NONE;

	if(motor >= MOTORS)
		return ERROR_MOTOR;

	// \xml <axis_parameters axis="0">
	switch(type)
	{
	case 0:
		// \xml <tmc:axis_parameter type="0" name="Target position" identifier="par::TargetPosition">
		// \xml   <desc>The desired target position in position mode.</desc>
		// \xml   <tmc:value_signed min="#s32_MIN#" max="#s32_MAX#" default="0" unit="[usteps]" access="RW"></tmc:value_signed>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->targetPosition;
		} else if(readWrite == WRITE) {
			StepDirCh[motor]->targetPosition = *value;
		}
		break;
	case 1:
		// \xml <tmc:axis_parameter type="1" name="Actual position" identifier="par::ActualPosition">
		// \xml   <desc>The actual position of the motor. Stop the motor before overwriting it. Should normally only be overwritten for reference position setting.</desc>
		// \xml   <tmc:value_signed min="#s32_MIN#" max="#s32_MAX#" default="0" unit="[usteps]" access="RW"></tmc:value_signed>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->actualPosition;
		} else if(readWrite == WRITE) {
			StepDirCh[motor]->actualPosition = *value;
		}
		break;
	case 2:
		// \xml <tmc:axis_parameter type="2" name="Target speed" identifier="par::TargetVelocity">
		// \xml   <desc>The desired speed in velocity mode. Not valid in position mode.</desc>
		// \xml   <tmc:value_signed min="-#TMC2660_MAX_VELOCITY#" max="#TMC2660_MAX_VELOCITY#" default="0" unit="[pps]" access="RW"></tmc:value_signed>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->targetVelocity;
		} else if(readWrite == WRITE) {
			StepDirCh[motor]->targetVelocity = *value;
		}
		break;
	case 3:
		// \xml <tmc:axis_parameter type="3" name="Actual speed" identifier="par::ActualVelocity">
		// \xml   <desc>The actual speed of the motor.</desc>
		// \xml   <tmc:value_signed min="-#TMC2660_MAX_VELOCITY#" max="#TMC2660_MAX_VELOCITY#" default="0" unit="[pps]" access="R"></tmc:value_signed>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->actualVelocity;
		} else if(readWrite == WRITE) {
			errors |= ERROR_TYPE;
		}
		break;
	case 4:
		// \xml <tmc:axis_parameter type="4" name="Maximum speed" identifier="par::MaxVelocity">
		// \xml   <desc>The maximum speed used for positioning ramps.</desc>
		// \xml   <tmc:value_unsigned min="0" max="#TMC2660_MAX_VELOCITY#" default="150000" unit="[pps]" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->maxPositioningSpeed;
		} else if(readWrite == WRITE) {
			StepDirCh[motor]->maxPositioningSpeed = abs(*value);
		}
		break;
	case 5:
		// \xml <tmc:axis_parameter type="5" name="Maximum acceleration" identifier="par::MaxAcceleration">
		// \xml   <desc>Maximum acceleration in positioning ramps. Acceleration and deceleration value in velocity mode.</desc>
		// \xml   <tmc:value_unsigned min="0" max="#TMC2660_MAX_ACCELERATION#" default="1000" unit="[pps/s]" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->actualAcceleration;
		} else if(readWrite == WRITE) {
			StepDirCh[motor]->actualAcceleration = *value;
		}
		break;
	case 6:
		// \xml <tmc:axis_parameter type="6" name="Maximum current" identifier="par::MaxCurrent">
		// \xml   <desc>Motor current used when motor is running. Where 0 = 3.125% up to 31 = 100% of maximum module current. Calculated by: maximum module current * (value + 1)/ 32.</desc>
		// \xml   <tmc:value_unsigned min="0" max="31" default="24" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2660.runCurrentScale;
		} else if(readWrite == WRITE) {
			TMC2660.runCurrentScale = *value;
			TMC2660_FIELD_UPDATE(0, TMC2660_SGCSCONF, TMC2660_CS_MASK, TMC2660_CS_SHIFT, *value);
		}
		break;
	case 7:
		// \xml <tmc:axis_parameter type="7" name="Standby current" identifier="par::StandbyCurrent">
		// \xml   <desc>The current used when the motor is not running. Where 0 = 3.125% up to 31 = 100% of maximum module current. Calculated by: maximum module current * (value + 1)/ 32. This value should be as low as possible so that the motor can cool down when it is not moving. Please see also parameter 214.</desc>
		// \xml   <tmc:value_unsigned min="0" max="31" default="3" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2660.standStillCurrentScale;
		} else if(readWrite == WRITE) {
			TMC2660.standStillCurrentScale = *value;
		}
		break;
	case 8:
		// \xml <tmc:axis_parameter type="8" name="Position reached flag" identifier="par::PositionReachedFlag">
		// \xml   <desc>This flag is always set when target position and actual position are equal.</desc>
		// \xml   <tmc:value_bool desc_false="Target position not reached" desc_true="Target position reached" default="false" access="R"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->targetReached;
		} else if(readWrite == WRITE) {
			errors |= ERROR_TYPE;
		}
		break;
	case 29:
		// todo XML CHECK 2: check description #2
		// \xml <tmc:axis_parameter type="29" name="Measured Speed" identifier="par::MeasuredSpeed">
		// \xml   <desc>Current speed value measured from the TMC5130 current position or from encoder.</desc>
		// \xml   <tmc:value_signed min="#s32_MIN#" max="#s32_MAX#" default="0" unit="" access="R"></tmc:value_signed>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->actualVelocity;
		} else if(readWrite == WRITE) {
			errors |= ERROR_TYPE;
		}
		break;
	case 140:
		// \xml <tmc:axis_parameter type="140" name="Microstep Resolution" identifier="par::MicrostepResolution">
		// \xml   <desc>Microstep resolutions per full step:</desc>
		// \xml   <tmc:value_choice default="256" unit="" access="RW">
		// \xml     <choice_element value="1" desc="fullstep"></choice_element>
		// \xml     <choice_element value="2" desc="halfstep"></choice_element>
		// \xml     <choice_element value="4" desc="4 microsteps"></choice_element>
		// \xml     <choice_element value="8" desc="8 microsteps"></choice_element>
		// \xml     <choice_element value="16" desc="16 microsteps"></choice_element>
		// \xml     <choice_element value="32" desc="32 microsteps"></choice_element>
		// \xml     <choice_element value="64" desc="64 microsteps"></choice_element>
		// \xml     <choice_element value="128" desc="128 microsteps"></choice_element>
		// \xml     <choice_element value="256" desc="256 microsteps"></choice_element>
		// \xml   </tmc:value_choice>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = 8 - TMC2660_FIELD_READ(0, TMC2660_DRVCTRL | TMC2660_WRITE, TMC2660_MRES_MASK, TMC2660_MRES_SHIFT);
		} else if(readWrite == WRITE) {
			switch(*value)
			{
			case 1:    *value = 8;   break;
			case 2:    *value = 7;   break;
			case 4:    *value = 6;   break;
			case 8:    *value = 5;   break;
			case 16:   *value = 4;   break;
			case 32:   *value = 3;   break;
			case 64:   *value = 2;   break;
			case 128:  *value = 1;   break;
			case 256:  *value = 0;   break;
			default:   *value = -1;  break;
			}

			if(*value != -1)
			{
				TMC2660_FIELD_UPDATE(0, TMC2660_DRVCTRL, TMC2660_MRES_MASK, TMC2660_MRES_SHIFT, *value);
			}
			else
			{
				errors |= ERROR_VALUE;
			}
		}
		break;
	case 160:
		// \xml <tmc:axis_parameter type="160" name="Microstep Interpolation" identifier="par::Intpol">
		// \xml   <desc>Interpolation of the current microstep resolution to 256 microsteps:</desc>
		// \xml   <tmc:value_bool desc_false="Disable STEP pulse interpolation." desc_true="Enable STEP pulse multiplication by 16." default="true" access="RWE"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2660_FIELD_READ(0, TMC2660_DRVCTRL | TMC2660_WRITE, TMC2660_INTPOL_MASK, TMC2660_INTPOL_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2660_FIELD_UPDATE(0, TMC2660_DRVCTRL, TMC2660_INTPOL_MASK, TMC2660_INTPOL_SHIFT, *value);
		}
		break;
	case 161:
		// \xml <tmc:axis_parameter type="161" name="Double Edge Steps" identifier="par::DoubleEdgeSteps">
		// \xml   <desc>Step impulse at each step edge to reduce step frequency requirement.</desc>
		// \xml   <tmc:value_bool desc_false="Rising edge active, falling inactive." desc_true="Both edges active." default="false" access="RWE"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2660_FIELD_READ(0, TMC2660_DRVCTRL | TMC2660_WRITE, TMC2660_DEDGE_MASK, TMC2660_DEDGE_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2660_FIELD_UPDATE(0, TMC2660_DRVCTRL, TMC2660_DEDGE_MASK, TMC2660_DEDGE_SHIFT, *value);
		}
		break;
	case 162:
		// \xml <tmc:axis_parameter type="162" name="Chopper blank time" identifier="par::ChopperBlankTime">
		// \xml   <desc>Selects the comparator blank time. This time needs to safely cover the switching event and the duration of the ringing on the sense resistor. Normally leave at the default value.</desc>
		// \xml   <tmc:value_unsigned min="0" max="3" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2660_FIELD_READ(0, TMC2660_CHOPCONF | TMC2660_WRITE, TMC2660_TBL_MASK, TMC2660_TBL_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2660_FIELD_UPDATE(0, TMC2660_CHOPCONF, TMC2660_TBL_MASK, TMC2660_TBL_SHIFT, *value);
		}
		break;
	case 163:
		// \xml <tmc:axis_parameter type="163" name="Constant TOff Mode" identifier="par::ConstantTOffMode">
		// \xml   <desc>Selection of the chopper mode.</desc>
		// \xml   <tmc:value_bool desc_false="spread cycle" desc_true="classic constant off time" default="false" access="RW"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2660_FIELD_READ(0, TMC2660_CHOPCONF | TMC2660_WRITE, TMC2660_CHM_MASK, TMC2660_CHM_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2660_FIELD_UPDATE(0, TMC2660_CHOPCONF, TMC2660_CHM_MASK, TMC2660_CHM_SHIFT, *value);
		}
		break;
	case 164:
		// \xml <tmc:axis_parameter type="164" name="Disable fast decay comparator" identifier="par::DisableFastDecayComparator">
		// \xml   <desc>See parameter 163. For "classic const. off time", setting this parameter to "1" will disable current comparator usage for termination of fast decay cycle.</desc>
		// \xml   <tmc:value_bool desc_false="enable comparator" desc_true="disable comparator" default="false" access="RW"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2660_FIELD_READ(0, TMC2660_CHOPCONF | TMC2660_WRITE, TMC2660_HDEC_MASK, TMC2660_HDEC_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2660_FIELD_UPDATE(0, TMC2660_CHOPCONF, TMC2660_HDEC_MASK, TMC2660_HDEC_SHIFT, *value);
		}
		break;
	case 165:
		// \xml <tmc:axis_parameter type="165" name="Chopper hysteresis end / fast decay time" identifier="par::ChopperHysteresisEnd">
		// \xml   <desc>See parameter 163. For "spread cycle" chopper mode this parameter will set / return the hysteresis end setting (hysteresis end value after a number of decrements). For "classic const. off time" chopper mode this parameter will set / return the fast decay time.</desc>
		// \xml   <tmc:value_unsigned min="0" max="15" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2660_FIELD_READ(0, TMC2660_CHOPCONF | TMC2660_WRITE, TMC2660_HEND_MASK, TMC2660_HEND_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2660_FIELD_UPDATE(0, TMC2660_CHOPCONF, TMC2660_HEND_MASK, TMC2660_HEND_SHIFT, *value);
		}
		break;
	case 166:
		// \xml <tmc:axis_parameter type="166" name="Chopper hysteresis start / sine wave offset" identifier="par::ChopperHysteresisStart">
		// \xml   <desc>See parameter 163. For "spread cycle" chopper mode this parameter will set / return the Hysteresis start setting (please note that this value is an offset to the hysteresis end value). For "classic const. off time" chopper mode this parameter will set / return the sine wave offset.</desc>
		// \xml   <tmc:value_unsigned min="0" max="8" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2660_FIELD_READ(0, TMC2660_CHOPCONF | TMC2660_WRITE, TMC2660_HSTRT_MASK, TMC2660_HSTRT_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2660_FIELD_UPDATE(0, TMC2660_CHOPCONF, TMC2660_HSTRT_MASK, TMC2660_HSTRT_SHIFT, *value);
		}
		break;
	case 167:
		// \xml <tmc:axis_parameter type="167" name="Chopper off time" identifier="par::TOff">
		// \xml   <desc>The off time setting controls the minimum chopper frequency. An off time within the range of 5us to 20us will fit. Off time setting for constant t Off chopper: NCLK = 12 + 32  tOFF (Minimum is 64 clocks) Setting this parameter to zero completely disables all driver transistors and the motor can freewheel.</desc>
		// \xml   <tmc:value_unsigned min="0" max="15" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2660_FIELD_READ(0, TMC2660_CHOPCONF | TMC2660_WRITE, TMC2660_TOFF_MASK, TMC2660_TOFF_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2660_FIELD_UPDATE(0, TMC2660_CHOPCONF, TMC2660_TOFF_MASK, TMC2660_TOFF_SHIFT, *value);
		}
		break;
	case 168:
		// \xml <tmc:axis_parameter type="168" name="smartEnergy current minimum (SEIMIN)" identifier="par::SEIMIN">
		// \xml   <desc>Sets the lower motor current limit for coolStep operation by scaling the maximum current (see axis parameter 6) value.</desc>
		// \xml   <tmc:value_bool desc_false="1/2 of CS" desc_true="1/4 of CS" default="false" access="RW"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2660_FIELD_READ(0, TMC2660_SMARTEN, TMC2660_SEIMIN_MASK, TMC2660_SEIMIN_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2660_FIELD_UPDATE(0, TMC2660_SMARTEN, TMC2660_SEIMIN_MASK, TMC2660_SEIMIN_SHIFT, *value);
		}
		break;
	case 169:
		// \xml <tmc:axis_parameter type="169" name="smartEnergy current down step" identifier="par::SECDS">
		// \xml   <desc>Sets the number of stallGuard2 readings above the upper threshold necessary for each current decrement of the motor current. Number of stallGuard2 measurements per decrement: Scaling: 0. . . 3: 32, 8, 2, 1. 0: slow decrement, 3: fast decrement</desc>
		// \xml   <tmc:value_unsigned min="0" max="3" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2660_FIELD_READ(0, TMC2660_SMARTEN, TMC2660_SEDN_MASK, TMC2660_SEDN_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2660_FIELD_UPDATE(0, TMC2660_SMARTEN, TMC2660_SEDN_MASK, TMC2660_SEDN_SHIFT, *value);
		}
		break;
	case 170:
		// \xml <tmc:axis_parameter type="170" name="smartEnergy hysteresis" identifier="par::smartEnergyHysteresis">
		// \xml   <desc>Sets the distance between the lower and the upper threshold for stallGuard2 reading. Above the upper threshold the motor current becomes decreased. Hysteresis: ([AP172] + 1) * 32. Upper stallGuard threshold: ([AP172] + [AP170] + 1) * 32</desc>
		// \xml   <tmc:value_unsigned min="0" max="15" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2660_FIELD_READ(0, TMC2660_SMARTEN, TMC2660_SEMAX_MASK, TMC2660_SEMAX_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2660_FIELD_UPDATE(0, TMC2660_SMARTEN, TMC2660_SEMAX_MASK, TMC2660_SEMAX_SHIFT, *value);
		}
		break;
	case 171:
		// \xml <tmc:axis_parameter type="171" name="smartEnergy current up step" identifier="par::SECUS">
		// \xml   <desc>Sets the current increment step. The current becomes incremented for each measured stallGuard2 value below the lower threshold see smartEnergy hysteresis start). Current increment step size: Scaling: 0. . . 3: 1, 2, 4, 8. 0: slow increment, 3: fast increment / fast reaction to rising load</desc>
		// \xml   <tmc:value_unsigned min="0" max="3" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2660_FIELD_READ(0, TMC2660_SMARTEN, TMC2660_SEUP_MASK, TMC2660_SEUP_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2660_FIELD_UPDATE(0, TMC2660_SMARTEN, TMC2660_SEUP_MASK, TMC2660_SEUP_SHIFT, *value);
		}
		break;
	case 172:
		// \xml <tmc:axis_parameter type="172" name="smartEnergy hysteresis start" identifier="par::smartEnergyHysteresisStart">
		// \xml   <desc>The lower threshold for the stallGuard2 value (see smart Energy current up step).</desc>
		// \xml   <tmc:value_unsigned min="0" max="15" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDir.ch1->coolStepActiveValue;
		} else if(readWrite == WRITE) {
			StepDir.ch1->coolStepActiveValue = *value;
		}
		break;
	case 173:
		// \xml <tmc:axis_parameter type="173" name="stallGuard2 filter enable" identifier="par::SG2FilterEnable">
		// \xml   <desc>Enables the stallGuard2 filter for more precision of the movement. If set, reduces the measurement frequency to one measurement per four fullsteps. In most cases it is expedient to set the filtered mode before using coolStep. Use the standard mode for step loss detection.</desc>
		// \xml   <tmc:value_bool desc_false="standard mode" desc_true="filtered mode" default="false" access="RW"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2660_FIELD_READ(0, TMC2660_SGCSCONF, TMC2660_SFILT_MASK, TMC2660_SFILT_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2660_FIELD_UPDATE(0, TMC2660_SGCSCONF, TMC2660_SFILT_MASK, TMC2660_SFILT_SHIFT, *value);
		}
		break;
	case 174:
		// \xml <tmc:axis_parameter type="174" name="stallGuard2 threshold" identifier="par::SG2Threshold">
		// \xml   <desc>This signed value controls stallGuard2 threshold level for stall output and sets the optimum measurement range for readout. A lower value gives a higher sensitivity. Zero is the starting value. A higher value makes stallGuard2 less sensitive and requires more torque to indicate a stall.</desc>
		// \xml   <tmc:value_signed min="-64" max="63" default="0" unit="" access="RW"></tmc:value_signed>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2660_FIELD_READ(0, TMC2660_SGCSCONF, TMC2660_SGT_MASK, TMC2660_SGT_SHIFT);
			*value = CAST_Sn_TO_S32(*value, 7);
		} else if(readWrite == WRITE) {
			TMC2660_FIELD_UPDATE(0, TMC2660_SGCSCONF, TMC2660_SGT_MASK, TMC2660_SGT_SHIFT, *value);
		}
		break;
	case 175:
		// \xml <tmc:axis_parameter type="175" name="Slope control, high side" identifier="par::SLPH">
		// \xml   <desc>In temperature compensated mode (tc), the MOSFET gate driver strength is increased if the over-temperature warning temperature is reached. This compensates for temperature dependency of high-side slope control.</desc>
		// \xml   <tmc:value_choice default="0" unit="" access="RW">
		// \xml     <choice_element value="0" desc="Minimum"></choice_element>
		// \xml     <choice_element value="1" desc="Minimum temperature compensation mode"></choice_element>
		// \xml     <choice_element value="2" desc="Medium temperature compensation mode"></choice_element>
		// \xml     <choice_element value="3" desc="Maximum"></choice_element>
		// \xml   </tmc:value_choice>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2660_FIELD_READ(0, TMC2660_DRVCONF, TMC2660_SLPH_MASK, TMC2660_SLPH_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2660_FIELD_UPDATE(0, TMC2660_DRVCONF, TMC2660_SLPH_MASK, TMC2660_SLPH_SHIFT, *value);
		}
		break;
	case 176:
		// \xml <tmc:axis_parameter type="176" name="Slope control, low side" identifier="par::SLPL">
		// \xml   <desc>In temperature compensated mode (tc), the MOSFET gate driver strength is increased if the over-temperature warning temperature is reached. This compensates for temperature dependency of low-side slope control.</desc>
		// \xml   <tmc:value_choice default="0" unit="" access="RW">
		// \xml     <choice_element value="0" desc="Minimum"></choice_element>
		// \xml     <choice_element value="1" desc="Minimum temperature compensation mode"></choice_element>
		// \xml     <choice_element value="2" desc="Medium temperature compensation mode"></choice_element>
		// \xml     <choice_element value="3" desc="Maximum"></choice_element>
		// \xml   </tmc:value_choice>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2660_FIELD_READ(0, TMC2660_DRVCONF, TMC2660_SLPL_MASK, TMC2660_SLPL_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2660_FIELD_UPDATE(0, TMC2660_DRVCONF, TMC2660_SLPL_MASK, TMC2660_SLPL_SHIFT, *value);
		}
		break;
	case 177:
		// \xml <tmc:axis_parameter type="177" name="Short to Ground Protection" identifier="par::ShortToGroundProtection">
		// \xml   <desc>Disable short to ground protection.</desc>
		// \xml   <tmc:value_bool desc_false="Short to GND protection is enabled" desc_true="Short to GND protection is disabled" default="false" access="RW"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2660_FIELD_READ(0, TMC2660_DRVCONF, TMC2660_DISS2G_MASK, TMC2660_DISS2G_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2660_FIELD_UPDATE(0, TMC2660_DRVCONF, TMC2660_DISS2G_MASK, TMC2660_DISS2G_SHIFT, *value);
		}
		break;
	case 178:
		// \xml <tmc:axis_parameter type="178" name="Short-to-ground detection timer" identifier="par::SLPL">
		// \xml   <desc>Timing for the short-to-ground detection.</desc>
		// \xml   <tmc:value_choice default="0" unit="" access="RW">
		// \xml     <choice_element value="0" desc="3.2 us"></choice_element>
		// \xml     <choice_element value="1" desc="1.6 us"></choice_element>
		// \xml     <choice_element value="2" desc="1.2 us"></choice_element>
		// \xml     <choice_element value="3" desc="0.8 us"></choice_element>
		// \xml   </tmc:value_choice>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2660_FIELD_READ(0, TMC2660_DRVCONF, TMC2660_TS2G_MASK, TMC2660_TS2G_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2660_FIELD_UPDATE(0, TMC2660_DRVCONF, TMC2660_TS2G_MASK, TMC2660_TS2G_SHIFT, *value);
		}
		break;
	case 179:
		// \xml <tmc:axis_parameter type="179" name="VSense" identifier="par::VSense">
		// \xml   <desc>Sense resistor voltage-based current scaling. (Full-scale refers to a current setting of 31 and a DAC value of 255.)</desc>
		// \xml   <tmc:value_bool desc_false="Full-scale sense resistor voltage is 305mV." desc_true="Full-scale sense resistor voltage is 165mV." default="false" access="RW"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2660_FIELD_READ(0, TMC2660_DRVCONF, TMC2660_VSENSE_MASK, TMC2660_VSENSE_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2660_FIELD_UPDATE(0, TMC2660_DRVCONF, TMC2660_VSENSE_MASK, TMC2660_VSENSE_SHIFT, *value);
		}
		break;
	case 180:
		// \xml <tmc:axis_parameter type="180" name="smartEnergy actual current" identifier="par::smartEnergyActualCurrent">
		// \xml   <desc>This status value provides the actual motor current setting as controlled by coolStep. The value goes up to the CS value and down to the portion of CS as specified by SEIMIN. Actual motor current scaling factor: 0. . . 31: 1/32, 2/32, . . . 32/32</desc>
		// \xml   <tmc:value_unsigned min="0" max="31" default="0" unit="" access="R"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2660_FIELD_READ(0, TMC2660_RESPONSE2, TMC2660_SE_MASK, TMC2660_SE_SHIFT);
		} else if(readWrite == WRITE) {
			errors |= ERROR_TYPE;
		}
		break;
	case 181:
		// \xml <tmc:axis_parameter type="181" name="smartEnergy stall velocity" identifier="par::smartEnergyStallVelocity">
		// \xml   <desc>Velocity from which stop on stall feature is active</desc>
		// \xml   <tmc:value_unsigned min="0" max="#TMC2660_MAX_VELOCITY#" default="0" unit="[pps]" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->stopOnStallVelocityTreshold;
		} else if(readWrite == WRITE) {
			StepDirCh[motor]->stopOnStallVelocityTreshold  = *value;
			StepDirCh[motor]->stallFlagIn                  = 0;
		}
		break;
	case 182:
		// \xml <tmc:axis_parameter type="182" name="smartEnergy threshold speed" identifier="par::smartEnergyThresholdSpeed">
		// \xml   <desc>Above this speed coolStep becomes enabled.</desc>
		// \xml   <tmc:value_unsigned min="0" max="7999774" default="0" unit="[pps]" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = StepDirCh[motor]->coolStepVelocityTreshold;
		} else if(readWrite == WRITE) {
			StepDirCh[motor]->coolStepVelocityTreshold = *value;
		}
		break;
	case 183:
		// \xml <tmc:axis_parameter type="183" name="Disable step/dir interface" identifier="par::SDOFF">
		// \xml   <desc>Disable step/dir interace and use SPI interface to move the motor.</desc>
		// \xml   <tmc:value_bool desc_false="Enable step/dir" desc_true="Disable step/dir" default="false" access="RW"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2660_FIELD_READ(0, TMC2660_DRVCONF, TMC2660_SDOFF_MASK, TMC2660_SDOFF_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2660_FIELD_UPDATE(0, TMC2660_DRVCONF, TMC2660_SDOFF_MASK, TMC2660_SDOFF_SHIFT, *value);
		}
		break;
	case 184:
		// \xml <tmc:axis_parameter type="184" name="Random TOff mode" identifier="par::RandomTOffMode">
		// \xml   <desc>Enable / disable random TOff mode.</desc>
		// \xml   <tmc:value_bool desc_false="Chopper off time is fixed" desc_true="Chopper off time is random" default="false" access="RW"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2660_FIELD_READ(0, TMC2660_CHOPCONF, TMC2660_RNDTF_MASK, TMC2660_RNDTF_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2660_FIELD_UPDATE(0, TMC2660_CHOPCONF, TMC2660_RNDTF_MASK, TMC2660_RNDTF_SHIFT, *value);
		}
		break;
	case 185:
		// Reserved test mode: leave undocumented?
		if(readWrite == READ) {
			*value = TMC2660_FIELD_READ(0, TMC2660_DRVCONF, TMC2660_TST_MASK, TMC2660_TST_SHIFT);
		} else if(readWrite == WRITE) {
			TMC2660_FIELD_UPDATE(0, TMC2660_DRVCONF, TMC2660_TST_MASK, TMC2660_TST_SHIFT, *value);
		}
		break;
	case 206:
		// \xml <tmc:axis_parameter type="206" name="Load value" identifier="par::LoadValue">
		// \xml   <desc>Actual current control scaling for monitoring smart energy current scaling or automatic current scaling.</desc>
		// \xml   <tmc:value_unsigned min="0" max="1023" default="0" unit="" access="R"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = (compatibilityMode) ?
					TMC2660_FIELD_READ(0, TMC2660_RESPONSE2, TMC2660_SGU_MASK, TMC2660_SGU_SHIFT)<<5 :
					TMC2660_FIELD_READ(0, TMC2660_RESPONSE1, TMC2660_SG2_MASK, TMC2660_SG2_SHIFT);
		} else if(readWrite == WRITE) {
			errors |= ERROR_TYPE;
		}
		break;
	case 208:
		// \xml <tmc:axis_parameter type="208" name="Status Flags" identifier="par::DrvStatusFlags">
		// \xml   <desc>TMC2660 status flags.</desc>
		// \xml   <tmc:value_unsigned min="0" max="255" default="0" unit="" access="R"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2660_FIELD_READ(0, TMC2660_RESPONSE_LATEST, TMC2660_STATUS_MASK, TMC2660_STATUS_SHIFT);
		} else if(readWrite == WRITE) {
			errors |= ERROR_TYPE;
		}
		break;
	case 214:
		// \xml <tmc:axis_parameter type="214" name="Power Down Delay" identifier="par::PowerDownDelay">
		// \xml   <desc>Standstill period before the current will be ramped down to standby current. The standard value is 200 (which means 2000ms).</desc>
		// \xml   <tmc:value_signed min="0" max="#u16_MAX#" default="0" unit="[10 ms]" access="RW"></tmc:value_signed>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC2660.standStillTimeout;
		} else if(readWrite == WRITE) {
			TMC2660.standStillTimeout = *value;
		}
		break;
	default:
		errors |= ERROR_TYPE;
		break;
	}
	// \xml </axis_parameters>
	return errors;
}
// \xml </tmc:product_module>

static uint32 SAP(uint8 type, uint8 motor, int32 value)
{
	return handleParameter(WRITE, motor, type, &value);
}

static uint32 GAP(uint8 type, uint8 motor, int32 *value)
{
	return handleParameter(READ, motor, type, value);
}

static void writeRegister(uint8 address, int32 value)
{
	tmc2660_writeInt(0, address, value);
}

static void readRegister(uint8 address, int32 *value)
{
	*value = tmc2660_readInt(0, address);
}

static uint32 getMeasuredSpeed(uint8 motor, int32 *value)
{
	if(motor >= MOTORS)
		return ERROR_MOTOR;

	switch(motor)
	{
	case 0:
		*value = StepDir.ch1->actualVelocity;
		break;
	default:
		return ERROR_MOTOR;
		break;
	}
	return ERROR_NONE;
}

static void coolStepActivationChanged(uint32 value)
{
	TMC2660_FIELD_UPDATE(0, TMC2660_SMARTEN, TMC2660_SEMIN_MASK, TMC2660_SEMIN_SHIFT, value);
}

static void deInit(void)
{
	enableDriver(DRIVER_DISABLE);

	HAL.IOs->config->setHigh(Pins.ENN);

	HAL.IOs->config->reset(Pins.CSN);
	HAL.IOs->config->reset(Pins.DIR);
	HAL.IOs->config->reset(Pins.ENN);
	HAL.IOs->config->reset(Pins.SG_TST);
	HAL.IOs->config->reset(Pins.STEP);

	StepDir_deInit();
}

static void periodicJob(uint32 tick)
{
	tmc2660_periodicJob(DEFAULT_MOTOR, tick, &TMC2660, TMC2660_config);

	Evalboards.ch2.errors = (TMC2660.isStandStillOverCurrent) 	? (Evalboards.ch2.errors | ERRORS_I_STS) 			: (Evalboards.ch2.errors & ~ERRORS_I_STS);
	Evalboards.ch2.errors = (TMC2660.isStandStillCurrentLimit) 	? (Evalboards.ch2.errors | ERRORS_I_TIMEOUT_STS) 	: (Evalboards.ch2.errors & ~ERRORS_I_TIMEOUT_STS);


//  Uncomment keep motors halted when no supply is detected to not to loose steps n this case.
//	vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
	//	if(VitalSignsMonitor.brownOut)								StepDir.ch1->haltMask 	|= TMC2660_SD_HALT_BROWNOUT;
	//	else if(StepDir.ch1->haltMask & TMC2660_SD_HALT_BROWNOUT)	StepDir.ch1->haltMask 	&= 	~TMC2660_SD_HALT_BROWNOUT;
// ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^


}

static uint8 reset()
{
	if(StepDirCh[0]->actualVelocity != 0)
		return 0;

	tmc2660_reset(&TMC2660, TMC2660_config);
	StepDirCh[0]->stallFlagIn = 0;
	compatibilityMode = 1;
	enableDriver(DRIVER_USE_GLOBAL_ENABLE);

	return 1;
}

static uint8 restore()
{
	return tmc2660_restore(TMC2660_config);
}

static void enableDriver(DriverState state)
{
	if(state == DRIVER_USE_GLOBAL_ENABLE)
		state = Evalboards.driverEnable;

	if(state == DRIVER_DISABLE)
		HAL.IOs->config->setHigh(Pins.ENN);
	else if((state == DRIVER_ENABLE) && (Evalboards.driverEnable == DRIVER_ENABLE))
		HAL.IOs->config->setLow(Pins.ENN);
}

void TMC2660_init(void)
{
	compatibilityMode = 1;

	tmc2660_initConfig(&TMC2660);

	Pins.ENN     = &HAL.IOs->pins->DIO0;
	Pins.SG_TST  = &HAL.IOs->pins->DIO1;
	Pins.STEP    = &HAL.IOs->pins->DIO6;
	Pins.DIR     = &HAL.IOs->pins->DIO7;
	Pins.CSN     = &HAL.IOs->pins->SPI2_CSN0;

	HAL.IOs->config->toOutput(Pins.STEP);
	HAL.IOs->config->toOutput(Pins.DIR);
	HAL.IOs->config->toOutput(Pins.ENN);
	HAL.IOs->config->toInput(Pins.SG_TST);
	HAL.IOs->config->toOutput(Pins.CSN);

#if defined(Startrampe)
	Pins.TEMP_BRIDGE	= &HAL.IOs->pins->AIN0;
	HAL.IOs->config->reset(Pins.TEMP_BRIDGE);
#endif

	TMC2660_SPIChannel = &HAL.SPI->ch2;
	TMC2660_SPIChannel->CSN = Pins.CSN;

	TMC2660.standStillCurrentScale  = I_STAND_STILL;
	TMC2660.standStillTimeout       = T_STAND_STILL;

	StepDir_init(MOTORS);
	StepDirCh[0] = StepDir.ch1;
	StepDirCh[0]->stepOut                    = Pins.STEP;
	StepDirCh[0]->dirOut                     = Pins.DIR;
	StepDirCh[0]->stallSignalIn              = Pins.SG_TST;
	StepDirCh[0]->coolStepActivationChanged  = coolStepActivationChanged;
	StepDirCh[0]->coolStepInactiveValue      = 0;

	TMC2660_config = Evalboards.ch2.config;

	Evalboards.ch2.config->restore      = restore;
	Evalboards.ch2.config->reset        = reset;
	Evalboards.ch2.config->state        = CONFIG_READY; // Not used, leave this as CONFIG_READY to indicate chip not being busy
	Evalboards.ch2.config->configIndex  = 0;

	Evalboards.ch2.rotate               = rotate;
	Evalboards.ch2.right                = right;
	Evalboards.ch2.left                 = left;
	Evalboards.ch2.stop                 = stop;
	Evalboards.ch2.GAP                  = GAP;
	Evalboards.ch2.SAP                  = SAP;
	Evalboards.ch2.moveTo               = moveTo;
	Evalboards.ch2.moveBy               = moveBy;
	Evalboards.ch2.writeRegister        = writeRegister;
	Evalboards.ch2.readRegister         = readRegister;
	Evalboards.ch2.periodicJob          = periodicJob;
	Evalboards.ch2.userFunction         = userFunction;
	Evalboards.ch2.getMeasuredSpeed     = getMeasuredSpeed;
	Evalboards.ch2.enableDriver         = enableDriver;
	Evalboards.ch2.numberOfMotors       = MOTORS;
	Evalboards.ch2.VMMin                = VM_MIN;
	Evalboards.ch2.VMMax                = VM_MAX;
	Evalboards.ch2.deInit               = deInit;

	enableDriver(DRIVER_USE_GLOBAL_ENABLE);
}

