#include "Board.h"
#include "tmc/ic/TMC5062/TMC5062.h"

#define ERRORS_VM        (1<<0)
#define ERRORS_VM_UNDER  (1<<1)
#define ERRORS_VM_OVER   (1<<2)

#define VM_MIN  50   // VM[V/10] min
#define VM_MAX  222  // VM[V/10] max +10%

#define MOTORS 2

#define TMC5062_MAX_VELOCITY 8388096

static uint32 right(uint8 motor, int32 velocity);
static uint32 left(uint8 motor, int32 velocity);
static uint32 rotate(uint8 motor, int32 velocity);
static uint32 stop(uint8 motor);
static uint32 moveTo(uint8 motor, int32 position);
static uint32 moveBy(uint8 motor, int32 *ticks);
static uint32 GAP(uint8 type, uint8 motor, int32 *value);
static uint32 SAP(uint8 type, uint8 motor, int32 value);

static void readRegister(uint8 address, int32 *value);
static void writeRegister(uint8 address, int32 value);
static uint32 getMeasuredSpeed(uint8 motor, int32 *value);

static void periodicJob(uint32 tick);
static void checkErrors	(uint32 tick);
static void deInit(void);
static uint32 userFunction(uint8 type, uint8 motor, int32 *value);

static uint8 reset();
static void enableDriver(DriverState state);

SPIChannelTypeDef *TMC5062_SPIChannel;
TMC5062TypeDef TMC5062;
ConfigurationTypeDef *TMC5062_config;

typedef struct
{
	IOPinTypeDef  *DRV_ENN;
	IOPinTypeDef  *INT_ENCA;
	IOPinTypeDef  *PP_ENCB;
	IOPinTypeDef  *SWSEL;
	IOPinTypeDef  *SWIOP1;
	IOPinTypeDef  *SWIOP2;
	IOPinTypeDef  *SWION;
} PinsTypeDef;

static PinsTypeDef Pins;

void tmc5062_writeDatagram(u8 motor, uint8 address, uint8 x1, uint8 x2, uint8 x3, uint8 x4)
{
	UNUSED(motor);
	int value;

	TMC5062_SPIChannel->readWrite(address | TMC5062_WRITE_BIT, FALSE);
	TMC5062_SPIChannel->readWrite(x1, FALSE);
	TMC5062_SPIChannel->readWrite(x2, FALSE);
	TMC5062_SPIChannel->readWrite(x3, FALSE);
	TMC5062_SPIChannel->readWrite(x4, TRUE);

	value = x1;
	value <<= 8;
	value |= x2;
	value <<= 8;
	value |= x3;
	value <<= 8;
	value |= x4;
	TMC5062_config->shadowRegister[address & 0x7F] = value;
}

void tmc5062_writeInt(u8 motor, uint8 Address, int Value)
{
	tmc5062_writeDatagram(motor, Address, 0xFF & (Value>>24), 0xFF & (Value>>16), 0xFF & (Value>>8), 0xFF & (Value>>0));
}

int tmc5062_readInt(u8 motor, uint8 address)
{
	UNUSED(motor);
	int value;

	address &= 0x7F;

	// register not readable -> shadow register copy
	if(!IS_READABLE(TMC5062.registerAccess[address]))
		return TMC5062_config->shadowRegister[address];

	TMC5062_SPIChannel->readWrite(address, FALSE);
	TMC5062_SPIChannel->readWrite(0, FALSE);
	TMC5062_SPIChannel->readWrite(0, FALSE);
	TMC5062_SPIChannel->readWrite(0, FALSE);
	TMC5062_SPIChannel->readWrite(0, TRUE);

	TMC5062_SPIChannel->readWrite(address, FALSE);
	value = TMC5062_SPIChannel->readWrite(0, FALSE);
	value <<= 8;
	value |= TMC5062_SPIChannel->readWrite(0, FALSE);
	value <<= 8;
	value |= TMC5062_SPIChannel->readWrite(0, FALSE);
	value <<= 8;
	value |= TMC5062_SPIChannel->readWrite(0, TRUE);

	return value;
}

static uint32 rotate(uint8 motor, int32 velocity)
{
	if(motor >= MOTORS)
		return ERROR_MOTOR;

	tmc5062_writeInt(motor, TMC5062_VMAX(motor), abs(velocity));
	TMC5062.vMaxModified[motor] = TRUE;
	tmc5062_writeDatagram(motor, TMC5062_RAMPMODE(motor), 0, 0, 0, (velocity >= 0)? TMC5062_MODE_VELPOS:TMC5062_MODE_VELNEG);

	return ERROR_NONE;
}

static uint32 right(uint8 motor, int32 velocity)
{
	return rotate(motor, velocity);
}

static uint32 left(uint8 motor, int32 velocity)
{
	return rotate(motor, -velocity);
}

static uint32 stop(uint8 motor)
{
	return rotate(motor, 0);
}

static uint32 moveTo(uint8 motor, int32 position)
{
	if(motor >= MOTORS)
		return ERROR_MOTOR;

	if(TMC5062.vMaxModified[motor])
	{
		tmc5062_writeInt(motor, TMC5062_VMAX(motor), TMC5062_config->shadowRegister[TMC5062_VMAX(motor)]);
		TMC5062.vMaxModified[motor] = FALSE;
	}
	tmc5062_writeInt(motor, TMC5062_XTARGET(motor), position);
	tmc5062_writeDatagram(motor, TMC5062_RAMPMODE(motor), 0, 0, 0, TMC5062_MODE_POSITION);

	return ERROR_NONE;
}

static uint32 moveBy(uint8 motor, int32 *ticks)
{
	// determine actual position and add numbers of ticks to move
	*ticks = tmc5062_readInt(motor, TMC5062_XACTUAL(motor)) + *ticks;

	return moveTo(motor, *ticks);
}

// \xml <?xml version="1.0" encoding="UTF-8" standalone="no"?>
// \xml <tmc:product_module author="?" xmlns:tmc="http://www.trinamic.com" xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" xsi:schemaLocation="http://www.trinamic.com https://www.trinamic.com/fileadmin/xml/TMC_Schema.xsd">
// \xml <name>TMC5062-EVAL</name>
// \xml <name_fw_version>5062</name_fw_version>
// \xml <hardware_version>1.0</hardware_version>
// \xml <software_version>2.91</software_version>
static uint32 handleParameter(u8 readWrite, u8 motor, u8 type, int32 *value)
{
	u32 errors = ERROR_NONE;
	int tempValue;

	if(motor >= MOTORS)
		return ERROR_MOTOR;

	// \xml <axis_parameters axis="0">
	switch(type)
	{
	case 0:
		// \xml <tmc:axis_parameter type="0" name="Target position" identifier="par::TargetPosition">
		// \xml   <desc>The desired target position in position mode.</desc>
		// \xml   <tmc:value_signed min="#s32_MIN#" max="#s32_MAX#" default="0" unit="[usteps]" access="RW"></tmc:value_signed>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = tmc5062_readInt(motor, TMC5062_XTARGET(motor));
		} else if(readWrite == WRITE) {
			tmc5062_writeInt(motor, TMC5062_XTARGET(motor), *value);
		}
		break;
	case 1:
		// \xml <tmc:axis_parameter type="1" name="Actual position" identifier="par::ActualPosition">
		// \xml   <desc>The actual position of the motor. Stop the motor before overwriting it. Should normally only be overwritten for reference position setting.</desc>
		// \xml   <tmc:value_signed min="#s32_MIN#" max="#s32_MAX#" default="0" unit="[usteps]" access="RW"></tmc:value_signed>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = tmc5062_readInt(motor, TMC5062_XACTUAL(motor));
		} else if(readWrite == WRITE) {
			tmc5062_writeInt(motor, TMC5062_XACTUAL(motor), *value);
		}
		break;
	case 2:
		// \xml <tmc:axis_parameter type="2" name="Target speed" identifier="par::TargetVelocity">
		// \xml   <desc>The desired speed in velocity mode. Not valid in position mode.</desc>
		// \xml   <tmc:value_unsigned min="0" max="#TMC5062_MAX_VELOCITY#" default="0" unit="[pps]" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = tmc5062_readInt(motor, TMC5062_VMAX(motor));
		} else if(readWrite == WRITE) {
			TMC5062.vMaxModified[motor] = TRUE;
			tmc5062_writeInt(motor, TMC5062_VMAX(motor), abs(*value));
		}
		break;
	case 3:
		// todo CHECK 3: min max actually velocity min and velocity max? (JE) #4
		// \xml <tmc:axis_parameter type="3" name="Actual speed" identifier="par::ActualVelocity">
		// \xml   <desc>The actual speed of the motor.</desc>
		// \xml   <tmc:value_signed min="#s32_MIN#" max="#s32_MAX#" default="0" unit="[pps]" access="R"></tmc:value_signed>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = tmc5062_readInt(motor, TMC5062_VACTUAL(motor));
			*value = CAST_Sn_TO_S32(*value, 24);
		} else if(readWrite == WRITE) {
			errors |= ERROR_TYPE;
		}
		break;
	case 4:
		// \xml <tmc:axis_parameter type="4" name="Maximum speed" identifier="par::MaxVelocity">
		// \xml   <desc>The maximum speed used for positioning ramps.</desc>
		// \xml   <tmc:value_unsigned min="0" max="#TMC5062_MAX_VELOCITY#" default="0" unit="[pps]" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC5062_config->shadowRegister[TMC5062_VMAX(motor)];
		} else if(readWrite == WRITE) {
			TMC5062_config->shadowRegister[TMC5062_VMAX(motor)] = abs(*value);
			if(tmc5062_readInt(motor, TMC5062_RAMPMODE(motor)) == TMC5062_MODE_POSITION)
				tmc5062_writeInt(motor, TMC5062_VMAX(motor), abs(*value));
		}
		break;
	case 5:
		// \xml <tmc:axis_parameter type="5" name="Maximum acceleration" identifier="par::MaxAcceleration">
		// \xml   <desc>Maximum acceleration in positioning ramps. Acceleration and deceleration value in velocity mode.</desc>
		// \xml   <tmc:value_unsigned min="0" max="#u16_MAX#" default="0" unit="[pps/s]" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = tmc5062_readInt(motor, TMC5062_AMAX(motor));
		} else if(readWrite == WRITE) {
			tmc5062_writeInt(motor, TMC5062_AMAX(motor), *value);
		}
		break;
	case 6:
		// \xml <tmc:axis_parameter type="6" name="Maximum current" identifier="par::MaxCurrent">
		// \xml   <desc>Motor current used when motor is running. Where 0 = 3.125% up to 31 = 100% of maximum module current. Calculated by: maximum module current * (value + 1)/ 32.</desc>
		// \xml   <tmc:value_unsigned min="0" max="31" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC5062_FIELD_READ(motor, TMC5062_IHOLD_IRUN(motor), TMC5062_IRUN_MASK, TMC5062_IRUN_SHIFT);
		} else if(readWrite == WRITE) {
			TMC5062_FIELD_UPDATE(motor, TMC5062_IHOLD_IRUN(motor), TMC5062_IRUN_MASK, TMC5062_IRUN_SHIFT, *value);
		}
		break;
	case 7:
		// \xml <tmc:axis_parameter type="7" name="Standby current" identifier="par::StandbyCurrent">
		// \xml   <desc>The current used when the motor is not running. Where 0 = 3.125% up to 31 = 100% of maximum module current. Calculated by: maximum module current * (value + 1)/ 32. This value should be as low as possible so that the motor can cool down when it is not moving. Please see also parameter 214.</desc>
		// \xml   <tmc:value_unsigned min="0" max="31" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC5062_FIELD_READ(motor, TMC5062_IHOLD_IRUN(motor), TMC5062_IHOLD_MASK, TMC5062_IHOLD_SHIFT);
		} else if(readWrite == WRITE) {
			TMC5062_FIELD_UPDATE(motor, TMC5062_IHOLD_IRUN(motor), TMC5062_IHOLD_MASK, TMC5062_IHOLD_SHIFT, *value);
		}
		break;
	case 8:
		// \xml <tmc:axis_parameter type="8" name="Position reached flag" identifier="par::PositionReachedFlag">
		// \xml		<desc>This flag is always set when target position and actual position are equal.</desc>
		// \xml		<tmc:value_bool desc_false="Target position not reached" desc_true="Target position reached" default="false" access="R"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC5062_FIELD_READ(motor, TMC5062_RAMPSTAT(motor), TMC5062_POSITION_REACHED_MASK, TMC5062_POSITION_REACHED_SHIFT);
		} else if(readWrite == WRITE) {
			errors |= ERROR_TYPE;
		}
		break;
	case 10:
		// \xml <tmc:axis_parameter type="10" name="Right endstop" identifier="par::RightEndstop">
		// \xml   <desc>Reference switch right status.</desc>
		// \xml   <tmc:value_bool desc_false="Inactive" desc_true="Active" default="false" access="R"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = !TMC5062_FIELD_READ(motor, TMC5062_RAMPSTAT(motor), TMC5062_STATUS_STOP_R_MASK, TMC5062_STATUS_STOP_R_SHIFT);
		} else if(readWrite == WRITE) {
			errors |= ERROR_TYPE;
		}
		break;
	case 11:
		// \xml <tmc:axis_parameter type="11" name="Left endstop" identifier="par::LeftEndstop">
		// \xml   <desc>Reference switch left status.</desc>
		// \xml   <tmc:value_bool desc_false="Inactive" desc_true="Active" default="false" access="R"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = !TMC5062_FIELD_READ(motor, TMC5062_RAMPSTAT(motor), TMC5062_STATUS_STOP_L_MASK, TMC5062_STATUS_STOP_L_SHIFT);
		} else if(readWrite == WRITE) {
			errors |= ERROR_TYPE;
		}
		break;
	case 12:
		// \xml <tmc:axis_parameter type="12" name="Automatic right stop" identifier="par::AutomaticRightStop">
		// \xml   <desc>Enables automatic motor stop during active right reference switch input.</desc>
		// \xml   <tmc:value_bool desc_false="Inactive" desc_true="Active" default="false" access="RW"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC5062_FIELD_READ(motor, TMC5062_SWMODE(motor), TMC5062_STOP_R_ENABLE_MASK, TMC5062_STOP_R_ENABLE_SHIFT);
		} else if(readWrite == WRITE) {
			TMC5062_FIELD_UPDATE(motor, TMC5062_SWMODE(motor), TMC5062_STOP_R_ENABLE_MASK, TMC5062_STOP_R_ENABLE_SHIFT, *value);
		}
		break;
	case 13:
		// \xml <tmc:axis_parameter type="13" name="Automatic left stop" identifier="par::AutomaticLeftStop">
		// \xml   <desc>Enables automatic motor stop during active left reference switch input.</desc>
		// \xml   <tmc:value_bool desc_false="Inactive" desc_true="Active" default="false" access="RW"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC5062_FIELD_READ(motor, TMC5062_SWMODE(motor), TMC5062_STOP_L_ENABLE_MASK, TMC5062_STOP_L_ENABLE_SHIFT);
		} else if(readWrite == WRITE) {
			TMC5062_FIELD_UPDATE(motor, TMC5062_SWMODE(motor), TMC5062_STOP_L_ENABLE_MASK, TMC5062_STOP_L_ENABLE_SHIFT, *value);
		}
		break;
	case 14:
		// \xml <tmc:axis_parameter type="14" name="SW_MODE Register" identifier="par::SW_MODE">
		// \xml   <desc>Complete SW_MODE reference switch and StallGuard2 event configuration register.</desc>
		// \xml   <tmc:value_unsigned min="0" max="2048" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = tmc5062_readInt(motor, TMC5062_SWMODE(motor));
		} else if(readWrite == WRITE) {
			tmc5062_writeInt(motor, TMC5062_SWMODE(motor), *value);
		}
		break;
	case 15:
		// \xml <tmc:axis_parameter type="15" name="Acceleration A1" identifier="par::A1">
		// \xml   <desc>First acceleration between VSTART and V1 (in position mode only).</desc>
		// \xml   <tmc:value_unsigned min="0" max="#u16_MAX#" default="100" unit="[pps/s]" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = tmc5062_readInt(motor, TMC5062_A1(motor));
		} else if(readWrite == WRITE) {
			tmc5062_writeInt(motor, TMC5062_A1(motor), *value);
		}
		break;
	case 16:
		// \xml <tmc:axis_parameter type="16" name="Velocity V1" identifier="par::V1">
		// \xml   <desc>First acceleration / deceleration phase target velocity (in position mode only). Setting this value to 0 turns off the first acceleration / deceleration phase, maximum acceleration (axis parameter 5) and maximum deceleration (axis parameter 17) are used only.</desc>
		// \xml   <tmc:value_unsigned min="0" max="#u20_MAX#" default="0" unit="[pps]" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = tmc5062_readInt(motor, TMC5062_V1(motor));
		} else if(readWrite == WRITE) {
			tmc5062_writeInt(motor, TMC5062_V1(motor), *value);
		}
		break;
	case 17:
		// \xml <tmc:axis_parameter type="17" name="Maximum Deceleration" identifier="par::MaxDeceleration">
		// \xml   <desc>Maximum deceleration in positioning ramps. Used to decelerate from maximum positioning speed (axis parameter 4) to velocity V1</desc>
		// \xml   <tmc:value_unsigned min="0" max="#u16_MAX#" default="0" unit="[pps/s]" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = tmc5062_readInt(motor, TMC5062_DMAX(motor));
		} else if(readWrite == WRITE) {
			tmc5062_writeInt(motor, TMC5062_DMAX(motor), *value);
		}
		break;
	case 18:
		// \xml <tmc:axis_parameter type="18" name="Deceleration D1" identifier="par::D1">
		// \xml   <desc>Deceleration between V1 and VSTOP (in positioning mode only).</desc>
		// \xml   <tmc:value_unsigned min="1" max="#u16_MAX#" default="200" unit="[pps/s]" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = tmc5062_readInt(motor, TMC5062_D1(motor));
		} else if(readWrite == WRITE) {
			tmc5062_writeInt(motor, TMC5062_D1(motor), *value);
		}
		break;
	case 19:
		// \xml <tmc:axis_parameter type="19" name="VSTART" identifier="par::StartVelocity">
		// \xml   <desc>Motor start velocity.</desc>
		// \xml   <tmc:value_unsigned min="0" max="262143" default="0" unit="[ppT]" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = tmc5062_readInt(motor, TMC5062_VSTART(motor));
		} else if(readWrite == WRITE) {
			tmc5062_writeInt(motor, TMC5062_VSTART(motor), *value);
		}
		break;
	case 20:
		// \xml <tmc:axis_parameter type="20" name="Velocity VSTOP" identifier="par::StopVelocity">
		// \xml   <desc>Motor stop velocity (in position mode only).</desc>
		// \xml   <tmc:value_unsigned min="0" max="262143" default="0" unit="[pps]" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = tmc5062_readInt(motor, TMC5062_VSTOP(motor));
		} else if(readWrite == WRITE) {
			tmc5062_writeInt(motor, TMC5062_VSTOP(motor), *value);
		}
		break;
	case 21:
		// \xml <tmc:axis_parameter type="21" name="Waiting time after ramp down" identifier="par::RampWaitTime">
		// \xml   <desc>Defines the waiting time after ramping down to zero velocity before next movement or direction inversion can start. Time range is about 0 to 2 seconds.</desc>
		// \xml   <tmc:value_unsigned min="0" max="255" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = tmc5062_readInt(motor, TMC5062_TZEROWAIT(motor));
		} else if(readWrite == WRITE) {
			tmc5062_writeInt(motor, TMC5062_TZEROWAIT(motor), *value);
		}
		break;
	case 22:
		// \xml <tmc:axis_parameter type="22" name="smartEnergy threshold speed" identifier="par::smartEnergyThresholdSpeed">
		// \xml   <desc>Above this speed coolStep and stallGuard become enabled.</desc>
		// \xml   <tmc:value_unsigned min="0" max="#u20_MAX#" default="0" unit="[pps]" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = tmc5062_readInt(motor, TMC5062_VCOOLTHRS(motor));
		} else if(readWrite == WRITE) {
			tmc5062_writeInt(motor, TMC5062_VCOOLTHRS(motor), *value);
		}
		break;
	case 23:
		// \xml <tmc:axis_parameter type="23" name="Speed threshold for high speed mode" identifier="par::THIGH">
		// \xml   <desc>Speed threshold for de-activating coolStep and switching into a different chopper mode and fullstepping to maximize torque.</desc>
		// \xml   <tmc:value_unsigned min="0" max="#u20_MAX#" default="0" unit="[pps]" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = tmc5062_readInt(motor, TMC5062_VHIGH(motor));
		} else if(readWrite == WRITE) {
			tmc5062_writeInt(motor, TMC5062_VHIGH(motor), *value);
		}
		break;
	case 24:
		// \xml <tmc:axis_parameter type="24" name="Minimum speed for switching to dcStep" identifier="par::VDCMIN">
		// \xml   <desc>DcStep is enabled above this velocity.</desc>
		// \xml   <tmc:value_unsigned min="0" max="8388096" default="0" unit="[pps]" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = tmc5062_readInt(motor, TMC5062_VDCMIN(motor));
		} else if(readWrite == WRITE) {
			tmc5062_writeInt(motor, TMC5062_VDCMIN(motor), *value);
		}
		break;
	case 28:
		// \xml <tmc:axis_parameter type="28" name="High speed fullstep mode" identifier="par::HighSpeedFullstepMode">
		// \xml   <desc>Switch to fullstep mode when measured speed is higher than axis parameter 23 when set to 1.</desc>
		// \xml   <tmc:value_bool desc_false="Inactive" desc_true="Active" default="false" access="RW"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC5062_FIELD_READ(motor, TMC5062_CHOPCONF(motor), TMC5062_VHIGHFS_MASK, TMC5062_VHIGHFS_SHIFT);
		} else if(readWrite == WRITE) {
			TMC5062_FIELD_UPDATE(motor, TMC5062_CHOPCONF(motor), TMC5062_VHIGHFS_MASK, TMC5062_VHIGHFS_SHIFT, *value);
		}
		break;
	case 29:
		if(readWrite == READ) {
			*value = TMC5062.velocity[motor];
		} else if(readWrite == WRITE) {
			errors |= ERROR_TYPE;
		}
		break;
	case 140:
		// \xml <tmc:axis_parameter type="140" name="Microstep Resolution" identifier="par::MicrostepResolution">
		// \xml   <desc>Microstep resolutions per full step:</desc>
		// \xml   <tmc:value_choice default="256" unit="" access="RW">
		// \xml     <choice_element value="1" desc="fullstep"></choice_element>
		// \xml     <choice_element value="2" desc="halfstep"></choice_element>
		// \xml     <choice_element value="4" desc="4 microsteps"></choice_element>
		// \xml     <choice_element value="8" desc="8 microsteps"></choice_element>
		// \xml     <choice_element value="16" desc="16 microsteps"></choice_element>
		// \xml     <choice_element value="32" desc="32 microsteps"></choice_element>
		// \xml     <choice_element value="64" desc="64 microsteps"></choice_element>
		// \xml     <choice_element value="128" desc="128 microsteps"></choice_element>
		// \xml     <choice_element value="256" desc="256 microsteps"></choice_element>
		// \xml   </tmc:value_choice>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = 256 >> TMC5062_FIELD_READ(motor, TMC5062_CHOPCONF(motor), TMC5062_MRES_MASK, TMC5062_MRES_SHIFT);
		} else if(readWrite == WRITE) {
			switch(*value)
			{
			case 1:    *value = 8;   break;
			case 2:    *value = 7;   break;
			case 4:    *value = 6;   break;
			case 8:    *value = 5;   break;
			case 16:   *value = 4;   break;
			case 32:   *value = 3;   break;
			case 64:   *value = 2;   break;
			case 128:  *value = 1;   break;
			case 256:  *value = 0;   break;
			default:   *value = -1;  break;
			}

			if(*value != -1)
			{
				TMC5062_FIELD_UPDATE(motor, TMC5062_CHOPCONF(motor), TMC5062_MRES_MASK, TMC5062_MRES_SHIFT, *value);
			}
		}
		break;
	case 162:
		// \xml <tmc:axis_parameter type="162" name="Chopper blank time" identifier="par::ChopperBlankTime">
		// \xml   <desc>Selects the comparator blank time. This time needs to safely cover the switching event and the duration of the ringing on the sense resistor. Normally leave at the default value.</desc>
		// \xml   <tmc:value_unsigned min="0" max="3" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC5062_FIELD_READ(motor, TMC5062_CHOPCONF(motor), TMC5062_TBL_MASK, TMC5062_TBL_SHIFT);
		} else if(readWrite == WRITE) {
			TMC5062_FIELD_UPDATE(motor, TMC5062_CHOPCONF(motor), TMC5062_TBL_MASK, TMC5062_TBL_SHIFT, *value);
		}
		break;
	case 163:
		// \xml <tmc:axis_parameter type="163" name="Constant TOff Mode" identifier="par::ConstantTOffMode">
		// \xml   <desc>Selection of the chopper mode.</desc>
		// \xml   <tmc:value_bool desc_false="spread cycle" desc_true="classic constant off time" default="false" access="RW"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC5062_FIELD_READ(motor, TMC5062_CHOPCONF(motor), TMC5062_CHM_MASK, TMC5062_CHM_SHIFT);
		} else if(readWrite == WRITE) {
			TMC5062_FIELD_UPDATE(motor, TMC5062_CHOPCONF(motor), TMC5062_CHM_MASK, TMC5062_CHM_SHIFT, *value);
		}
		break;
	case 164:
		// \xml <tmc:axis_parameter type="164" name="Disable fast decay comparator" identifier="par::DisableFastDecayComparator">
		// \xml   <desc>See parameter 163. For "classic const. off time", setting this parameter to "1" will disable current comparator usage for termination of fast decay cycle.</desc>
		// \xml   <tmc:value_bool desc_false="enable comparator" desc_true="disable comparator" default="false" access="RW"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC5062_FIELD_READ(motor, TMC5062_CHOPCONF(motor), TMC5062_DISFDCC_MASK, TMC5062_DISFDCC_SHIFT);
		} else if(readWrite == WRITE) {
			TMC5062_FIELD_UPDATE(motor, TMC5062_CHOPCONF(motor), TMC5062_DISFDCC_MASK, TMC5062_DISFDCC_SHIFT, *value);
		}
		break;
	case 165:
		// \xml <tmc:axis_parameter type="165" name="Chopper hysteresis end / fast decay time" identifier="par::ChopperHysteresisEnd">
		// \xml   <desc>See parameter 163. For "spread cycle" chopper mode this parameter will set / return the hysteresis end setting (hysteresis end value after a number of decrements). For "classic const. off time" chopper mode this parameter will set / return the fast decay time.</desc>
		// \xml   <tmc:value_unsigned min="0" max="15" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			if(TMC5062_FIELD_READ(motor, TMC5062_CHOPCONF(motor), TMC5062_CHM_MASK, TMC5062_CHM_SHIFT))
			{
				*value = TMC5062_FIELD_READ(motor, TMC5062_CHOPCONF(motor), TMC5062_HEND_MASK, TMC5062_HEND_SHIFT);
			}
			else
			{
				*value = TMC5062_FIELD_READ(motor, TMC5062_CHOPCONF(motor), TMC5062_TFD_ALL_MASK, TMC5062_TFD_ALL_SHIFT);
				if(TMC5062_FIELD_READ(motor, TMC5062_CHOPCONF(motor), TMC5062_TFD_3_MASK, TMC5062_TFD_3_SHIFT))
					*value |= 1<<3; // MSB wird zu value hinzugefgt
			}
		} else if(readWrite == WRITE) {
			if(tmc5062_readInt(motor, TMC5062_CHOPCONF(motor)) & (1<<14))
			{
				TMC5062_FIELD_UPDATE(motor, TMC5062_CHOPCONF(motor), TMC5062_HEND_MASK, TMC5062_HEND_SHIFT, *value);
			}
			else
			{
				TMC5062_FIELD_UPDATE(motor, TMC5062_CHOPCONF(motor), TMC5062_TFD_3_MASK, TMC5062_TFD_3_SHIFT, (*value & 1<<3)? 1:0);
				TMC5062_FIELD_UPDATE(motor, TMC5062_CHOPCONF(motor), TMC5062_TFD_ALL_MASK, TMC5062_TFD_ALL_SHIFT, *value);
			}
		}
		break;
	case 166:
		// \xml <tmc:axis_parameter type="166" name="Chopper hysteresis start / sine wave offset" identifier="par::ChopperHysteresisStart">
		// \xml   <desc>See parameter 163. For "spread cycle" chopper mode this parameter will set / return the Hysteresis start setting (please note that this value is an offset to the hysteresis end value). For "classic const. off time" chopper mode this parameter will set / return the sine wave offset.</desc>
		// \xml   <tmc:value_unsigned min="0" max="8" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			if(tmc5062_readInt(motor, TMC5062_CHOPCONF(motor)) & (1<<14))
			{
				*value = TMC5062_FIELD_READ(motor, TMC5062_CHOPCONF(motor), TMC5062_HSTRT_MASK, TMC5062_HSTRT_SHIFT);
			}
			else
			{
				*value = TMC5062_FIELD_READ(motor, TMC5062_CHOPCONF(motor), TMC5062_OFFSET_MASK, TMC5062_OFFSET_SHIFT);
				if(TMC5062_FIELD_READ(motor, TMC5062_CHOPCONF(motor), TMC5062_TFD_3_MASK, TMC5062_TFD_3_SHIFT))
					*value |= 1<<3;
			}
		} else if(readWrite == WRITE) {
			if(tmc5062_readInt(motor, TMC5062_CHOPCONF(motor)) & (1<<14))
			{
				TMC5062_FIELD_UPDATE(motor, TMC5062_CHOPCONF(motor), TMC5062_HSTRT_MASK, TMC5062_HSTRT_SHIFT, *value);
			}
			else
			{
				TMC5062_FIELD_UPDATE(motor, TMC5062_CHOPCONF(motor), TMC5062_OFFSET_MASK, TMC5062_OFFSET_SHIFT, *value);
			}
		}
		break;
	case 167:
		// \xml <tmc:axis_parameter type="167" name="Chopper off time" identifier="par::TOff">
		// \xml   <desc>The off time setting controls the minimum chopper frequency. An off time within the range of 5us to 20us will fit. Off time setting for constant t Off chopper: NCLK = 12 + 32  tOFF (Minimum is 64 clocks) Setting this parameter to zero completely disables all driver transistors and the motor can freewheel.</desc>
		// \xml   <tmc:value_unsigned min="0" max="15" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC5062_FIELD_READ(motor, TMC5062_CHOPCONF(motor), TMC5062_TOFF_MASK, TMC5062_TOFF_SHIFT);
		} else if(readWrite == WRITE) {
			TMC5062_FIELD_UPDATE(motor, TMC5062_CHOPCONF(motor), TMC5062_TOFF_MASK, TMC5062_TOFF_SHIFT, *value);
		}
		break;
	case 168:
		// \xml <tmc:axis_parameter type="168" name="smartEnergy current minimum (SEIMIN)" identifier="par::SEIMIN">
		// \xml   <desc>Sets the lower motor current limit for coolStep operation by scaling the maximum current (see axis parameter 6) value.</desc>
		// \xml   <tmc:value_bool desc_false="1/2 of CS" desc_true="1/4 of CS" default="false" access="RW"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC5062_FIELD_READ(motor, TMC5062_COOLCONF(motor), TMC5062_SEIMIN_MASK, TMC5062_SEIMIN_SHIFT);
		} else if(readWrite == WRITE) {
			TMC5062_FIELD_UPDATE(motor, TMC5062_COOLCONF(motor), TMC5062_SEIMIN_MASK, TMC5062_SEIMIN_SHIFT, *value);
		}
		break;
	case 169:
		// \xml <tmc:axis_parameter type="169" name="smartEnergy current down step" identifier="par::SECDS">
		// \xml   <desc>Sets the number of stallGuard2 readings above the upper threshold necessary for each current decrement of the motor current. Number of stallGuard2 measurements per decrement: Scaling: 0. . . 3: 32, 8, 2, 1. 0: slow decrement, 3: fast decrement</desc>
		// \xml   <tmc:value_unsigned min="0" max="3" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC5062_FIELD_READ(motor, TMC5062_COOLCONF(motor), TMC5062_SEDN_MASK, TMC5062_SEDN_SHIFT);
		} else if(readWrite == WRITE) {
			TMC5062_FIELD_UPDATE(motor, TMC5062_COOLCONF(motor), TMC5062_SEDN_MASK, TMC5062_SEDN_SHIFT, *value);
		}
		break;
	case 170:
		// \xml <tmc:axis_parameter type="170" name="smartEnergy hysteresis" identifier="par::smartEnergyHysteresis">
		// \xml   <desc>Sets the distance between the lower and the upper threshold for stallGuard2 reading. Above the upper threshold the motor current becomes decreased. Hysteresis: ([AP172] + 1) * 32. Upper stallGuard threshold: ([AP172] + [AP170] + 1) * 32</desc>
		// \xml   <tmc:value_unsigned min="0" max="15" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC5062_FIELD_READ(motor, TMC5062_COOLCONF(motor), TMC5062_SEMAX_MASK, TMC5062_SEMAX_SHIFT);
		} else if(readWrite == WRITE) {
			TMC5062_FIELD_UPDATE(motor, TMC5062_COOLCONF(motor), TMC5062_SEMAX_MASK, TMC5062_SEMAX_SHIFT, *value);
		}
		break;
	case 171:
		// \xml <tmc:axis_parameter type="171" name="smartEnergy current up step" identifier="par::SECUS">
		// \xml   <desc>Sets the current increment step. The current becomes incremented for each measured stallGuard2 value below the lower threshold see smartEnergy hysteresis start). Current increment step size: Scaling: 0. . . 3: 1, 2, 4, 8. 0: slow increment, 3: fast increment / fast reaction to rising load</desc>
		// \xml   <tmc:value_unsigned min="0" max="3" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC5062_FIELD_READ(motor, TMC5062_COOLCONF(motor), TMC5062_SEUP_MASK, TMC5062_SEUP_SHIFT);
		} else if(readWrite == WRITE) {
			TMC5062_FIELD_UPDATE(motor, TMC5062_COOLCONF(motor), TMC5062_SEUP_MASK, TMC5062_SEUP_SHIFT, *value);
		}
		break;
	case 172:
		// \xml <tmc:axis_parameter type="172" name="smartEnergy hysteresis start" identifier="par::smartEnergyHysteresisStart">
		// \xml   <desc>The lower threshold for the stallGuard2 value (see smart Energy current up step).</desc>
		// \xml   <tmc:value_unsigned min="0" max="15" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC5062_FIELD_READ(motor, TMC5062_COOLCONF(motor), TMC5062_SEMIN_MASK, TMC5062_SEMIN_SHIFT);
		} else if(readWrite == WRITE) {
			TMC5062_FIELD_UPDATE(motor, TMC5062_COOLCONF(motor), TMC5062_SEMIN_MASK, TMC5062_SEMIN_SHIFT, *value);
		}
		break;
	case 173:
		// \xml <tmc:axis_parameter type="173" name="stallGuard2 filter enable" identifier="par::SG2FilterEnable">
		// \xml   <desc>Enables the stallGuard2 filter for more precision of the movement. If set, reduces the measurement frequency to one measurement per four fullsteps. In most cases it is expedient to set the filtered mode before using coolStep. Use the standard mode for step loss detection.</desc>
		// \xml   <tmc:value_bool desc_false="standard mode" desc_true="filtered mode" default="false" access="RW"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC5062_FIELD_READ(motor, TMC5062_COOLCONF(motor), TMC5062_SFILT_MASK, TMC5062_SFILT_SHIFT);
		} else if(readWrite == WRITE) {
			TMC5062_FIELD_UPDATE(motor, TMC5062_COOLCONF(motor), TMC5062_SFILT_MASK, TMC5062_SFILT_SHIFT, *value);
		}
		break;
	case 174:
		// \xml <tmc:axis_parameter type="174" name="stallGuard2 threshold" identifier="par::SG2Threshold">
		// \xml   <desc>This signed value controls stallGuard2 threshold level for stall output and sets the optimum measurement range for readout. A lower value gives a higher sensitivity. Zero is the starting value. A higher value makes stallGuard2 less sensitive and requires more torque to indicate a stall.</desc>
		// \xml   <tmc:value_signed min="-64" max="63" default="0" unit="" access="RW"></tmc:value_signed>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC5062_FIELD_READ(motor, TMC5062_COOLCONF(motor), TMC5062_SGT_MASK, TMC5062_SGT_SHIFT);
			*value = CAST_Sn_TO_S32(*value, 7);
		} else if(readWrite == WRITE) {
			TMC5062_FIELD_UPDATE(motor, TMC5062_COOLCONF(motor), TMC5062_SGT_MASK, TMC5062_SGT_SHIFT, *value);
		}
		break;
	case 179:
		// \xml <tmc:axis_parameter type="179" name="VSense" identifier="par::VSense">
		// \xml   <desc>Sense resistor voltage:</desc>
		// \xml   <tmc:value_bool desc_false="High sense resistor voltage" desc_true="Low sense resistor voltage" default="false" access="RW"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC5062_FIELD_READ(motor, TMC5062_CHOPCONF(motor), TMC5062_VSENSE_MASK, TMC5062_VSENSE_SHIFT);
		} else if(readWrite == WRITE) {
			TMC5062_FIELD_UPDATE(motor, TMC5062_CHOPCONF(motor), TMC5062_VSENSE_MASK, TMC5062_VSENSE_SHIFT, *value);
		}
		break;
	case 180:
		// \xml <tmc:axis_parameter type="180" name="smartEnergy actual current" identifier="par::smartEnergyActualCurrent">
		// \xml   <desc>This status value provides the actual motor current setting as controlled by coolStep. The value goes up to the CS value and down to the portion of CS as specified by SEIMIN. Actual motor current scaling factor: 0. . . 31: 1/32, 2/32, . . . 32/32</desc>
		// \xml   <tmc:value_unsigned min="0" max="31" default="0" unit="" access="R"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC5062_FIELD_READ(motor, TMC5062_DRVSTATUS(motor), TMC5062_CS_ACTUAL_MASK, TMC5062_CS_ACTUAL_SHIFT);
		} else if(readWrite == WRITE) {
			errors |= ERROR_TYPE;
		}
		break;
	case 181:
		// \xml <tmc:axis_parameter type="181" name="reset stall" identifier="par::resetStall">
		// \xml   <desc>Resets stallguard stall so that the motor can move again.</desc>
		// \xml   <tmc:value_unsigned min="0" max="#TMC5062_MAX_VELOCITY#" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			if( tmc5062_readInt(motor, TMC5062_SWMODE(motor)) & (1<<10))
			{
				*value = tmc5062_readInt(motor, TMC5062_VCOOLTHRS(motor));
			}
			else
				tempValue = 0;
		} else if(readWrite == WRITE) {
			tmc5062_writeInt(motor, TMC5062_VCOOLTHRS(motor),*value);
			TMC5062_FIELD_UPDATE(motor, TMC5062_SWMODE(motor), TMC5062_SG_STOP_MASK, TMC5062_SG_STOP_SHIFT, (*value)? 1:0);
		}
		break;
	case 182:
		// \xml <tmc:axis_parameter type="182" name="smartEnergy threshold speed" identifier="par::smartEnergyThresholdSpeed">
		// \xml   <desc>Above this speed coolStep becomes enabled.</desc>
		// \xml   <tmc:value_unsigned min="0" max="#TMC5062_MAX_VELOCITY#" default="0" unit="[pps]" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = tmc5062_readInt(motor, TMC5062_VCOOLTHRS(motor));
		} else if(readWrite == WRITE) {
			tmc5062_writeInt(motor, TMC5062_VCOOLTHRS(motor),*value);
		}
		break;
	case 184:
		// \xml <tmc:axis_parameter type="184" name="Random TOff mode" identifier="par::RandomTOffMode">
		// \xml   <desc>Enable / disable random TOff mode.</desc>
		// \xml   <tmc:value_bool desc_false="Chopper off time is fixed" desc_true="Chopper off time is random" default="false" access="RW"></tmc:value_bool>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC5062_FIELD_READ(motor, TMC5062_CHOPCONF(motor), TMC5062_RNDTF_MASK, TMC5062_RNDTF_SHIFT);
		} else if(readWrite == WRITE) {
			TMC5062_FIELD_UPDATE(motor, TMC5062_CHOPCONF(motor), TMC5062_RNDTF_MASK, TMC5062_RNDTF_SHIFT, *value);
		}
		break;
	case 185:
		// \xml <tmc:axis_parameter type="185" name="Chopper synchronization" identifier="par::ChopperSynchronization">
		// \xml   <desc>This parameter allows synchronization of the chopper for both phases of a two phase motor in order to avoid the occurrence of a beat, especially at low velocities. 0: chopper sync function chopSync off 1. . . 15: chopper synchronization</desc>
		// \xml   <tmc:value_unsigned min="0" max="15" default="0" unit="" access="RW"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC5062_FIELD_READ(motor, TMC5062_CHOPCONF(motor), TMC5062_SYNC_MASK, TMC5062_SYNC_SHIFT);
		} else if(readWrite == WRITE) {
			TMC5062_FIELD_UPDATE(motor, TMC5062_CHOPCONF(motor), TMC5062_SYNC_MASK, TMC5062_SYNC_SHIFT, *value);
		}
		break;
	case 206:
		// \xml <tmc:axis_parameter type="206" name="Load value" identifier="par::LoadValue">
		// \xml   <desc>Actual current control scaling for monitoring smart energy current scaling or automatic current scaling.</desc>
		// \xml   <tmc:value_unsigned min="0" max="#u10_MAX#" default="0" unit="" access="R"></tmc:value_unsigned>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = TMC5062_FIELD_READ(motor, TMC5062_DRVSTATUS(motor), TMC5062_SG_RESULT_MASK, TMC5062_SG_RESULT_SHIFT);
		} else if(readWrite == WRITE) {
			errors |= ERROR_TYPE;
		}
		break;
	case 209:
		// \xml <tmc:axis_parameter type="209" name="Encoder position" identifier="par::EncoderPosition">
		// \xml   <desc>The current encoder position</desc>
		// \xml   <tmc:value_signed min="#s32_MIN#" max="#s32_MAX#" default="0" unit="[usteps]" access="RW"></tmc:value_signed>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = tmc5062_readInt(motor, TMC5062_XENC(motor));
		} else if(readWrite == WRITE) {
			tmc5062_writeInt(motor, TMC5062_XENC(motor),*value);
		}
		break;
	case 210:
		// \xml <tmc:axis_parameter type="210" name="Encoder Resolution" identifier="par::EncoderResolution">
		// \xml   <desc>Resolution of the encoder.</desc>
		// \xml   <tmc:value_signed min="#s32_MIN#" max="#s32_MAX#" default="0" unit="" access="RW"></tmc:value_signed>
		// \xml </tmc:axis_parameter>
		if(readWrite == READ) {
			*value = tmc5062_readInt(motor, TMC5062_ENC_CONST(motor));
		} else if(readWrite == WRITE) {
			tmc5062_writeInt(motor, TMC5062_ENC_CONST(motor),*value);
		}
		break;
	case 211:
		if(readWrite == READ) {
			// encoder enable
			switch(motor)
			{
			case 0:
				tempValue = tmc5062_readInt(motor, TMC5062_GCONF);
				tempValue &= (1<<3) | (1<<4);
				*value = (tempValue == (1<<4))? 1 : 0;
				break;
			case 1:
				tempValue = tmc5062_readInt(motor, TMC5062_GCONF);
				tempValue &= (1<<5) | (1<<6);
				*value = (tempValue == ((1<<5) | (0<<6)))? 1 : 0;
				break;
			}
		} else if(readWrite == WRITE) {
			// encoder enable
			switch(motor)
			{
			case 0:
				tempValue = tmc5062_readInt(motor, TMC5062_GCONF);
				if(*value)
					tempValue = (tempValue & ~(1<<3)) | (1<<4);
				else
					tempValue = (tempValue | (1<<3)) & ~(1<<4);
				tmc5062_writeInt(motor, TMC5062_GCONF, tempValue);
				break;
			case 1:		// enable ENCODER2 - disable REF
				tempValue = tmc5062_readInt(motor, TMC5062_GCONF);
				if(*value)
					//tempValue = (tempValue | (1<<5)) & ~(5<<5);
					tempValue = (tempValue | (1<<5)) & ~(1<<6); //todo: CHECK 3: Sind die nderungen richtig? Codemig macht es so Sinn, aber die Bits sind in der Dokumentation als reserved markiert (LH) #3
				else
					//tempValue = (tempValue & ~(1<<6)) | ~(1<<6);
					tempValue = (tempValue & ~(1<<5)) | (1<<6); //todo: CHECK 3: Sind die nderungen richtig? Codemig macht es so Sinn, aber die Bits sind in der Dokumentation als reserved markiert (LH) #4
				tmc5062_writeInt(motor, TMC5062_GCONF, tempValue);
				break;
			}
		}
		break;
	default:
		errors |= ERROR_TYPE;
		break;
	}
	// \xml </axis_parameters>

	// \xml <axis_parameters axis="1">
	// \xml 	~~<axis_parameters axis="0">
	// \xml </axis_parameters>
	return errors;
}
// \xml </tmc:product_module>

static uint32 SAP(uint8 type, uint8 motor, int32 value)
{
	return handleParameter(WRITE, motor, type, &value);
}

static uint32 GAP(uint8 type, uint8 motor, int32 *value)
{
	return handleParameter(READ, motor, type, value);
}

static uint32 getMeasuredSpeed(uint8 motor, int32 *value)
{
	if(motor >= MOTORS)
		return ERROR_MOTOR;

	*value = TMC5062.velocity[motor];

	return ERROR_NONE;
}

static void writeRegister(uint8 address, int32 value)
{
	tmc5062_writeInt(0, address, value);
}

static void readRegister(uint8 address, int32 *value)
{
	*value	= tmc5062_readInt(0, address);
}

static void periodicJob(uint32 tick)
{
	for(int motor = 0; motor < MOTORS; motor++)
		tmc5062_periodicJob(motor, tick, &TMC5062, TMC5062_config);
}

static void checkErrors(uint32 tick)
{
	UNUSED(tick);
	Evalboards.ch1.errors = 0;
}

static uint32 userFunction(uint8 type, uint8 motor, int32 *value)
{
	uint32 errors = 0;

	UNUSED(motor);

	switch(type)
	{
	case 0:		// enable Single Wire Mode
		if(*value)
			HAL.IOs->config->setHigh(Pins.SWSEL);
		else
			HAL.IOs->config->setLow(Pins.SWSEL);
		break;
	case 1:		// read interrupt pin INT
		*value = (HAL.IOs->config->isHigh(Pins.INT_ENCA))? 1 : 0;
		break;
	case 2:		// read position compare pin PP
		*value = (HAL.IOs->config->isHigh(Pins.PP_ENCB))? 1 : 0;
		break;
	default:
		errors |= ERROR_TYPE;
		break;
	}
	return errors;
}

static void deInit(void)
{
	enableDriver(DRIVER_DISABLE);
	HAL.IOs->config->reset(Pins.DRV_ENN);
	HAL.IOs->config->reset(Pins.INT_ENCA);
	HAL.IOs->config->reset(Pins.PP_ENCB);
	HAL.IOs->config->reset(Pins.SWION);
	HAL.IOs->config->reset(Pins.SWIOP1);
	HAL.IOs->config->reset(Pins.SWIOP2);
	HAL.IOs->config->reset(Pins.SWSEL);
};

static uint8 reset()
{
	for(uint8 motor = 0; motor < MOTORS; motor++)
		if(tmc5062_readInt(motor, TMC5062_VACTUAL(motor)) != 0)
			return 0;

	return tmc5062_reset(TMC5062_config);
}

static uint8 restore()
{
	return tmc5062_restore(TMC5062_config);
}

static void enableDriver(DriverState state)
{
	if(state == DRIVER_USE_GLOBAL_ENABLE)
		state = Evalboards.driverEnable;

	if(state ==  DRIVER_DISABLE)
		HAL.IOs->config->setHigh(Pins.DRV_ENN);
	else if((state == DRIVER_ENABLE) && (Evalboards.driverEnable == DRIVER_ENABLE))
		HAL.IOs->config->setLow(Pins.DRV_ENN);
}

void TMC5062_init(void)
{
	tmc5062_initConfig(&TMC5062);

	Pins.DRV_ENN   = &HAL.IOs->pins->DIO0;
	Pins.INT_ENCA  = &HAL.IOs->pins->DIO5;
	Pins.PP_ENCB   = &HAL.IOs->pins->DIO6;
	Pins.SWSEL     = &HAL.IOs->pins->DIO16;
	Pins.SWIOP1    = &HAL.IOs->pins->DIO17;
	Pins.SWIOP2    = &HAL.IOs->pins->DIO18;
	Pins.SWION     = &HAL.IOs->pins->DIO19;

	HAL.IOs->config->toOutput(Pins.DRV_ENN);
	HAL.IOs->config->toOutput(Pins.SWSEL);

	HAL.IOs->config->setLow(Pins.SWSEL);

	HAL.IOs->config->toInput(Pins.INT_ENCA);
	HAL.IOs->config->toInput(Pins.PP_ENCB);
	HAL.IOs->config->toInput(Pins.SWION);
	HAL.IOs->config->toInput(Pins.SWIOP1);
	HAL.IOs->config->toInput(Pins.SWIOP2);

	TMC5062_SPIChannel = &HAL.SPI->ch1;
	TMC5062_SPIChannel->CSN = &HAL.IOs->pins->SPI1_CSN;

	TMC5062_config = Evalboards.ch1.config;

	Evalboards.ch1.config->reset        = reset;
	Evalboards.ch1.config->restore      = restore;
	Evalboards.ch1.config->state        = CONFIG_RESET;
	Evalboards.ch1.config->configIndex  = 0;

	Evalboards.ch1.rotate               = rotate;
	Evalboards.ch1.right                = right;
	Evalboards.ch1.left                 = left;
	Evalboards.ch1.stop                 = stop;
	Evalboards.ch1.GAP                  = GAP;
	Evalboards.ch1.SAP                  = SAP;
	Evalboards.ch1.moveTo               = moveTo;
	Evalboards.ch1.moveBy               = moveBy;
	Evalboards.ch1.writeRegister        = writeRegister;
	Evalboards.ch1.readRegister         = readRegister;
	Evalboards.ch1.periodicJob          = periodicJob;
	Evalboards.ch1.userFunction         = userFunction;
	Evalboards.ch1.getMeasuredSpeed     = getMeasuredSpeed;
	Evalboards.ch1.enableDriver         = enableDriver;
	Evalboards.ch1.checkErrors          = checkErrors;
	Evalboards.ch1.numberOfMotors       = MOTORS;
	Evalboards.ch1.VMMin                = VM_MIN;
	Evalboards.ch1.VMMax                = VM_MAX;
	Evalboards.ch1.deInit               = deInit;

	enableDriver(DRIVER_ENABLE);
};
